#include <cstdio>
#include <cctype>
#include <algorithm>
#include <functional>
#include <unordered_set>
#include <set>

#include "assert.h"

#include "binding.hh"
#include "typenameof.hh"
#include "compile.hh"
#include "dprintf.hh"
#include "util.hh"

static bool DynamicMode = false;
static struct Scope RootScope_Actual;
static Scope* RootScope    = &RootScope_Actual;
static Scope* CurrentScope = &RootScope_Actual;
static std::string CurrentSubBaseName;
extern unsigned option_base; // from parser.cc

void SetMetaDynamic()
{
    //fprintf(stderr, "Meta $DYNAMIC\n");
    DynamicMode = true;
}

void SetMetaStatic()
{
    //fprintf(stderr, "Meta $STATIC\n");
    DynamicMode = false;
}

bool IsMetaDynamic() { return DynamicMode; }

Scope* GetCurrentScope()
{
    return CurrentScope;
}

Scope* GetRootScope()
{
    return RootScope;
}

std::string DeclareAsLabelInCurrentScope(const std::string& varname, bool is_definition)
{
    auto i = CurrentScope->labelnames.find(varname);
    if(i != CurrentScope->labelnames.end()) return i->second;

    i = RootScope->labelnames.find(varname);
    if(i != RootScope->labelnames.end()) return i->second;
    //if(&scope == &*RootScope && is_definition) return "";

    std::string name = "L_" + UCase(varname);
    CurrentScope->labelnames[varname] = name;
    CurrentScope->labels[name].defined = false;
    return name;
}

bool AnchorLabelHere(const std::string& l)
{
    /* Find from an existing scope */
    auto i = CurrentScope->labels.find(l);
    if(i == CurrentScope->labels.end())
    {
        i = RootScope->labels.find(l);
        if(i == RootScope->labels.end())
        {
            LabelDef& def = CurrentScope->labels[l];
        #ifdef BINDING_DEBUG
            dprintf(dp_declare, "Defining label %s\n", l.c_str());
        #endif
            def.defined = true;
            def.dataptr = global_data.size();
            return true;
        }
    }
    if(i->second.defined)
    {
        dprintf(dp_declare, "Label %s already defined\n", i->first.c_str());
        return false;
    }
    i->second.defined = true;
    i->second.dataptr = global_data.size();
    return true;
}

const UserType* FindUserType(const std::string& name)
{
    const std::string basename = UCase(StripTypeSuffix(name));

    auto i = CurrentScope->types.find(basename);
    if(i != CurrentScope->types.end()) return i->second.get();

    i = RootScope->types.find(basename);
    if(i != RootScope->types.end()) return i->second.get();

    return nullptr;
}

bool DefineUserType(const std::string& name, UtypeP type)
{
    if(FindUserType(name)) return false;

    const std::string basename = UCase(StripTypeSuffix(name));

    GetRootScope()->types[basename] = type;

    return true;
}

expr* FindConstant(const std::string& name, bool duplicate_it)
{
    expr* result = nullptr;
    const std::string basename = UCase(StripTypeSuffix(name));

    auto i = CurrentScope->constants.find(basename);
    if(i != CurrentScope->constants.end())
        result = i->second.get();
    else
    {
        i = RootScope->constants.find(basename);
        if(i != RootScope->constants.end())
            result = i->second.get();
    }
    if(result && duplicate_it)
    {
        if(expr_number* n = dynamic_cast<expr_number*> (result))
            return new expr_number(*n);
        if(expr_string* n = dynamic_cast<expr_string*> (result))
            return new expr_string(*n);

        assert(!"Internal Error: Constants should be either numeric or numeric");
    }
    return result;
}

int DefineConstant(const std::string& name, exprp&& p)
{
    const std::string basename = UCase(StripTypeSuffix(name));

    auto i = CurrentScope->usage.find(basename);
    if(i != CurrentScope->usage.end())
    {
        const TypeFlagSet& tf = i->second;
        if(tf.is_scalar || tf.is_const || tf.is_subroutine) return 0;
    }
    i = RootScope->usage.find(basename);
    if(i != RootScope->usage.end())
    {
        const TypeFlagSet& tf = i->second;
        if(tf.is_scalar || tf.is_const || tf.is_subroutine) return 0;
    }

    Scope* scope = GetCurrentScope();

    if(HasTypeSuffix(name))
    {
        TypeP target_type = CreateTypeTemplate(name);
        TypeP orig_type   = p->GetType();
        if(*target_type != *orig_type)
        {
            switch(target_type->Type)
            {
                case TypeType::String:
                {
                    /* If the target type is String, and the expression
                     * is not a String, there's nothing we can do to make
                     * it a String.
                     */
                    return -1;
                }
                /* With numeric types, we can delegate the work to these functions. */
                case TypeType::Byte:
                case TypeType::UnsignedByte:
                case TypeType::Integer: p.reset(new expr_builtin_numeric(LexLocation(), CINT, std::move(p))); break;

                case TypeType::Single:  p.reset(new expr_builtin_numeric(LexLocation(), CSNG, std::move(p))); break;

                case TypeType::UnsignedLong:
                case TypeType::Double:  p.reset(new expr_builtin_numeric(LexLocation(), CDBL, std::move(p))); break;

                case TypeType::UnsignedInteger:
                case TypeType::Long:    p.reset(new expr_builtin_numeric(LexLocation(), CLNG, std::move(p))); break;
                /* If the functions comply, they will return a const, which we check
                 * below. If there's a type mismatch, they will not return a const.
                 */
                default: break;
            }
            OptimizeExprp(p);
            if(!p->is_const()) return -1;
        }
    }
    scope->constants[basename] = std::move(p);
    scope->usage[basename].is_const = true;

    return 1;
}

VarContext FindVariableFromScope(const std::string& name, bool match_arrays)
{
    const std::string basename = UCase(StripTypeSuffix(name));

    VarContext result;
    result.is_global = false;

    Scope* scope = GetCurrentScope();
    bool force_suffix = false;

recheck_context:;
    auto i = scope->usage.find(basename);
    if(i != scope->usage.end())
    {
        const TypeFlagSet& tf = i->second;

        if(match_arrays ? tf.is_array : tf.is_scalar)
            goto found_context;

        if(match_arrays ? tf.is_global_array : tf.is_global_scalar)
        {
            if(scope != GetRootScope())
            {
                scope = GetRootScope();
                result.is_global = true;
                goto recheck_context;
            }
            goto found_context;
        }
    }
    return result;

found_context:
recheck_typesuffix:
    std::string typesuffix = CalcTypeSuffix(name, force_suffix);
    /* First attempt to find the right scope using explicit type, if there is one */
    const auto& c = scope->Scopes[(unsigned)TypeSuffixIntoType(typesuffix).Type][match_arrays];
    /*fprintf(stderr, "Trying to find '%s'(%s) -> %s\n", basename.c_str(), typesuffix.c_str(),
                    c.find(basename) != c.end() ? "found" : "not found");*/

    auto j = c.find(basename);
    if(j != c.end())
    {
        result.info = &j->second;
    }
    else if(typesuffix.empty())
    {
        /* If it doesn't work, tack a type suffix and try again */
        force_suffix = true;
        goto recheck_typesuffix;
    }
    return result;
}

TypeP ImplicitlyDefineScalar(const std::string& name, bool is_static, bool is_shared, TypeP type, LexLocation context)
{
    if(!type) type = CreateTypeTemplate(name);
#ifdef BINDING_DEBUG
    dprintf(dp_declare, "Implicit scalar: %s, type = %s\n", name.c_str(), type ? type->Dump().c_str() : "(null)");
#endif
    return ExplicitlyDefine(name, type, is_static, is_shared, true, context);
}

TypeP ImplicitlyDefineArray(const std::string& name,
                            unsigned num_dimensions,
                            long starting_index,
                            unsigned ending_index,
                            bool is_static,
                            bool is_shared, TypeP type, LexLocation context)
{
#ifdef BINDING_DEBUG
    dprintf(dp_declare, "Implicit array: %s\n", name.c_str());
#endif
    if(!type) type = CreateTypeTemplate(name);
    type->arraytype = IsMetaDynamic() ? TypeInfo::DynamicArray : TypeInfo::StaticArray;
    type->dimensions.resize(num_dimensions,
                            std::make_pair((long)(starting_index),
                                           (unsigned)(ending_index-starting_index+1))
                           );
    return ExplicitlyDefine(name, type, is_static, is_shared, true, context);
}

TypeP ExplicitlyDefine(const std::string& name, TypeP type,
                       bool is_static,
                       bool is_shared,
                       bool was_implicit,
                       LexLocation context)
{
    auto res = ExplicitlyDefine_C(context, name,type,is_static,is_shared, was_implicit);
    if(res) (const_cast<VarContextDef*> (res))->defined_at = context;
    return res ? res->type : nullptr;
}

const VarContextDef* ExplicitlyDefine_C(const LexLocation& lexloc,
                                        const std::string& name, TypeP type,
                                        bool is_static,
                                        bool is_shared,
                                        bool was_implicit,
                                        int param_index)
{
    const std::string basename = UCase(StripTypeSuffix(name));

#ifdef BINDING_DEBUG
    dprintf(dp_declare, "Defining '%s'%s as '%s'%s%s (paramindex=%d)\n",
        name.c_str(),
        was_implicit ? " implicitly" : "",
        type ? type->Dump().c_str() : "(null)",
        is_static ? " (static)":"",
        is_shared ? " (shared)":"",
        param_index
    );
#endif
    if(was_implicit && param_index==-1) // param_index check should be unnecessary here
    {
        dprintf(dp_declare, "%s: Warning: In %s, implicitly declaring %s%s %s%s\n",
            lexloc.stringify().c_str(),
            CurrentSubBaseName.empty() ? "global scope" : CurrentSubBaseName.c_str(),
            is_static ? "STATIC ":"",
            type ? type->Dump().c_str() : "(null)",
            name.c_str(),
            is_shared ? " (SHARED)":"");
    }

    bool is_array = type->is_array();
    if(FindVariableFromScope(name, is_array)) return 0;

    // If it's a const or a subroutine in any scope, bail out
    auto i = CurrentScope->usage.find(basename);
    if(i != CurrentScope->usage.end())
    {
        const TypeFlagSet& tf = i->second;
        if(tf.is_subroutine || (!is_array && tf.is_const)) return nullptr;
    }
    i = RootScope->usage.find(basename);
    if(i != RootScope->usage.end())
    {
        const TypeFlagSet& tf = i->second;
        if(tf.is_subroutine || (!is_array && tf.is_const)) return nullptr;
    }

    Scope* scope = GetCurrentScope();

    VarContextDef& context = (scope->Scopes[(unsigned)type->Type][is_array]) [basename];

    context.type = std::make_shared<TypeInfo>(*type);
    context.is_static = is_static;
    context.is_shared = is_shared;
    context.param_index = param_index;

    if(is_array)
    {
        if(was_implicit)
        {
            if(type->arraytype != TypeInfo::ParamArray)
                type->arraytype = DynamicMode ? TypeInfo::DynamicArray : TypeInfo::StaticArray;
        }
        scope->usage[basename].is_array  = true;

        if(is_shared) scope->usage[basename].is_global_array = true;
    }
    else
    {
        scope->usage[basename].is_scalar = true;

        if(is_shared) scope->usage[basename].is_global_scalar = true;
    }

    if(!HasTypeSuffix(name))
    {
        (scope->Scopes[(unsigned)TypeType::UserType][is_array]) [basename] = context;
        /* Explicit type: Add to Utypes,
         * so that it is matched when type suffix is omitted.
         */
    }
    return &context;
}

const SubRoutine* DefineSubFunc(bool is_declare,
                                const std::string& name,
                                TypeP returntype,
                                const std::vector<SubRoutine::ParamSpec>& params,
                                bool is_static,
                                int callconvention,
                                const std::string& linkername)
{
    const std::string basename = UCase(StripTypeSuffix(name));

    Scope& scope = *GetRootScope();
    TypeFlagSet& typeset = scope.usage[basename];

    bool first_def = !typeset.is_subroutine;

    // Not first definition?
    if(!first_def)
    {
        if(is_declare) return 0;        // DECLARE should be the first mention!
        if(typeset.has_body) return 0;  // Anything is useless after the sub has a body
    }

    if(typeset.is_const || typeset.is_scalar) return 0;

    if(!returntype && HasTypeSuffix(name)) return 0; // TODO: Should be caught by parser

    typeset.is_subroutine = true;

    SubRoutine& sub = scope.subroutines[basename];

    if(first_def)
    {
        sub.returntype     = returntype;
        sub.params         = params;
        sub.callconvention = callconvention;
        sub.linkername     = linkername;
    }
    else
    {
        /* Check that the return type still matches */
        if(!returntype != !sub.returntype
        || (returntype && *sub.returntype  != *returntype))
        {
            dprintf(dp_errors, "Error: Return type not matching\n");
            return 0;
        }

        /* Check that the parameter types still match */
        for(std::size_t a=0; a<params.size(); ++a)
        {
            if(params[a].passconvention != sub.params[a].passconvention
            && params[a].passconvention != SubRoutine::ParamSpec::ByRef)
            {
                // If BYVAL was used in DECLARE but omitted in SUB, we allow that.
                dprintf(dp_errors, "Error: Param %u: Passing convention does not match?\n", (unsigned)a);
                return 0;
            }
            if(!params[a].type)
            {
                dprintf(dp_errors, "Error: Param %u: No type?\n", (unsigned)a);
                return 0;
            }

            if(!sub.params[a].type)
            {
                dprintf(dp_warnings, "Warning: Param %u refined as %s\n", (unsigned)a, params[a].type->Dump().c_str());
                sub.params[a].type = params[a].type;
            }
            else if(*params[a].type != *sub.params[a].type)
            {
                dprintf(dp_errors, "Error: Param %u: Type mismatch? (%s vs %s)\n", (unsigned)a,
                    sub.params[a].type->Dump().c_str(),
                    params[a].type->Dump().c_str());
                return 0;
            }
        }
    }

    for(std::size_t a=0; a<params.size(); ++a)
    {
        const std::string oldname = sub.params[a].name;

        sub.params[a].name = UCase(StripTypeSuffix(params[a].name));

        if(!first_def && oldname != sub.params[a].name)
        {
            dprintf(dp_warnings, "Warning: Param %u name changed from %s to %s\n",
                (unsigned)a,
                oldname.c_str(),
                sub.params[a].name.c_str());
        }
    }

    if(!is_declare)
    {
        /* Check for duplicate parameter names */
        std::unordered_set<std::string> names;
        for(std::size_t a=0; a<params.size(); ++a)
        {
            const std::string& name = sub.params[a].name;
            if(names.find(name) != names.end()) return 0;
            names.insert(name);
        }
    }

    sub.is_static = is_static;

    return &sub;
}

void EnterSubFuncScope(const std::string& name, bool everything_is_shared)
{
    const std::string basename = UCase(StripTypeSuffix(name));
    Scope& rootscope = *GetRootScope();
    assert(&*GetCurrentScope() == &rootscope);

    //dprintf(dp_rulecomego, "Enter sub '%s' (%s)\n", name.c_str(), basename.c_str());

    TypeFlagSet& typeset = rootscope.usage[basename];
    assert(typeset.is_subroutine);
    assert(!typeset.has_body);
    assert(CurrentSubBaseName.empty());

    CurrentSubBaseName = basename;

    SubRoutine& subdata = rootscope.subroutines[basename];

    typeset.has_body = true;

    Scope& subscope = *(rootscope.subroutines[basename].sub_scope = new Scope);

    std::unordered_set<std::string> param_names[2];
    for(const auto& p: subdata.params) param_names[p.is_array].insert(p.name);

    // Import scopes.
    for(const auto& a: rootscope.Scopes) // type => scalars,arrays
        for(unsigned is_array=0; is_array<2; ++is_array)
            for(const auto& t: a[is_array])
                if(t.second.is_shared || everything_is_shared)
                {
                    if(param_names[is_array].find(t.first) != param_names[is_array].end())
                    {
                        dprintf(dp_warnings, "Warning: In sub '%s', %s parameter '%s' shadows a shared variable by the same name\n",
                            name.c_str(),
                            is_array ? "array" : "scalar",
                            t.first.c_str());
                        continue;
                    }

                    if(is_array)
                    {
                        //fprintf(stderr, "Imports global array '%s'\n", t.first.c_str());
                        subscope.usage[t.first].is_global_array = true;
                    }
                    else
                    {
                        //fprintf(stderr, "Imports global scalar '%s'\n", t.first.c_str());
                        subscope.usage[t.first].is_global_scalar = true;
                    }
                }
    CurrentScope = &subscope;

    for(std::size_t a=0; a<subdata.params.size(); ++a)
    {
        const SubRoutine::ParamSpec& p = subdata.params[a];
        // name, type.. ignore passconvention
        ExplicitlyDefine_C(LexLocation(), p.name, p.type, false, false, false, int(a));
    }
}

void ExitSubFuncScope(const std::string& name, stmtp&& code)
{
    const std::string basename = UCase(StripTypeSuffix(name));
    Scope& rootscope = *GetRootScope();
    Scope& curscope  = *GetCurrentScope();
    assert(&curscope != &rootscope);
    assert(basename == CurrentSubBaseName);

    //dprintf(dp_rulecomego, "Exiting sub '%s' (%s)\n", name.c_str(), basename.c_str());

    assert(rootscope.subroutines.find(basename) != rootscope.subroutines.end());

    rootscope.subroutines[basename].code = std::move(code);

    CurrentScope = &rootscope;
    CurrentSubBaseName.clear();
}

void SubScopeImportSharedVariable(const std::string& name, TypeP type, bool is_array)
{
    assert(!CurrentSubBaseName.empty());

    bool found = false;
    bool implicitly_defined = false;

    // If type was not specified (they're not looking for a particularly typed variable),
    // determine the type from the name.
    if(!type) type = CreateTypeTemplate(name);

try_finding_again:
    for(const auto& a: GetRootScope()->Scopes) // type => scalars,arrays
    {
        auto i = a[is_array].find(name);
        if(i != a[is_array].end())
        {
            if(*type == *i->second.type)
            {
                if(is_array)
                {
                #ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Imports global array '%s'\n", name.c_str());
                #endif
                    CurrentScope->usage[name].is_global_array = true;
                }
                else
                {
                #ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Imports global scalar '%s'\n", name.c_str());
                #endif
                    CurrentScope->usage[name].is_global_scalar = true;
                }
                found = true;
            }
        }
    }
    // FIXME TODO
    if(!found)
    {
        if(!implicitly_defined)
        {
            dprintf(dp_warnings, "Warning: Could not find SHARED %s variable by name '%s', type '%s'... Defining!\n",
                is_array ? "array" : "scalar", name.c_str(),
                type->Dump().c_str());

            auto backup = CurrentScope;
            CurrentScope = GetRootScope();

            TypeP deftype ( new TypeInfo(*type) );
            if(is_array)
                ImplicitlyDefineArray(name,1,option_base,10, false,false, deftype);
            else
                ImplicitlyDefineScalar(name, false,false, deftype);

            CurrentScope = backup;

            implicitly_defined = true;
            goto try_finding_again;
        }

        dprintf(dp_errors, "Error: Could not find SHARED %s variable by name '%s', type '%s'!\n",
                is_array ? "array" : "scalar", name.c_str(), type->Dump().c_str());
    }
}

const SubRoutine* FindSubFunc(const std::string& name)
{
    const std::string basename = UCase(StripTypeSuffix(name));

    Scope& rootscope = *GetRootScope();
    if(rootscope.usage[basename].is_subroutine)
        return &rootscope.subroutines[basename];

    return nullptr;
}

const std::string GetCurrentSubBaseName()
{
    return CurrentSubBaseName;
}

void AnalyzeSubParamUsage()
{
    Scope* const old_scope = CurrentScope;

    struct ParamReport
    {
        std::set<std::pair<const SubRoutine*, std::size_t>> PassedAsParamTo;
        bool AssignedTo;
    };
    std::unordered_map<const SubRoutine*, std::vector<ParamReport>> ParamsKnowledge;

    for(const auto& s: RootScope->subroutines)
    {
        std::vector<ParamReport>& report = ParamsKnowledge[&s.second];
        report.resize( s.second.params.size() );

        std::function<void(const class stmt*)> AnalyzeCode;
        std::function<void(const class expr*,tristate,const std::string&,unsigned)> AnalyzeExpr;

        AnalyzeCode = [&](const class stmt* stmt)
        {
            if(!stmt) return;
            stmt->ForEachSubStmt(AnalyzeCode);
            stmt->ForEachSubExpr(AnalyzeExpr);
        };
        AnalyzeExpr = [&](const class expr* expr, tristate write_target, const std::string& name, unsigned callee_param)
        {
            if(!expr) return;

            const SubRoutine* callee_func = name.empty() ? nullptr : FindSubFunc(name);

            if(!write_target.is_false())
            {
                /* Maybe or definitely a write target */
                if(dynamic_cast<const expr_returnvalueholder*> (&*expr)) return;
                if(const expr_var* var = dynamic_cast<const expr_var*> (&*expr))
                {
                    const std::string& name = var->getname();

                    VarContext ctx = FindVariableFromScope(name, var->isarray());

                    //dprintf(dp_param_tracking, "- Writes into variable '%s'%s\n", name.c_str(), write_target.is_maybe() ? " (maybe)" : "");
                    if(!ctx)
                    {
                        dprintf(dp_param_tracking, "- '%s' was not found in scope!\n", name.c_str());
                    }
                    else
                    {
                        if(ctx.info->param_index >= 0)
                        {
                            unsigned a = ctx.info->param_index;

                            if(write_target.is_maybe())
                            {
                                //dprintf(dp_param_tracking, "- Param %u passed to %p[%u]\n", a,
                                //    &*callee_func, callee_param);
                                report[a].PassedAsParamTo.insert( {callee_func,callee_param} );
                            }
                            else
                            {
                                //dprintf(dp_param_tracking, "- Param %u written to\n", a);
                                report[a].AssignedTo = true;
                            }
                        }
                    }
                }
            }
            expr->ForEachSubExpr(AnalyzeExpr);
        };

        CurrentScope = s.second.sub_scope;
        AnalyzeCode(s.second.code.get());
    }

    /* Optimisticly assume parameters are not modified */
    for(auto& s: RootScope->subroutines)
        for(auto& p: s.second.params)
            p.was_modified = false;

    for(;;)
    {
        bool changes = false;
        for(auto& s: RootScope->subroutines)
        {
            SubRoutine& sub = s.second;
            const std::string& subname = s.first;
            std::vector<ParamReport>& report = ParamsKnowledge[&sub];

            //dprintf(dp_param_tracking, "Checking sub %s\n", subname.c_str());
            for(std::size_t a=0; a<report.size(); ++a)
            {
                if(sub.params[a].was_modified) continue;

                auto& param_report = report[a];

                // Was the parameter definitely modified?
                if(param_report.AssignedTo)
                {
                    // Yes. Mark it as such.
                    dprintf(dp_param_tracking, "Subroutine %s: Parameter %s was assigned to; will not automatically convert to BYVAL\n", subname.c_str(), sub.params[a].name.c_str());
                    sub.params[a].was_modified = true;
                    changes = true;
                }
                else
                    for(const auto& p: param_report.PassedAsParamTo)
                    {
                        const SubRoutine* callee_func    = p.first;
                        const std::size_t callee_paramno = p.second;

                        if(callee_func->params[callee_paramno].was_modified)
                        {
                            dprintf(dp_param_tracking, "Subroutine %s: Pamameter %s was modified; will not automatically convert to BYVAL\n", subname.c_str(), sub.params[a].name.c_str());

                            sub.params[a].was_modified = true;
                            changes = true;
                        }
                    }
            }
        }
        if(!changes) break;
    }
    CurrentScope = old_scope;
}

void SetCurrentScope(Scope* scope)
{
    CurrentScope = scope;
}
