#include <ostream>
#include <sstream>
#include <unordered_set>

#include "assert.h"

#include "compile.hh"
#include "symtables.hh"
#include "binding.hh"
#include "tristate"
#include "dprintf.hh"

#include "util.hh"

#define CPP_SUBFUNC_NAME(x) (std::string(x)+"_f")
#define CPP_MEMBER_NAME(x) (std::string(x))
#define CPP_UTYPE_TYPENAME(x)  (std::string(x)+"_ut")
#define CPP_UTYPE_NAME(x)  (std::string(x)+"_u")
#define CPP_INT_NAME(x)    (std::string(x)+"_i")
#define CPP_SINGLE_NAME(x) (std::string(x)+"_z")
#define CPP_DOUBLE_NAME(x) (std::string(x)+"_d")
#define CPP_LONG_NAME(x)   (std::string(x)+"_l")
#define CPP_STRING_NAME(x) (std::string(x)+"_s")
#define CPP_ARRAY_NAME(x)  (std::string(x)+"_a")

#define CPP_BYTE_NAME(x)   (std::string(x)+"_y")
#define CPP_UBYTE_NAME(x)  (std::string(x)+"_Y")
#define CPP_UINT_NAME(x)   (std::string(x)+"_I")
#define CPP_ULONG_NAME(x)  (std::string(x)+"_U")
#define CPP_BIT_NAME(x)   (std::string(x)+"_b")

tristate DoResumeHandling = maybe;
tristate DoERLhandling    = true;
tristate DoEventChecks    = maybe;
bool CheckStaticSubscripts = false;
bool MakeLineLines = true;

std::vector<std::string> global_data;

static std::string FunctionReturnVarName;
static std::string Current_EXITDO_label;
static std::string Current_EXITFOR_label;
static unsigned EXITlabelCounter=0, GOSUBlabelCounter=0;
static unsigned EventLabelCounter=0, IFlabelCounter=0;
static unsigned ForStepTempCounter=0, ForEndTempCounter=0, SelectTempCounter=0;

static std::vector<std::string> ScopeTemporaries;

static std::vector<std::string> SystemCode;
static std::unordered_set<std::string> SystemCodeSet;

#include "libdata-common.inc"

static void AddSystemCode(const std::string& s)
{
    if(SystemCodeSet.insert(s).second)
        SystemCode.push_back(s);
}
static void AddSystemCode(const std::initializer_list<std::string>& codes)
{
    for(const auto& s: codes)
        AddSystemCode(s);
}

static const std::pair<std::string, std::string>
    GetTypeBaseNames(const TypeInfo& type, const std::string& name)
{
    switch(type.Type)
    {
        case TypeType::Integer: return std::make_pair("BAS_INT", CPP_INT_NAME(name));
        case TypeType::Single:  return std::make_pair("BAS_SGN", CPP_SINGLE_NAME(name));
        case TypeType::Double:  return std::make_pair("BAS_DBL", CPP_DOUBLE_NAME(name));
        case TypeType::Long:    return std::make_pair("BAS_LNG", CPP_LONG_NAME(name));
        case TypeType::String:
        {
            if(type.fixlength)
            {
                // FIXME: Unsupported for now
                return std::make_pair("std::string", CPP_STRING_NAME(name));
            }
            else
            {
                return std::make_pair("std::string", CPP_STRING_NAME(name));
            }
        }
        case TypeType::UserType:
            return std::make_pair(CPP_UTYPE_TYPENAME(type.utypename), CPP_UTYPE_NAME(name));

        case TypeType::Byte:            return std::make_pair("BAS_BYTE",  CPP_BYTE_NAME(name));
        case TypeType::UnsignedByte:    return std::make_pair("BAS_UBYTE", CPP_UBYTE_NAME(name));
        case TypeType::UnsignedInteger: return std::make_pair("BAS_UINT",  CPP_UINT_NAME(name));
        case TypeType::UnsignedLong:    return std::make_pair("BAS_ULONG", CPP_ULONG_NAME(name));

        case TypeType::Bit:
        {
            if(type.fixlength == 1) return std::make_pair("bool", CPP_BIT_NAME(name));
            if(type.fixlength <= 8) return std::make_pair("B_BIT<BAS_BYTE," + std::to_string(type.fixlength)+">", CPP_BIT_NAME(name));
            if(type.fixlength <=16) return std::make_pair("B_BIT<BAS_INT," + std::to_string(type.fixlength)+">", CPP_BIT_NAME(name));
            if(type.fixlength <=32) return std::make_pair("B_BIT<BAS_LNG," + std::to_string(type.fixlength)+">", CPP_BIT_NAME(name));
            return GetTypeBaseNames(*TypeLng, name);
        }
        case TypeType::UnsignedBit:
        {
            if(type.fixlength == 1) return std::make_pair("bool", CPP_BIT_NAME(name));
            if(type.fixlength <= 8) return std::make_pair("B_BIT<BAS_UBYTE," + std::to_string(type.fixlength)+">", CPP_BIT_NAME(name));
            if(type.fixlength <=16) return std::make_pair("B_BIT<BAS_UINT," + std::to_string(type.fixlength)+">", CPP_BIT_NAME(name));
            if(type.fixlength <=32) return std::make_pair("B_BIT<BAS_ULONG," + std::to_string(type.fixlength)+">", CPP_BIT_NAME(name));
            return GetTypeBaseNames(*TypeUlong, name);
        }
    }
    /* not reached, but keeps compiler happy */
    assert(!"not_reached");
    return std::make_pair(CPP_UTYPE_TYPENAME(type.utypename), CPP_UTYPE_NAME(name));
}

enum StrCompileDefs
{
    StrCompileMultipurpose,
    StrCompilePrinting,
    StrCompileConstOnly
};
static const std::string CompileStringConst
   (const std::string&str,
    StrCompileDefs defs = StrCompileMultipurpose,
    bool CareAboutNuls = true)
{
    char esc = '"';
    if(defs == StrCompilePrinting && str.size() == 1)
    {
        esc = '\'';
    }
    std::string result;
    result += esc;
    bool contains_nul_char = false;
    unsigned size_minus = 0;
    for(std::size_t a=0; a<str.size(); ++a)
    {
        unsigned char c = str[a];
        if((result.size()-size_minus) >= 77 && esc == '"')
        {
            result += "\"\n\"";
            size_minus = result.size()-1;
        }
        if(c=='\n') { result += "\\n"; continue; }
        if(c=='\r') { result += "\\r"; continue; }
        if(c=='\t') { result += "\\t"; continue; }
        if(c=='\b') { result += "\\b"; continue; }
        if(c=='\a') { result += "\\a"; continue; }
        if(c==esc || c=='\\') { result += '\\'; result += str[a]; continue; }
        if(c < 32 || c >= 127)
        {
            if(c == 0) contains_nul_char = true;
            char Buf[64];
            int width = 1;
            if(a+1 < str.size() && str[a+1]>='0' && str[a+1]<='7') width=3;
            sprintf(Buf, "\\%0*o", width, c);
            result += Buf;
            continue;
        }
        result += str[a];
    }
    result += esc;
    if(esc == '"' && defs != StrCompileConstOnly && contains_nul_char && CareAboutNuls)
    {
        return "std::string(" + result + ", " + std::to_string(str.size()) + ")";
    }
    return result;
}

const std::string LexLocation::make_cpp_line() const
{
    if(!MakeLineLines) return "";
    return "#line " + std::to_string(line) + " " + CompileStringConst(filename, StrCompileConstOnly) + "\n";
}


enum VarDeclType { Init_Var, Param_Ref, Param_Copy, Param_Const, Type_Member, Clear_Var };

static std::string CompileVarDeclaration
    (std::ostream& out,
     const std::string& name,
     TypeP typep,
     VarDeclType decl_type = Init_Var
    )
{
    std::string resulting_name;

    if(!typep)
    {
        out << name;
        dprintf(dp_errors, "Error: Type of '%s' is unknown\n", name.c_str());
        return name;
    }
    out << std::flush;

    assert(typep);
    const TypeInfo& type = *typep;

    const std::pair<std::string, std::string> base = GetTypeBaseNames(type, name);

    switch(type.arraytype)
    {
        case TypeInfo::NotArray:
        {
            if(decl_type == Clear_Var)
            {
                out << (resulting_name=base.second);
            }
            else
            {
                bool const_ref = false;
                if(decl_type == Param_Const)
                    const_ref = !type.is_num();

                if(const_ref) out << "const ";
                out << base.first << ' ';
                if(decl_type == Param_Ref || const_ref) out << "&";
                out << (resulting_name=base.second);
            }

            if(decl_type == Init_Var)
            {
                out << " { }";
            }
            else if(decl_type == Clear_Var)
            {
                switch(type.Type)
                {
                    case TypeType::Integer:
                    case TypeType::Long:
                        out << " = 0";
                        break;
                    case TypeType::Single:
                    case TypeType::Double:
                        out << " = 0.0";
                        break;
                    case TypeType::String:
                        if(decl_type == Clear_Var)
                            out << ".clear()";
                        break;
                    default:
                        out << " = " << base.first << "()";
                }
            }
            break;
        }
        case TypeInfo::StaticArray:
        {
            if(decl_type == Clear_Var)
            {
                // FIXME: Lexlocation
                expr_var *t = new expr_var(LexLocation(), name + TypeIntoTypeSuffix(typep->Type));
                t->set_type(typep);
                stmt_erase e{LexLocation(), exprp(t)};
                out << e.Compile();
            }
            else
            {
                out << base.first << ' ';
                if(decl_type == Param_Copy)
                {
                    dprintf(dp_warnings, "Warning: BYVAL Array parameters not supported\n");
                }
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                for(std::size_t a=0; a<type.dimensions.size(); ++a)
                    out << (!a ? "[" : "][") << type.dimensions[a].second;
                out << "]";

                if(decl_type == Init_Var)
                    out << " = { }";
            }
            break;
        }
        case TypeInfo::DynamicArray:
        {
            AddSystemCode(DYNARRAY_Common);

            if(decl_type == Clear_Var)
            {
                // Could also use stmt_erase here, but this is simpler.
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                out << ".Erase()";
            }
            else
            {
                if(decl_type == Param_Const) out << "const ";
                out << "BAS_DYNARRAY<" << base.first << ", " << type.dimensions.size() << "> ";
                if(decl_type == Param_Ref
                || decl_type == Param_Const) out << '&';
                out << (resulting_name=CPP_ARRAY_NAME(base.second));

                if(decl_type == Init_Var)
                {
                    out << "(";
                    for(std::size_t a=0; a<type.dimensions.size(); ++a)
                    {
                        if(a) out << ", ";
                        out
                            << type.dimensions[a].first
                            << ','
                            << type.dimensions[a].second;
                    }
                    out << ')';
                }
            }
            break;
        }
        case TypeInfo::ParamArray:
        {
            AddSystemCode(DYNARRAY_Common);
            AddSystemCode(MAGICARRAY_Common);

            if(decl_type == Param_Const) out << "const ";
            out << "BAS_MAGICARRAY<" << base.first << ">";
            if(decl_type == Param_Ref
            || decl_type == Param_Const) out << "&& ";
            out << (resulting_name=CPP_ARRAY_NAME(base.second));
            break;
        }
    }

    return resulting_name;
}

static void CompileTypeDefs(std::ostream& out, const Scope& scope)
{
    for(const auto& t: scope.types)
    {
        UtypeP p = t.second;
        assert(p);

        out << "struct " << CPP_UTYPE_TYPENAME(t.first) << " {\n";

        for(const auto& member: *p)
        {
            out << "  ";
            CompileVarDeclaration(out, CPP_MEMBER_NAME(member.first), member.second, Type_Member);
            out << ";\n";
        }

        out << "};\n";
    }
}

class ErrorHandlerCode
{
    std::string repeat_label, ignore_label, event_label;
    LexLocation loc;
public:
    std::string GetBegin(const LexLocation& context, bool enabled)
    {
        loc = context;

        if(enabled && DoResumeHandling)
        {
            repeat_label = "erp" + std::to_string(EventLabelCounter);
            ignore_label = "eri" + std::to_string(EventLabelCounter);
        }

        if(enabled && DoEventChecks)
            event_label  = "eve" + std::to_string(EventLabelCounter);

        if(!repeat_label.empty() || !event_label.empty())
            ++EventLabelCounter;

        std::string result;

        if(!repeat_label.empty())
            result += "B_ERR_STMT_BEGIN(" + repeat_label + "); ";

        return result;
    }
    std::string GetEnd()
    {
        std::string result;

        if(!event_label.empty())
        {
            result += " B_CHECK_EVENTS(" + event_label + ");";
        }

        if(!repeat_label.empty())
        {
            if(DoERLhandling)
                result += " B_ERR_STMT_END(" + repeat_label + "," + ignore_label + ","
                       + std::to_string(loc.line) + ","
                       + std::to_string(loc.col)  + ","
                       + CompileStringConst(loc.filename, StrCompileConstOnly)
                       + ");";
            else
                result += " B_ERR_STMT_END_BRIEF(" + repeat_label + "," + ignore_label + ");";
        }
        return result;
    }
};

static void ProcessScopeVars(
    std::ostream& out,
    const ScopeVarListType& vars,
    tristate shared_only,
    bool UtypesOnly,
    VarDeclType decl_type
#ifdef BINDING_DEBUG
   ,unsigned typeindex,unsigned is_array
#endif
    )
{
    for(const auto& v: vars)
    {
        const VarContextDef& ctx = v.second;
#ifdef BINDING_DEBUG
        dprintf(dp_declare, "Scope(type=%u,is_array=%u, decltype=%d,utypes=%d) var: %s (%s), static=%d,shared=%d,param=%d\n",
            typeindex,is_array,int(decl_type),UtypesOnly, v.first.c_str(), ctx.type->Dump().c_str(),
            ctx.is_static,ctx.is_shared,ctx.param_index);
#endif

        if(!shared_only.is_maybe() && ctx.is_shared != shared_only) continue;
        if(ctx.param_index >= 0) continue;
        if(UtypesOnly && ctx.type->Type != TypeType::UserType) continue;

        out << ctx.defined_at.make_cpp_line();

        if(ctx.is_static) out << "static ";

        CompileVarDeclaration(out, v.first, ctx.type, decl_type);
        out << ";\n";
    }
}

static void ProcessScopeVars(std::ostream& out, const Scope& scope, tristate shared_only, VarDeclType decl_type)
{
    /* When emitting Utypes, ignore non-Utypes (explicitly typed variables) */
    for(unsigned a=0; a<NumTypes; ++a)
        for(unsigned is_array=0; is_array<2; ++is_array)
            ProcessScopeVars(
                out,
                scope.Scopes[a][is_array], shared_only,
                (TypeType)a == TypeType::UserType,
                decl_type
#ifdef BINDING_DEBUG
                ,a,is_array
#endif
              );
}

static void CompileVars(std::ostream& out, const Scope& scope, tristate shared_only)
{
    ProcessScopeVars(out, scope, shared_only, Init_Var);
    if(shared_only.is_false() && !ScopeTemporaries.empty())
    {
        out << "// Automatically added temporaries:\n";
        for(const auto& s: ScopeTemporaries) out << s << "\n";
        ScopeTemporaries.clear();
    }
}

static void EmitClearVars(std::ostream& out, const Scope& scope)
{
    ProcessScopeVars(out, scope, maybe, Clear_Var);
}

static void CompileSubFunc(std::ostream& out, const std::string& name, const SubRoutine& sub, bool decl_only)
{
    LexLocation loc = decl_only ? sub.declared_at : sub.defined_at;
    out << loc.make_cpp_line();

    out << "static ";
    if(sub.returntype)
    {
        const std::pair<std::string, std::string> base = GetTypeBaseNames(*sub.returntype, name);
        /* no "const" here, to not hinder passing the result as reference */
        out << base.first;
    }
    else
        out << "void";

    out << " " << CPP_SUBFUNC_NAME(name) << "(";

    for(std::size_t a=0; a<sub.params.size(); ++a)
    {
        const SubRoutine::ParamSpec& param = sub.params[a];
        if(a) out << ", ";
        CompileVarDeclaration(out,
                              param.name,
                              param.type,
                              param.was_modified
                               ? param.passconvention ==
                                   SubRoutine::ParamSpec::ByVal
                                       ? Param_Copy
                                       : Param_Ref
                               : Param_Const);
    }

    out << ')';

    if(decl_only)
        out << ";\n";
    else
    {
        out << "\n";
    }
}

static void CompileSubDeclarations(std::ostream& out, const Scope& scope)
{
    for(const auto& s: scope.subroutines)
        CompileSubFunc(out, s.first, s.second, true);
}

static std::string CompileCode(const stmt* codep)
{
    /* The statements form a reverse single-linked list. Reverse that list first,
     * so we can compile them in real order.
     */
    std::list<const stmt*> stack;
    while(codep) { stack.push_front(codep); codep = &*codep->get_prev(); }

    std::string result;
    for(std::list<const stmt*>::iterator j, i = stack.begin(); i != stack.end(); i=j)
    {
        j=i; ++j;
        codep = *i;

        result += codep->GetLoc().make_cpp_line();

        ErrorHandlerCode errorframe;
        result += errorframe.GetBegin(codep->GetLoc(), codep->MayErr());
        result += codep->Compile();
        result += errorframe.GetEnd();
        result += "\n";

        stack.erase(i);
    }
    return result;
}

static const std::string GenerateLabelName(const std::string& lab)
{
    const Scope& scope = *GetCurrentScope();
    if(scope.labels.find(lab) == scope.labels.end())
    {
        dprintf(dp_errors, "Error: Label %s not defined\n", lab.c_str());
        return "L??";
    }

    if(lab[0] >= '0' && lab[0] <= '9') return "L" + lab;

    static std::map<std::pair<const Scope*, std::string>, unsigned> ldata;
    std::pair<const Scope*, std::string> seed(&scope, lab);

    unsigned lnum = 0;
    auto l = ldata.find(seed);
    if(l != ldata.end())
        lnum = l->second;
    else
        lnum = ldata[seed] = ldata.size() + 1;

    return "L" + std::to_string(lnum) + "_" + lab;
}

static std::string CompileExpr(const expr* e, ExprCompileDefs defs=exprdef_nothingparticular)
{
    if(!e) return "";

    if(defs == exprdef_cast_to_int)
    {
        TypeP type = e->GetType();
        if(type && !type->is_integer())
        {
            if(const expr_number* num = dynamic_cast<const expr_number*>(&*e))
            {
                if(num->is_int()) goto NormalExprCompile;
            }
            //std::cout << "/* operand type is " << type->Dump() << " */";

            defs = exprdef_nothingparticular;

            return "std::llround(" + e->Compile(defs) + ")";
        }
    }
NormalExprCompile:
    return e->Compile(defs);
}

std::string stmt_if::Compile() const
{
    /* We always have a condition, maybe a thenp and maybe an elsep.
     * We never have an elsep without thenp, but we don't necessarily have a thenp.
     */
    if(!thenp)
    {
        /* If we don't have a thenp, we just evaluate a condition. */
        assert(!elsep);

        ErrorHandlerCode error_handler;
        std::string result = error_handler.GetBegin(GetLoc(), condition->MayErr());
        result += CompileExpr(&*condition, exprdef_pure_condition);
        result += ';';
        result += error_handler.GetEnd();
        return result;
    }

    /* So, we have a thenp. And maybe an elsep. */

    /* If our condition does not generate errors, things are simple.
     *
     *    if(condition) { thenp; } else { elsep; }
     *
     */

    bool condition_may_err = condition->MayErr() && (DoResumeHandling || DoEventChecks);

    if(!condition_may_err)
    {
        std::string result = "if(" + CompileExpr(&*condition, exprdef_pure_condition) + ") {\n";
        if(thenp) result += CompileCode(&*thenp);
        result += "}";
        if(elsep) result += " else {\n" + CompileCode(&*elsep) + "}";
        return result;
    }

    /* Otherwise we need special code. With both thenp and elsep:
     *
     *    EB if(!(condition)) goto nobranch; EE
     *    yesbranch:   thenp; goto skipbranch;
     *    nobranch:    elsep;
     *    skipbranch:;
     *
     * With just thenp:
     *
     *    EB if(!(condition)) goto nobranch; EE
     *    yesbranch:   thenp;
     *    nobranch:;
     *
     * Why this complication? Because we cannot nest error-handling blocks.
     * We must ensure that the code that is between EB and EE will not involve
     * another pair of EB and EE. This is accomplished by using "goto nobranch;"
     * instead of "thenp".
     **/
    //std::string yeslabel = "ify" + std::to_string(IFlabelCounter);
    std::string  nolabel = "ifn" + std::to_string(IFlabelCounter);
    std::string endlabel = "ife" + std::to_string(IFlabelCounter);
    ++IFlabelCounter;

    ErrorHandlerCode error_handler;
    std::string result = error_handler.GetBegin(GetLoc(), condition->MayErr());
    result += "if(!(" + CompileExpr(&*condition, exprdef_pure_condition) + ")) goto " + nolabel + ";";
    result += error_handler.GetEnd();
    //result += " " + yeslabel + ":\n";
    if(thenp) result += "\n" + CompileCode(&*thenp);
    if(elsep)
    {
        result += "goto " + endlabel + "; " + nolabel + ":\n";
        result += CompileCode(&*elsep);
        result += endlabel + ":;";
    }
    else
    {
        result += nolabel + ":;";
    }
    return result;
}
std::string stmt_doloop::Compile() const
{
    std::string previous_EXITDO_label = Current_EXITDO_label;

    bool needs_do_exit_label = false;
    std::function<void(const stmt*)> scan_stmt = [&](const stmt* s) {
        if(!s) return;
        if(auto l = dynamic_cast<const stmt_labelref*>(s))
            if(l->token == EXITDO)
                { needs_do_exit_label = true; }
        if(dynamic_cast<const stmt_doloop*>(s)) return; // stop looking inside nested DOs
        s->ForEachSubStmt(scan_stmt);
    };
    scan_stmt(&*bodyp);

    std::ostringstream out;
    if(!is_while && needs_do_exit_label)
        Current_EXITDO_label = "do_exit" + std::to_string(EXITlabelCounter++);

    /* Different types of loops:
     *
     * Errorhandler, if used, is supposed to trap errors
     * happen in the evaluation of the loop condition.
     * The statements inside the loop have their own error traps;
     * we mustn't generate one to encompass them all, or we will get
     * compilation errors if there is a goto from outside to the inside.
     *
     *    precondition postcondition
     *    no           no                 for(;;) { stmt; }
     *    no           mayerr             for(;;) { stmt; EB if(!(condition2)) break; EE; }
     *    no           yes                do { stmt; } while(condition2);
     *
     *    mayerr       no                 for(;;) { EB if(!(condition)) break; EE; stmt; }
     *    mayerr       mayerr             for(;;) { EB if(!(condition)) break; EE; stmt; EB if(!condition2)) break; EE }
     *    mayerr       yes                do  { EB if(!(condition)) break; EE; stmt; } while(condition2);
     *
     *    yes          no                 while(condition) { stmt; }
     *    yes          mayerr             while(condition) { stmt; EB if(condition2) break; EE }
     *    yes          yes                while(condition) { stmt; if(condition2) break; }
     *
     */

    bool             precondition_may_error = precondition && precondition->MayErr() && (DoResumeHandling || DoEventChecks);
    ErrorHandlerCode precondition_error_handler;
    bool             postcondition_may_error = postcondition && postcondition->MayErr() && (DoResumeHandling || DoEventChecks);
    ErrorHandlerCode postcondition_error_handler;

    bool whileloop = precondition && !precondition_may_error;
    bool do_while  = postcondition && !postcondition_may_error && !whileloop;

    if(do_while)
        out << "do {\n";
    else if(whileloop)
        out << "while(" << CompileExpr(&*precondition, exprdef_pure_condition) << ") {\n";
    else
    {
        out << "for(;;) {";
        if(precondition)
        {
            out << precondition_error_handler.GetBegin(GetLoc(), precondition->MayErr());
            out << "if(!(" << CompileExpr(&*precondition, exprdef_pure_condition) << ")) break; ";
            out << precondition_error_handler.GetEnd();
        }
        out << "\n";
    }

    out << CompileCode(&*bodyp);

    if(do_while)
    {
        // TODO: Get a #line statement for the post-condition line separately
        out << GetLoc().make_cpp_line();

        out << "} while(" << CompileExpr(&*postcondition, exprdef_pure_condition) << ");";
    }
    else
    {
        if(postcondition)
        {
            // TODO: Get a #line statement for the post-condition line separately
            out << GetLoc().make_cpp_line();

            out << postcondition_error_handler.GetBegin(GetLoc(), postcondition->MayErr());
            out << "if(!(" << CompileExpr(&*postcondition, exprdef_pure_condition) << ")) break; ";
            out << postcondition_error_handler.GetEnd();
            out << "\n";
        }
        out << "}";
    }

    if(!is_while && needs_do_exit_label)
        out << ' ' << Current_EXITDO_label << ":;";

    Current_EXITDO_label = previous_EXITDO_label;
    return out.str();
}

std::string stmt_for::Compile() const
{
    std::string previous_EXITFOR_label = Current_EXITFOR_label;
    bool needs_for_exit_label = false;
    std::function<void(const stmt*)> scan_stmt = [&](const stmt* s) {
        if(!s) return;
        if(auto l = dynamic_cast<const stmt_labelref*>(s))
            if(l->token == EXITFOR)
                { needs_for_exit_label = true; }
        if(dynamic_cast<const stmt_for*>(s)) return; // stop looking inside nested FORs
        s->ForEachSubStmt(scan_stmt);
    };
    scan_stmt(&*bodyp);

    /* FOR loop has BEGIN, END and STEP.
     *
     * All of these are evaluated exactly ONCE during the loop.
     * In other words, the FOR-loop precalculates the number of iterations needed.
     *
     * However, it still uses the iteration variable and step, rather than
     * iterating an integer and assigning the iteration variable.
     * This means that e.g. FOR a! = 1 TO 100 STEP 0.2: NEXT accumulates floating point
     * inaccuracies, which wouldn't happen if it transformed the loop into:
     *    FOR internal% = 0 TO 999: a! = 1 + internal% * 0.2: NEXT
     * In addition, after the loop, the variable will contain a value of one step added.
     */
    bool end_is_constant   = end->is_const();
    bool step_is_constant  = step->is_const();
    bool begin_is_constant = begin->is_const();

    if(begin_is_constant && end_is_constant && step_is_constant)
    {
        // We know the number of iterations at compile time.
    }

    std::ostringstream out;
    if(!bodyp)
    {
        // If the loop does not contain a body, just create a delay call.
        ErrorHandlerCode error_handler;
        out << error_handler.GetBegin(GetLoc(), var->MayErr() || begin->MayErr() || end->MayErr() || step->MayErr());
        out << "{ auto dl_end = " << CompileExpr(&*end) << "; auto dl_step = " << CompileExpr(&*step) << "; ";
        out << "SDL_Delay((dl_end-" << CompileExpr(&*begin) << ") / dl_step); ";
        // Still do an assignment to the iteration variable though.
        out << CompileExpr(&*var) << " = dl_end + dl_step; }";
        out << error_handler.GetEnd();
        return out.str();
    }

    if(needs_for_exit_label)
        Current_EXITFOR_label = "for_exit" + std::to_string(EXITlabelCounter++);

    /* Note: In QuickBasic, the FOR-variable cannot be an array reference, and thus can never generate errors.
     *       In our compiler, it can however be an array reference.
     *
     *     EB                            -- If VAR, BEGIN, STEP or END can generate errors.
     *        var = begin;               -- If VAR or BEGIN can generate errors.
     *        GLOBAL_STEP = step;        -- if STEP is non-constant.
     *        GLOBAL_END = end;          -- if END is non-constant.
     *     EE                            -- see above.
     *     for(
     *          var = begin;             -- If neither VAR nor BEGIN can generate errors.
     *         ;
     *          var op end;              -- If VAR cannot generate errors, and END is constant.
     *          var op GLOBAL_END;       -- If VAR cannot generate errors, and END is non-constant.
     *         ;
     *          var = var + step;        -- if VAR cannot generate errors, and STEP is constant.
     *          var = var + GLOBAL_STEP  -- if VAR cannot generate errors, and STEP is non-constant.
     *        )
     *     {
     *          stmt;
     *
     *          EB                            -- If VAR can generate errors.
     *          var = var + step;             -- if VAR can generate errors, and STEP is constant.
     *          var = var + GLOBAL_STEP;      -- if VAR can generate errors, and STEP is non-constant.
     *          if(var op end) break;         -- if VAR can generate errors, and END is constant.
     *          if(var op GLOBAL_END) break;  -- if VAR can generate errors, and END is non-constant.
     *          EE                            -- see above.
     *     }
     *
     *     op is generated like:
     *         (step > 0) ? (var <= end) : (var >= end)
     */

    bool var_may_err   = var->MayErr()   && (DoResumeHandling || DoEventChecks);
    bool end_may_err   = end->MayErr()   && (DoResumeHandling || DoEventChecks);
    bool step_may_err  = step->MayErr()  && (DoResumeHandling || DoEventChecks);
    bool begin_may_err = begin->MayErr() && (DoResumeHandling || DoEventChecks);

    std::string stepvar_name;
    std::string endvar_name;

    if(!step_is_constant)
    {
        auto s = GetTypeBaseNames(*step->GetType(), "fstp" + std::to_string(ForStepTempCounter++));
        ScopeTemporaries.push_back(s.first + " " + s.second + " { };");
        stepvar_name = s.second;
    }
    if(!end_is_constant)
    {
        auto s = GetTypeBaseNames(*end->GetType(), "fend" + std::to_string(ForEndTempCounter++));
        ScopeTemporaries.push_back(s.first + " " + s.second + " { };");
        endvar_name = s.second;
    }

    ErrorHandlerCode loop_begin_errors;

    out << loop_begin_errors.GetBegin(GetLoc(), var_may_err || begin_may_err || end_may_err || step_may_err);
    if(var_may_err || begin_may_err)
        out << CompileExpr(&*var) << " = " << CompileExpr(&*begin) << "; ";
    if(!step_is_constant)
        out << stepvar_name << " = " << CompileExpr(&*step) << "; ";
    if(!end_is_constant)
        out << endvar_name << " = " << CompileExpr(&*end) << "; ";
    out << loop_begin_errors.GetEnd();

    out << "for(";
    if(!var_may_err && !begin_may_err)
        out << CompileExpr(&*var) << " = " << CompileExpr(&*begin);
    out << ';';

    bool step_maybe_negative = !step->GetType()->is_unsigned();
    bool step_maybe_positive = true;
    if(auto n = dynamic_cast<const expr_number*>(&*step))
    {
        if(n->get_f() < 0) step_maybe_positive = false;
        else               step_maybe_negative = false;
    }
    if(!var_may_err)
    {
        const char *incr = " <= ", *decr = " >= ";
        if(begin0step1loop) { incr = " < "; decr = nullptr; }
        if(step_maybe_negative && step_maybe_positive)
            out << "(" << (step_is_constant ? CompileExpr(&*step) : stepvar_name) << " > 0) ? (";
        if(step_maybe_positive)
            out << CompileExpr(&*var) << incr << (end_is_constant ? CompileExpr(&*end) : endvar_name);
        if(step_maybe_negative && step_maybe_positive)
            out << ") : (";
        if(step_maybe_negative)
            out << CompileExpr(&*var) << decr << (end_is_constant ? CompileExpr(&*end) : endvar_name);
        if(step_maybe_negative && step_maybe_positive)
            out << ")";
    }
    out << ';';
    if(!var_may_err) out << CompileExpr(&*var) << " += " << (step_is_constant ? CompileExpr(&*step) : stepvar_name);
    out << ") {\n";

    out << CompileCode(bodyp.get());

    if(var_may_err)
    {
        out << GetLoc().make_cpp_line();

        ErrorHandlerCode loop_end_errors;
        out << loop_end_errors.GetBegin(GetLoc(), var_may_err);

        const char *incr = " > ", *decr = " < ";
        if(begin0step1loop) { incr = " >= "; decr = nullptr; }

        out << CompileExpr(&*var) << " += " << (step_is_constant ? CompileExpr(&*step) : stepvar_name);
        out << "; if(";
        if(step_maybe_negative && step_maybe_positive)
            out << "(" << (step_is_constant ? CompileExpr(&*step) : stepvar_name) << " > 0) ? (";
        if(step_maybe_positive)
            out << CompileExpr(&*var) << incr << (end_is_constant ? CompileExpr(&*end) : endvar_name);
        if(step_maybe_negative && step_maybe_positive)
            out << ") : (";
        if(step_maybe_negative)
            out << CompileExpr(&*var) << decr << (end_is_constant ? CompileExpr(&*end) : endvar_name);
        if(step_maybe_negative && step_maybe_positive)
            out << ")";
        out << ") break;";

        out << loop_end_errors.GetEnd();
    }

    out << "}";
    if(needs_for_exit_label)
    {
        out << ' ' << Current_EXITFOR_label << ":;";
    }

    Current_EXITFOR_label = previous_EXITFOR_label;
    return out.str();
}
std::string stmt_let::Compile() const
{
    std::string result = CompileExpr(&*target);

    if(expr_binary* bin = dynamic_cast<expr_binary*> (&*value))
    {
        std::string op, cast1;
        ExprCompileDefs subdefs = exprdef_nothingparticular;
        switch(bin->token)
        {
            case XOR: op = "^="; break;
            case OR:  op = "|="; break;
            case AND: op = "&="; break;
            case (QBtoken)'+': op = "+="; break;
            case (QBtoken)'-': op = "-="; break;
            case (QBtoken)'*': op = "*="; break;
            case (QBtoken)'/': op = "/="; cast1 = "(BAS_DBL)"; break;
            default: break;
        }
        if(!op.empty())
        {
            /* Check if the first argument in this expr_binary
             * is the same as target, and if it is, act on it
             */
            if(expr_var* tvar = dynamic_cast<expr_var*>(&*target))
                if(expr_var* svar = dynamic_cast<expr_var*>(&*bin->p0))
                    if(tvar->IsSameAs(*svar))
                    {
                        result += ' ';
                        result += op;
                        result += ' ';
                        result += cast1;
                        result += CompileExpr(&*bin->p1, subdefs);
                        result += ";";
                        return result;
                    }
        }
    }

    return result + " = " + CompileExpr(&*value) + ";";
}

static std::string CompileCaseExpr(
    const expr* expr, bool IsSwitchCase,
    const std::unordered_map<std::string, unsigned>& string_keys)
{
    ExprCompileDefs defs = IsSwitchCase ? exprdef_cast_to_int : exprdef_nothingparticular;

    if(auto s = dynamic_cast<const expr_string*>(expr))
        if(!string_keys.empty())
            return std::to_string(string_keys.find(s->get())->second);
    return CompileExpr(expr, defs);
}

static void CompileCaseTest(
    std::ostream& out,
    const expr* expr, const std::string& varname,
    bool IsSwitchCase,
    const std::unordered_map<std::string, unsigned>& string_keys)
{
    if(const expr_caseor* e = dynamic_cast<const expr_caseor*>(&*expr))
    {
        CompileCaseTest(out, e->a.get(), varname, IsSwitchCase, string_keys);
        if(!IsSwitchCase) out << " || ";
        CompileCaseTest(out, e->b.get(), varname, IsSwitchCase, string_keys);
    }
    else if(const expr_caseis* e = dynamic_cast<const expr_caseis*>(&*expr))
    {
        std::string op = "==";
        switch(e->token)
        {
            case (QBtoken)'>':    op = ">"; break;
            case (QBtoken)'<':    op = "<"; break;
            case (QBtoken)'=':    op = "=="; break;
            case NOTEQUAL:        op = "!="; break;
            case LESSOREQUAL:     op = "<="; break;
            case GREATEROREQUAL:  op = ">="; break;
            default:
                dprintf(dp_errors, "Internal Error: Unknown expr_unary token: %s\n", GetTokenName(e->token).c_str());
                assert(!"unexpected caseis token");
        }
        if(IsSwitchCase && op == "==") op="";
        if(IsSwitchCase) out << "case " << op;
        else out << varname << ' ' << op << ' ';
        out << CompileCaseExpr(&*e->ref, IsSwitchCase, string_keys);
        if(IsSwitchCase) out << ":\n";
    }
    else if(const expr_caserange* e = dynamic_cast<const expr_caserange*>(&*expr))
    {
        if(IsSwitchCase)
        {
            out << "case " << CompileCaseExpr(&*e->min, IsSwitchCase, string_keys);
            out << " ... " << CompileCaseExpr(&*e->max, IsSwitchCase, string_keys);
            out << ":\n";
        }
        else
        {
            out << '('    << varname << " >= " << CompileCaseExpr(&*e->min, IsSwitchCase, string_keys);
            out << " && " << varname << " <= " << CompileCaseExpr(&*e->max, IsSwitchCase, string_keys);
            out << ')';
        }
    }
    else
    {
        assert(!"unexpected case type");
    }
}

std::string stmt_select::Compile() const
{
    std::ostringstream out;

    bool var_may_err = ref->MayErr() && (DoResumeHandling || DoEventChecks);
    std::string testvar_name;

    if(var_may_err || (!IsEntirelyIntegerCase() && !IsEntirelyStringCase()))
    {
        auto s = GetTypeBaseNames( IsEntirelyIntegerCase() && !ref->GetType()->is_integer() ? *TypeLng : *ref->GetType(),
                                   "selv" + std::to_string(SelectTempCounter++));
        ScopeTemporaries.push_back(s.first + " " + s.second + " { };");
        testvar_name = s.second;

        ErrorHandlerCode select_errors;
        out << select_errors.GetBegin(GetLoc(), var_may_err);
        if(IsEntirelyIntegerCase())
            out << testvar_name << " = " << CompileExpr(&*ref, exprdef_cast_to_int) << "; ";
        else
            out << testvar_name << " = " << CompileExpr(&*ref) << "; ";
        out << select_errors.GetEnd();
    }

    std::unordered_map<std::string, unsigned> string_keys;
    std::string hashkeytype;
    std::string hashmapname = "selm" + std::to_string(SelectTempCounter);
    //std::string hashitname  = "seli" + std::to_string(SelectTempCounter);
    unsigned num_string_keys = 0;

    if(IsEntirelyStringCase())
    {
        std::string hashmap_contents = "";
        std::function<void(const class expr*,tristate,const std::string&,unsigned)>
            AnalyzeExpr = [&](const class expr* expr, tristate, const std::string&, unsigned)
        {
            if(!expr) return;
            if(auto s = dynamic_cast<const expr_string*> (expr))
            {
                if(!hashmap_contents.empty()) hashmap_contents += ", ";
                hashmap_contents += "{ " + CompileStringConst(s->get()) + ", " + std::to_string(num_string_keys) + "}";
                string_keys.insert( { s->get(), num_string_keys } );
                ++num_string_keys;
            }
            expr->ForEachSubExpr(AnalyzeExpr);
        };
        for(const auto& c: cases) AnalyzeExpr(&*c.first,false,{},0);
        hashkeytype = GetTypeBaseNames(*expr_number((long)num_string_keys).GetType(), "").first;

        std::string hashmap = "static const std::unordered_map<std::string," + hashkeytype + "> " + hashmapname
                            + " { " + hashmap_contents + " };";
        ScopeTemporaries.push_back(hashmap);
        //ScopeTemporaries.push_back("std::unordered_map<std::string," + hashkeytype + ">::const_iterator " + hashitname + ";");
        ++SelectTempCounter;
        //
    }

    if(IsEntirelyIntegerCase() || IsEntirelyStringCase())
    {
        bool gosubcase = IsEntirelyGosubCase();
        bool had_caseelse = false;

        static unsigned gosubcase_counter = 0;
        std::string gosub_return_label = "on_gosub_return" + std::to_string(gosubcase_counter++);

        if(gosubcase)
        {
            std::string gosubcase_label = "gcl" + std::to_string(gosubcase_counter);

            out << "B_GOSUB(" << gosubcase_label << ", " << gosub_return_label
                << "); " << gosubcase_label << ": ";
            /* GCC already optimizes switch-case-gotos as jump tables.
             * GOSUB is not much different; we just need to push a return
             * address beforehand. We utilize GCC's existing ability, and
             * turn all GOSUBs into GOTOs.
             */
        }

        out << "switch(";
        if(IsEntirelyStringCase())
        {
            /* [&](){auto i = map.find(expr); return i == map.end() ? type(n) : i->second;}() */
            out << "[&]()->" << hashkeytype
                << "{auto i = " << hashmapname << ".find(";
            out << (!testvar_name.empty() ? testvar_name : CompileExpr(&*ref, exprdef_nothingparticular));
            out << "); "
                << " return i == " << hashmapname << ".end() ? " << num_string_keys
                << " : i->second;}()";
            //out << "likely(" << hashitname << " != " << hashmapname << ".end()) ? " << hashitname << "->second"
            //       " : " << hashkeytype << "(" << num_string_keys << ")";
        }
        else
            out << (!testvar_name.empty() ? testvar_name : CompileExpr(&*ref, exprdef_cast_to_int));
        out << ") {\n";

        for(const auto& c: cases)
        {
            if(gosubcase)
            {
                // FIXME: const-correctness broken here
                stmt_labelref* l = dynamic_cast<stmt_labelref*> (&*c.second);
                if(l && l->token == GOSUB) l->token = GOTO; // transform the GOSUBs to GOTOs
            }

            if(dynamic_cast<expr_caseelse*>(&*c.first))
            {
                out << "default:\n";
                out << CompileCode(c.second.get());
                had_caseelse = true;
                break;
            }
            CompileCaseTest(out, c.first.get(), "?", true, string_keys); // varname="?"
            out << CompileCode(c.second.get());
            out << "break;\n";
        }

        if(gosubcase && !had_caseelse)
        {
            out << "default: B_RETURN_TO(" << gosub_return_label << ");\n";
        }

        out << "}"; // end switch
        if(gosubcase)
        {
            out << " " << gosub_return_label << ":;";
        }
    }
    else
    {
        std::string endlabel;

        bool first = true;
        for(const auto& c: cases)
        {
            if(!first) out << "else ";
            if(dynamic_cast<expr_caseelse*>(&*c.first))
            {
                out << "{\n";
                out << CompileCode(c.second.get());
                out << "} ";
                break;
            }

            if(c.first->MayErr() && (DoResumeHandling || DoEventChecks))
            {
                if(endlabel.empty()) endlabel = "ife" + std::to_string(IFlabelCounter);
                std::string elselabel = "ifn" + std::to_string(IFlabelCounter++);

                ErrorHandlerCode select_errors;
                out << select_errors.GetBegin(GetLoc(), c.first->MayErr());
                out << "if(!(";
                CompileCaseTest(out, c.first.get(), testvar_name, false, string_keys);
                out << ")) goto " + elselabel << ";";
                out << select_errors.GetEnd();
                out << "\n" << CompileCode(c.second.get());
                out << " goto " << endlabel << "; " << elselabel << ": if(0) {} ";
            }
            else
            {
                out << "if(";
                CompileCaseTest(out, c.first.get(), testvar_name, false, string_keys);
                out << ") {\n";
                out << CompileCode(c.second.get());
                out << "} ";
            }
            first=false;
        }
        if(!endlabel.empty()) out << endlabel << ":;";
    }
    return out.str();
}

std::string stmt_print::Compile() const
{
    static unsigned UsingCounter=0;
    std::string OutTempName   = "O_" + std::to_string(UsingCounter);
    std::string UsingTempName = "U_" + std::to_string(UsingCounter++);

    std::ostringstream out;

    if(using_string)
    {
        AddSystemCode(Func_USING_Common);
        out << "UsingFormatter "/* << UsingTempName*/ << '(';
        out << CompileExpr(&*using_string);
        out << ").DoParam(";
    }

    AddSystemCode(Stream_Common);

    if(target)
    {
        out << "GET_STREAM_OUTPUT(" << CompileExpr(&*target) << ')';
    }
    else if(is_lprint)
    {
        out << "GET_PRINTER()";
    }
    else
    {
        AddSystemCode(Console_Common);
        out << "GET_STDOUT()";
    }

    enum { nothing, push, param } mode = using_string?param:push;

    for(const auto& p: params)
    {
        switch(p.first)
        {
            case SPC: // SPC function
            {
                if(mode == param) { out << ")"; mode = push; }
                if(mode == push)
                    { out << " << PrintSPCtag{" << CompileExpr(&*p.second) << "}"; }
                else
                    { out << OutTempName << ".DoSPC(" << CompileExpr(&*p.second) << ");\n"; }
                break;
            }
            case TAB: // TAB function
            {
                if(mode == param) { out << ")"; mode = push; }
                if(mode == push)
                    if(p.second)
                        { out << " << PrintTABtag{" << CompileExpr(&*p.second) << "}"; }
                    else
                        out << " << PrintNextTABtag{}";
                else
                    if(p.second)
                        { out << OutTempName << ".DoTAB(" << CompileExpr(&*p.second) << ");"; }
                    else
                        out << OutTempName << ".DoNextTAB();\n";
                break;
            }
            case CRLF: // terminator
            {
                if(mode == param) { out << ")"; mode = push; }
                if(mode == push)
                    out << " << PrintCRLFtag{}";
                else
                    out << OutTempName << ".DoCRLF(); ";
                break;
            }
            default:
                if(!p.first) // expression param
                {
                    if(using_string)
                    {
                        if(mode == push) { out << ";\n"; mode = nothing; }
                        if(mode != param) out << UsingTempName << ".DoParam(" << OutTempName;
                        out << ", ";
                        out << CompileExpr(&*p.second);
                        mode = param;
                    }
                    else
                    {
                        if(mode == nothing) { out << OutTempName; mode = push;}
                        out << " << ";
                        out << CompileExpr(&*p.second, exprdef_may_produce_char_from_string);
                    }
                    break;
                }
                assert(!"unexpected print operator");
        }
    }
    if(mode == param) out << "); ";
    if(mode == push) out << "; ";
    return out.str();
}

std::string stmt_input::Compile() const
{
    static unsigned UsingCounter=0;
    std::string InTempName   = "I_" + std::to_string(UsingCounter++);
    std::ostringstream out;

    AddSystemCode(Stream_Common);

    out << "{ InputStreamBase& " << InTempName << " = ";
    if(source)
    {
        out << "GET_STREAM_INPUT(" << CompileExpr(&*source);
        out << ')';
    }
    else
    {
        AddSystemCode(Console_Common);
        out << "GET_STDIN()";
    }
    out << ";\n";

    if(!source && !is_linput)  out << "for(;;) {";

    if(!source)
    {
        /* INPUT: Will display prompt string, and "? " is always appended.
         *        If no input string is provided, simply "? " is shown.
         * LINE INPUT: Will display prompt string verbatim if given.
         */

        expr* prompt_expr = nullptr;
        bool allocated = false;

        bool blank_prompt = !prompt.get()
        || ( dynamic_cast<const expr_string*>(prompt.get())
          && dynamic_cast<const expr_string*>(prompt.get())->get().empty());

        if(is_linput)
            { if(!blank_prompt) prompt_expr = prompt.get(); }
        else if(blank_prompt)
            { allocated=true; prompt_expr = new expr_string("? "); }
        else if(const expr_string* e = dynamic_cast<const expr_string*> (prompt.get()))
            { allocated=true; prompt_expr = new expr_string( e->get() + "? " ); }
        else
        {
            out << "GET_STDOUT() << ";
            out << CompileExpr(prompt.get(), exprdef_may_produce_char_from_string);
            out << " << \"? \";\n";
            blank_prompt = true;
        }

        if(!blank_prompt)
        {
            out << "GET_STDOUT() << ";
            out << CompileExpr(prompt_expr, exprdef_may_produce_char_from_string);
            out << ";\n";
        }
        if(allocated) delete prompt_expr;
    }

    if(!is_linput)
        out << InTempName << ".BeginInput();\n";

    for(const auto& p: params)
    {
        std::string method = is_linput ? "ReadLine" : "InputTo";
        out << InTempName << '.' << method
            << '(' << CompileExpr(&*p) << ");\n";
    }

    /* Input error from console results in "?Redo from start"
     * Input error from file results in error 62 */

    if(!source && !is_linput)
        out << "if(" << InTempName << ".EndInputOkp()) break; \n"
            << "GET_STDOUT() << " << InTempName << ".GetRedoMsg(); }\n";
    else
        out << "if(!" << InTempName << ".EndInputOkp()) B_ERROR(62);\n";

    if(!source && !keep_cursor)
        out << "GET_STDOUT().DoCRLF();\n";

    out << "}";
    return out.str();
}


std::string stmt_labelref::Compile() const
{
    std::ostringstream out;

    #define tokcase(t, p, code) case t: AddSystemCode(code); break;
    switch(token)
    {
#include "libdata-stmt-label.inc"
        default:
            dprintf(dp_errors, "Internal Error: Unknown stmt_labelref token: %s\n", GetTokenName(token).c_str());
            assert(!"unknown stmt_labelref token");
    }
    #undef tokcase

    expr_label* e = 0;
    std::string lab;
    std::string lname;
    if(label)
    {
        e = dynamic_cast<expr_label*> (&*label);
        assert(e);
        lab = e->get_label();
        lname = GenerateLabelName(lab);
    }
    if(lname.empty()
    && token != ON_PLAY
    && token != ON_ERROR
    && token != ON_KEY
    && token != ON_TIMER
    && token != EXITDO && token != EXITFOR && token != EXITSUB
    && token != EXITDEF && token != EXITFUNCTION
    && token != RESUME && token != RESTORE && token != RETURN)
    {
        dprintf(dp_errors, "%s: Error: 0 is a permitted label only in ON ERROR GOTO and similar constructs.\n", GetLoc().stringify().c_str());
    }

    switch(token)
    {
        case EXITDO:
            if(Current_EXITDO_label.empty()) dprintf(dp_errors, "%s: Error: EXIT DO without DO\n", GetLoc().stringify().c_str());
            out << "goto " << Current_EXITDO_label << ";";
            break;
        case EXITFOR:
            if(Current_EXITFOR_label.empty())
                dprintf(dp_errors, "%s: Error: EXIT FOR without FOR\n", GetLoc().stringify().c_str());
            out << "goto " << Current_EXITFOR_label << ";";
            break;
        case EXITSUB:
        case EXITFUNCTION:
        case EXITDEF:
            out << "goto subfunc_exit;";
            break;
        case RETURN:
            if(label)
                out << "B_RETURN_TO(" << lname << ");";
            else
                out << "B_RETURN();";
            break;
        case RESUME:
            if(label)
                out << "B_RESUME_TO(" << lname << ");";
            else
                out << "B_RESUME();";
            break;
        case RESUME_NEXT:
            out << "B_RESUME_NEXT();";
            break;
        case GOTO:
            out << "goto " << lname << ";";
            break;
        case GOSUB:
            out << "B_GOSUB(" << lname << ", gex" << GOSUBlabelCounter << "); gex" << GOSUBlabelCounter << ":;";
            ++GOSUBlabelCounter;
            break;

        case ON_ERROR:
            if(label)
                out << "ON_ERROR_GOTO(" << lname << ");";
            else
                out << "ON_ERROR_GOTO_NOWHERE();";
            break;
        case ON_PLAY: /* FIXME, not fully considered */
            if(label)
                out << "ON_PLAY_GOSUB(" << CompileExpr(&*param) << "," << lname << ");";
            else
                out << "ON_PLAY_GOSUB_NOWHERE();";
            break;
        case ON_KEY: /* FIXME, not fully considered */
            if(label)
                out << "ON_KEY_GOTO(" << lname << ");";
            else
                out << "ON_KEY_GOTO_NOWHERE();";
            break;
        case ON_TIMER: /* FIXME, not fully considered */
            if(label)
                out << "ON_TIMER_GOTO(" << lname << ");";
            else
                out << "ON_TIMER_GOTO_NOWHERE();";
            break;
        case RESTORE:
            out << "B_RESTORE(";
            if(!label) out << '0';
            else
            {
                const Scope& scope = *GetCurrentScope();
                auto i = scope.labels.find(lab);
                if(i != scope.labels.end())
                    out << i->second.dataptr;

                out << " /* " << lname << " */";
            }
            out << ");";
            break;
        default: assert(!"unexpected stmt_goto token");
    }
    return out.str();
}
std::string stmt_label::Compile() const
{
    return GenerateLabelName(label) + ": ;";
}

std::string stmt_fileopen::Compile() const
{
    AddSystemCode(Stream_Common);
    AddSystemCode(R"(
}
#include <unistd.h>   // open,close
#include <fcntl.h>    // O_RDONLY
#include <sys/file.h> // flock
namespace
{
    void B_OPEN(const char* filename, BAS_INT filenumber,  unsigned mode_bitmask, BAS_LNG reclen=0)
    {
        if(fdmap.find(filenumber) != fdmap.end()) B_ERROR(55); // file already open
        if(reclen < 0) B_ERROR(63); // bad record length

        unsigned mode    = mode_bitmask >> 4;

        if(mode < 1 || mode > 4 || (mode_bitmask&3) == 0) B_ERROR(54); // bad file mode

        bool access_read  = mode_bitmask&1;
        bool access_write = mode_bitmask&2;
        bool lock_read    = mode_bitmask&4;
        bool lock_write   = mode_bitmask&8;

        // modes:   
        //   1=RANDOM
        //   2=INPUT and OUTPUT
        //   3=BINARY
        //   4=APPEND

        bool create   = access_write;
        bool truncate = access_write && !access_read && mode!=4;
        bool append   = access_write && mode==4;

        unsigned open_flags = access_read ? access_write ? O_RDWR : O_RDONLY : O_WRONLY;
        if(create)    open_flags |= O_CREAT;
        if(truncate)  open_flags |= O_TRUNC;
        if(append)    open_flags |= O_APPEND;

        if(reclen == 0) reclen = mode==2 ? 512 : 128;
        if(mode == 3) reclen = 1; // For BINARY files, ignore the record length

        int fd = open(filename, open_flags, 0644);
        if(fd < 0) B_ERROR_Errno();

        int lock_status = flock(fd, (lock_write || lock_read) ? LOCK_EX : LOCK_SH);
        if((lock_read || lock_write) && lock_status < 0)
        {
            close(fd);
            B_ERROR_Errno();
        }
        if(fd >= 0)
        {
            fdmap.emplace(filenumber, std::make_shared<FileDescriptor>(fd,reclen, access_read,access_write));
        }
    }

    void B_OPEN(const std::string& filename,BAS_INT filenumber,  unsigned mode_bitmask, BAS_LNG reclen=0)
    {
        B_OPEN(filename.c_str(), filenumber, mode_bitmask, reclen);
    }

    void B_OPEN(const std::string& filename,BAS_INT filenumber, const std::string& mode, unsigned bitmask, BAS_LNG reclen=0)
    {
        if(mode.size() == 1)
            switch(mode[0])
            {
                case 'O': B_OPEN(filename,filenumber, (2 << 4) + (bitmask|2), reclen); return;
                case 'I': B_OPEN(filename,filenumber, (2 << 4) + (bitmask|1), reclen); return;
                case 'R': B_OPEN(filename,filenumber, (1 << 4) + (bitmask|3), reclen); return;
                case 'B': B_OPEN(filename,filenumber, (3 << 4) + (bitmask|3), reclen); return;
                case 'A': B_OPEN(filename,filenumber, (4 << 4) + (bitmask|2), reclen); return;
            }
        B_ERROR(54); // bad file mode
    }
)");
    // B_OPEN(filename,filenumber, mode,     bitmask)
    // B_OPEN(filename,filenumber, mode,     bitmask, reclen)
    // B_OPEN(filename,filenumber, mode_str, bitmask)
    // B_OPEN(filename,filenumber, mode_str, bitmask, reclen)

    // reclen is a runtime parsed integer expression
    // mode_str is a runtime parsed string variable, expect "R","O","I","A","B"
    // mode is:
    //    1 for RANDOM
    //    2 for INPUT and OUTPUT
    //    3 for BINARY
    //    4 for APPEND
    // bitmask is:
    //  &3:
    //    1=reading    (INPUT,         or when overridden by ACCESS: RANDOM,BINARY)
    //    2=writing    (OUTPUT,APPEND, or when overridden by ACCESS: RANDOM,BINARY)
    //    3=read+write (RANDOM,BINARY)
    //    0 = when mode_str is used
    //  &12=
    //    0=shared
    //    4=lock read
    //    8=lock write
    //   12=lock read+write
    std::ostringstream out;
    out << "B_OPEN(" << CompileExpr(&*filename);
    out << "," << CompileExpr(&*filenumber);

    if(access)
        switch(mode)
        {
            case RANDOM: out << ",1/*RANDOM*/*16+"; break;
            default:
            case INPUT:
            case OUTPUT: out << ",2/*IN/OUT*/*16+"; break;
            case BINARY: out << ",3/*BINARY*/*16+"; break;
            case APPEND: out << ",4/*APPEND*/*16+"; break;
        }
    else
    {
        out << "," << CompileExpr(&*mode_str);
        out << ",";
    }

    out << (access + 4*locking);
    out << "/*";
        if(access&1) out << "READ ";
        if(access&2) out << "WRITE ";
        if(access&4) out << "LOCKREAD ";
        if(access&8) out << "LOCKWRITE ";
    out << "*/";

    if(reclen)
    {
        out << "," << CompileExpr(&*reclen);
    }
    out << ");";
    return out.str();
}

std::string stmt_subcall::Compile() const
{
    // temporary sharing of unique_ptr
    std::vector<exprp> f_params( params.size() );
    for(std::size_t a=0; a<params.size(); ++a)
        f_params[a].reset( params[a].get() );

    expr_userfun f(GetLoc(), name, nullptr, std::move(f_params));
    std::string result = f.Compile(exprdef_nothingparticular, true);

    for(std::size_t a=0; a<params.size(); ++a)
        f.get_param(a).release();

    result += ';';
    return result;
}

std::string stmt_redim::Compile() const
{
    expr_var* v = dynamic_cast<expr_var*>(&*target);
    assert(v);
    const std::string& name = v->getname();

    VarContext ctx = FindVariableFromScope(name, true);
    if(!ctx)
        dprintf(dp_errors, "%s: Error: Cannot find definition of '%s' in stmt_redim\n", GetLoc().stringify().c_str(), name.c_str());
    TypeP type = ctx.info->type;

    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*type, UCase(StripTypeSuffix(name)));

    switch(type->arraytype)
    {
        case TypeInfo::NotArray:
        case TypeInfo::StaticArray:
            return "/* REDIM eaten by grue? */";

        case TypeInfo::ParamArray:
            return "FIXME"; // FIXME, TODO

        case TypeInfo::DynamicArray:
        {
            std::string result;
            result += CPP_ARRAY_NAME(base.second);
            result += ".Redim(0,1";
            for(std::size_t a=0; a<dimensions.size(); ++a)
            {
                result += ", " + CompileExpr(&*dimensions[a].first, exprdef_cast_to_int);
                result += ", " + CompileExpr(&*dimensions[a].second, exprdef_cast_to_int);
            }
            result += ");";
            return result;
        }
    }
    return "";
}
std::string stmt_erase::Compile() const
{
    expr_var* v = dynamic_cast<expr_var*>(&*target);
    assert(v);
    const std::string& name = v->getname();
    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*v->GetType(), UCase(StripTypeSuffix(name)));
    VarContext ctx = FindVariableFromScope(name, true);
    if(!ctx)
        dprintf(dp_errors, "%s: Error: Cannot find definition of '%s' in stmt_erase\n", GetLoc().stringify().c_str(), name.c_str());
    TypeP type = ctx.info->type;
    switch(type->arraytype)
    {
        case TypeInfo::NotArray:
            return "/* ERASE eaten by grue? */";

        case TypeInfo::StaticArray:
        {
            std::ostringstream out;
            if(type->dimensions.size() != 1)
            {
                out << "{ " << base.first << "* ep = &";
                out << CPP_ARRAY_NAME(base.second);
                size_t size = 1;
                for(std::size_t a=0; a<type->dimensions.size(); ++a)
                {
                    size *= type->dimensions[a].second;
                    out << "[0]";
                }
                out << "; ";
                out << "std::fill(ep, ep + " << size << ", " << base.first << "{}); }";
            }
            else
            {
                out << "std::fill(" << CPP_ARRAY_NAME(base.second) << ", "
                                    << CPP_ARRAY_NAME(base.second) << " + " << type->dimensions[0].second
                                    << ", " << base.first << "{})";
            }
            return out.str();
        }
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            return CPP_ARRAY_NAME(base.second) + ".Erase();";
    }
    return "";
}

static std::string CompileFunctionCall
    (const LexLocation& lexloc,
     const std::string& functionname,
     const std::vector<exprp>& params,
     const std::vector<SubRoutine::ParamSpec>& paramdefs,
     bool is_statement
       /* Paramdefs components used:
           * type->arraytype == ParamArray
           * passconvention != ByVal
             && was_modified (if needs to be passed by reference)
           * type (for temporary VarDeclaration)
        */
    )
{
    std::ostringstream out;

    out << functionname;

    if(functionname == "B_StmtCLEAR")
    {
        // special case
        bool in_sub = &*GetCurrentScope() != &*GetRootScope();

        if(in_sub)
            out << "<true>";
        else
            out << "<false>";
    }

    out << '(';
    for(std::size_t a=0; a<params.size(); ++a)
    {
        if(a) out << ", ";

        if(!paramdefs[a].type)
        {
            dprintf(dp_errors, "%s: Error: Type of parameter %u of function %s is undefined\n",
                    lexloc.stringify().c_str(),
                    (unsigned)a, functionname.c_str());
            out << CompileExpr(&*params[a], exprdef_nothingparticular);
            continue;
        }

        expr_var* v = dynamic_cast<expr_var*>(&*params[a]);
        if(paramdefs[a].type->arraytype == TypeInfo::ParamArray && v->isarray())
        {
            const TypeInfo& vtype = *v->GetType();
            switch(vtype.arraytype)
            {
                case TypeInfo::NotArray:
                {
                    dprintf(dp_errors, "%s: Error: Attempt to pass a non-array to a function that expects array\n",
                            lexloc.stringify().c_str());
                    break;
                }
                case TypeInfo::StaticArray:
                {
                    const auto base = GetTypeBaseNames(*paramdefs[a].type, "");
                    if(paramdefs[a].type->Type != vtype.Type)
                    {
                        dprintf(dp_errors, "%s: Error: Attempt to pass an array of incorrect type (expected %s, got %s)\n",
                            lexloc.stringify().c_str(),
                            base.first.c_str(), GetTypeBaseNames(vtype,"").first.c_str());
                    }

                    out << "BAS_MAGICARRAY<" << base.first << ">(";
                    out << "&";
                    out << CompileExpr(&*params[a], exprdef_nothingparticular);
                    for(std::size_t a=0; a<vtype.dimensions.size(); ++a)
                        out << "[0]";
                    out << ", BAS_MAGICARRAY_SIZETAG<" << vtype.dimensions.size() << ">()";
                    for(std::size_t a=0; a<vtype.dimensions.size(); ++a)
                        out << ", " << vtype.dimensions[a].first
                            << ','  << vtype.dimensions[a].second;
                    out << ")";
                    continue;
                }

                case TypeInfo::DynamicArray:
                    /* DynamicArray can be passed using an implicit constructor. */
                    break;

                case TypeInfo::ParamArray:
                    /* Param array does not need special care
                     * when the source is also a param array
                     */

                    // Note: QuickBASIC does not support this!
                    // TODO: Check whether we do either.
                    break;
            }
        }

        bool needed_cast = false;

        if(paramdefs[a].passconvention != SubRoutine::ParamSpec::ByVal
        && paramdefs[a].was_modified
        && paramdefs[a].type->arraytype != TypeInfo::ParamArray)
        {
            /* Check if the parameter to be passed is a temporary. */
            /* If it is, a temporary holder needs to be constructed. */
            if(!dynamic_cast<expr_var*>(&*params[a]))
            {
                /* It's not a var, so we need a temporary for the reference to work. */
                /* Let's have the C++ compiler create the temporary for us.
                 * We need to prod a bit the compiler towards creating a temporary
                 * reference with static_cast<const type&> (which is permitted
                 * by C++ standard and happens pretty frequently), and then we
                 * just cast the const away and everyone is happy.
                 */
                needed_cast = true;
                std::string typestr = GetTypeBaseNames(*paramdefs[a].type, "").first;
                out << "const_cast<" << typestr << "&>(static_cast<const " << typestr << "&>(";
            }
        }

        out << CompileExpr(&*params[a], exprdef_nothingparticular);

        if(needed_cast)
            out << "))";
    }
    out << ')';

    return out.str();
}

static std::vector<SubRoutine::ParamSpec>
    CreateParamDefsFromPattern(const std::string& pattern,
                               std::vector<exprp>& params)
{
    std::vector<SubRoutine::ParamSpec> paramdefs(params.size());

    unsigned paramno=0;
    const char* defs = pattern.c_str();
    while(*defs)
    {
        #define EP() do { \
            while(paramno >= params.size()) params.push_back(nullptr); \
            if(paramno >= paramdefs.size()) paramdefs.resize(paramno+1); } while(0)
        #define PTYPE() do { \
            if(params[paramno]) \
                paramdefs[paramno].type = params[paramno]->GetType(); \
            else \
                dprintf(dp_warnings, "Warning: param %u is empty\n", paramno); \
            } while(0)
        #define SetByVal() do { \
            paramdefs[paramno].was_modified = false; \
            paramdefs[paramno].passconvention = SubRoutine::ParamSpec::ByVal; } while(0)
        #define SetByRef() do { \
            paramdefs[paramno].was_modified = true; \
            paramdefs[paramno].passconvention = SubRoutine::ParamSpec::ByRef; } while(0)
        #define InsInt(v) params.emplace(params.begin()+paramno, new expr_number(v))
        #define InsStr(s) params.emplace(params.begin()+paramno, new expr_string(s))

        if(paramno+1 > params.size()) params.push_back(nullptr);
        switch(*defs++)
        {
            // Optional numeric argument: Defaults to -1
            case 'n': case 'x':
                EP(); if(!params[paramno]) { params[paramno].reset( new expr_number(-1) ); }
                //passthru
            case 'N': case 'S': case 'X': // number or string
              GotS:
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            // Optional string argument. Defaults to empty string.
            case 's':
                EP(); if(!params[paramno]) { params[paramno].reset( new expr_string("") ); }
                goto GotS;
            // Optional step-coordinate argument. Defaults to 0,0,-1
            case 'p':
                EP(); if(!params[paramno]) { params[paramno].reset( new expr_number(0) );
                                             InsInt(0); InsInt(-1); }
                //passthru
            case 'P': // step coordinate
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            // Optional coordinate arguments. Defaults to 0,-1
            case 'c':
                EP(); if(!params[paramno]) { params[paramno].reset( new expr_number(0) );
                                             InsInt(-1); }
                //passthru
            case 'C': // coordinate
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            case 'V': // lvalue
                EP(); PTYPE(); SetByRef(); ++paramno;
                break;
            case 'A': // array target or source (GET, PUT, PALETTE USING)
                EP(); PTYPE(); SetByRef(); ++paramno;
                break;
            case 'L': // label
            case 'l':
                assert(!"Label params should only occur in stmt_labelref");
            default:
                dprintf(dp_errors, "Internal Error: pattern '%s': unknown char '%c'\n", pattern.c_str(), *defs);
        }

        #undef EP
        #undef PTYPE
        #undef SetByVal
        #undef SetByRef
        #undef InsInt
        #undef InsStr
    }
    /*
    fprintf(stderr, "params.size=%u, paramno=%u, paramdefs.size=%u, pattern=%s\n",
        (unsigned)params.size(), paramno, (unsigned)paramdefs.size(), pattern.c_str());
    */
    return paramdefs;
}

std::string stmt_lib_stmt::Compile() const
{
    const char* funcname = 0;
    #define tokcase(t, p, code) \
        case t: funcname = "B_Stmt"#t; AddSystemCode(code); goto DoParams;
    switch(token)
    {
/* #include "libdata-stmt-label.inc" - handled separately */
#include "libdata-stmt-stmt.inc"

        default:
            dprintf(dp_errors, "Internal Error: Unknown stmt_lib_stmt token: %s\n", GetTokenName(token).c_str());
            assert(!"unknown stmt_lib_stmt token");
    }
    return "";
DoParams:
    // Temporary sharing of unique_ptr
    std::vector<exprp> plist;
    for(std::size_t a=0; a<params.size(); ++a) plist.emplace_back( params[a].get() );

    std::vector<SubRoutine::ParamSpec> paramdefs = CreateParamDefsFromPattern(pattern, plist);

    std::string result = CompileFunctionCall(
        GetLoc(),
        funcname,
        plist,
        paramdefs,
        true
    );
    #undef tokcase

    for(std::size_t a=0; a<params.size(); ++a) if(params[a]) plist[a].release();

    result += ';';
    return result;
}

std::string expr_string::Compile(ExprCompileDefs defs) const
{
    return CompileStringConst
        (this->value,
           defs==exprdef_may_produce_char_from_string
           ? StrCompilePrinting
           : StrCompileMultipurpose);
}
std::string expr_number::Compile(ExprCompileDefs defs) const
{
    std::ostringstream out;
    if(is_int())
    {
        out << get_i();
        if(type->is_unsigned()) out << 'u';
        if(type->Type == TypeType::Long
        || type->Type == TypeType::UnsignedLong) out << 'l';
    }
    else
    {
        out.precision(20);
        out << get_f();
        /* FIXME: Prevent integer overflows */
    }
    return out.str();
}
std::string expr_unary::Compile(ExprCompileDefs defs) const
{
    ExprCompileDefs subdefs = exprdef_nothingparticular;
    std::string prefix = "";
    switch(token)
    {
        case NOT:
            if(defs == exprdef_pure_condition)
                { prefix = "!"; subdefs=defs; break; }
            prefix = "~";
            subdefs = exprdef_cast_to_int;
            break;
        case MINUS:
            prefix = "-";
            break;
        default:
            dprintf(dp_errors, "Internal Error: Unknown expr_unary token: %s\n", GetTokenName(token).c_str());
            assert(!"unexpected expr_unary token");
    }
    return "(" + prefix + CompileExpr(&*p, subdefs) + ")";
}

/* OptimizeStringComparison():
 *    Given a string expression, produce:
 *              a string expression
 *              character offset
 *              maximum number of characters
 */
struct SubStringExpr
{
    const expr* expression;
    std::size_t offset = 0;
    std::size_t limit  = ~std::size_t(0);
};
static void AcquireSubStringExpression(SubStringExpr& expr)
{
    // At entry: offset=0, limit=~0
    if(const expr_string* s = dynamic_cast<const expr_string*> (expr.expression))
    {
        expr.limit = s->get().size();
        return;
    }

    const expr_builtin_string* e = dynamic_cast<const expr_builtin_string*>(expr.expression);
    if(!e) return;

    if(e->GetToken() == LEFT_S                              // LEFT$(stringexpr, n)
    || (e->GetToken() == MID_S && e->GetNumParams() == 2))  // MID$(stringexpr, n)
    {
        const expr_number* n1 = dynamic_cast<const expr_number*> (e->GetParam(1).get());
        if(!n1)

        // At this point, expr.offset and limit should still be 0 and ~0 respectively.
        expr.expression = e->GetParam(0).get();
        AcquireSubStringExpression(expr);

        std::size_t limit_expr = std::size_t( n1->get_i() );
        if(limit_expr > expr.limit)
        {
            // TODO: context
            dprintf(dp_warnings, "Warning: LEFT$() number is out of range\n");
        }
        expr.limit = std::min(expr.limit, limit_expr);
        return;
    }
    if(e->GetToken() == MID_S)
    {
        const expr_number* n1 = dynamic_cast<const expr_number*> (e->GetParam(1).get());
        const expr_number* n2 = dynamic_cast<const expr_number*> (e->GetParam(2).get());
        if(!n1 || !n2)
        {
            return;
        }

        // At this point, expr.offset and limit should still be 0 and ~0 respectively.
        expr.expression = e->GetParam(0).get();
        AcquireSubStringExpression(expr);

        std::size_t offset_expr = std::size_t( n1->get_i() - 1);
        std::size_t limit_expr  = std::size_t( n2->get_i() );

        if(expr.limit != ~std::size_t(0))
        {
            if(offset_expr > expr.limit)
                expr.limit = 0;
            else
                expr.limit -= offset_expr;
        }

        expr.offset += offset_expr;
        expr.limit = std::min(expr.limit, limit_expr);

        return;
    }
}

std::string expr_binary::Compile(ExprCompileDefs defs) const
{
    std::string fucast, funame, cast0, op, cast1, finish;
    ExprCompileDefs subdefs = exprdef_nothingparticular;

    std::ostringstream out;

    switch(token)
    {
        case XOR:
            op = "^";
        FinishInteger:
            subdefs = exprdef_cast_to_int;
            break;
        case qbcSHL: op = "<<"; goto FinishInteger;
        case qbcSHR: op = ">>"; goto FinishInteger;
        case OR:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "||"; subdefs=defs; break; }
            op = "|"; goto FinishInteger;
        case AND:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "&&"; subdefs=defs; break; }
            op = "&"; goto FinishInteger;
        case GREATERTHAN:
            op = ">";
        FinishCompare:
            if(defs == exprdef_pure_condition)
                subdefs = defs;
            else
            {
                fucast = "(";
                finish = ") ? -1 : 0";
            }
            break;
        case LESSTHAN:        op = "<"; goto FinishCompare;
        case EQUALS:          op = "=="; goto FinishCompare;
        case NOTEQUAL:        op = "!="; goto FinishCompare;
        case LESSOREQUAL:     op = "<="; goto FinishCompare;
        case GREATEROREQUAL:  op = ">="; goto FinishCompare;
        case PLUS:            op = "+"; break;
        case MINUS:           op = "-"; break;
        case BKSLASH:         op = "/"; subdefs = exprdef_cast_to_int; break;
        case ASTERISK:        op = "*"; break;
        case SLASH:           op = "/"; cast1 = "(BAS_DBL)"; break;
        case CIRCUMFLEX:      funame = "std::pow"; op=","; finish=""; break;
        case MOD:             op = "%"; goto FinishInteger;
        default:
            dprintf(dp_errors, "Internal Error: Unknown expr_binary token: %s\n", GetTokenName(token).c_str());
            assert(!"unexpected expr_binary token");
    }

    if(p0->GetType()->is_string()
    && p1->GetType()->is_string())
    {
        int gt=0, eq=-1;
        switch(token)
        {
            case GREATERTHAN:    gt=1, eq=0; break;
            case LESSTHAN:       gt=-1,eq=0; break;
            case EQUALS:         gt=0, eq=1; break;
            case NOTEQUAL:       gt=0, eq=0; break;
            case LESSOREQUAL:    gt=-1,eq=1; break;
            case GREATEROREQUAL: gt=1, eq=1; break;
            default:  goto not_string_compare;
        }
        if(eq != -1)
        {
            /* Compile into std::string::compare() */

            out << fucast << funame << "(";

            SubStringExpr lhs; lhs.expression = p0.get();
            SubStringExpr rhs; rhs.expression = p1.get();

            /*
            Due to optimization, it shouldn't be possible for both
            strings to be constants. If the lefthand side is a constant,
            swap the operands so that the lefthand side is an actual
            object.
            */
            if(p0->is_const())
            {
                std::swap(lhs, rhs);
                gt = -gt;
            }

            AcquireSubStringExpression(lhs);
            AcquireSubStringExpression(rhs);

            if(lhs.offset == 0 && lhs.limit == ~std::size_t(0)
            && (rhs.offset != 0 || rhs.limit != ~std::size_t(0)))
            {
                if(!rhs.expression->is_const())
                {
                    /* Swap them again to get less parameters */
                    std::swap(lhs, rhs);
                    gt = -gt;
                }
            }

            op = "";
            if(gt != 0)            op += (gt > 0  ? '>' : '<');
            else                   op += (eq != 0 ? '=' : '!');
            if(eq != 0 || gt == 0) op += '=';

            out << cast0;

            if(const expr_string* s = dynamic_cast<const expr_string*> (lhs.expression))
                out << CompileStringConst(s->get(), StrCompileConstOnly, false);
            else
                out << CompileExpr(lhs.expression, subdefs);

            if(rhs.limit == 0 && lhs.offset == 0 && lhs.limit == ~std::size_t(0) && gt == 0)
            {
                out << ".empty()";
                if(!eq) out << "==false";
                out << finish << ")";
                return out.str();
            }

            out << ".compare(";
                if(lhs.offset != 0
                || lhs.limit   != ~std::size_t(0)
                || rhs.offset != 0
                || rhs.limit   != ~std::size_t(0)) out << lhs.offset << ", ";
                if(lhs.limit  != ~std::size_t(0)) out << lhs.limit << ", ";
                else if(rhs.offset != 0
                     || rhs.limit   != ~std::size_t(0))
                     {
                         out << "std::string::npos, ";
                     }

                out << cast1;

                bool rhs_const = false;
                if(const expr_string* s = dynamic_cast<const expr_string*> (rhs.expression))
                {
                    out << CompileStringConst(s->get(), StrCompileConstOnly, false);
                    rhs_const = true;
                }
                else
                    out << CompileExpr(rhs.expression, subdefs);

                if(rhs.offset != 0 || (!rhs_const && rhs.limit  != ~std::size_t(0)))
                    out << ", " << rhs.offset;

                if(rhs.limit  != ~std::size_t(0)) out << ", " << rhs.limit;
            out << ")" << op << "0";

            out << finish << ")";
            return out.str();
        }
    not_string_compare:;
    }

    out << fucast << funame << "("
        << cast0 << CompileExpr(&*p0, subdefs);
    out << ' ' << op << ' '
        << cast1 << CompileExpr(&*p1, subdefs);
    out << finish
        << ')';

    return out.str();
}

std::string expr_userfun::Compile(ExprCompileDefs defs) const
{
    return Compile(defs, false);
}

std::string expr_userfun::Compile(ExprCompileDefs defs, bool is_statement) const
{
    std::string basename = UCase(StripTypeSuffix(name));
    const SubRoutine* subp = FindSubFunc(basename);
    if(!subp)
    {
        dprintf(dp_errors, "%s: Error: Sub/Func %s (%s) not found\n", lexloc.stringify().c_str(), name.c_str(), basename.c_str());
        return "";
    }
    const auto& paramdefs = subp->params;

    if(params.size() != paramdefs.size())
    {
        dprintf(dp_errors, "%s: Error: Function %s is given %u arguments, expects only %u\n", lexloc.stringify().c_str(),
            name.c_str(), (unsigned)params.size(), (unsigned)paramdefs.size());
        return "";
    }

    return CompileFunctionCall(
        lexloc,
        CPP_SUBFUNC_NAME(basename),
        params,
        paramdefs,
        is_statement
    );
}

std::string expr_builtin_function::Compile(ExprCompileDefs defs) const
{
    std::string result;
    if(lexloc.line && MakeLineLines)
    {
        result += "\n" + lexloc.make_cpp_line();
    }

    const char* funcname = 0;
    #define tokcase(t, p, code) \
        case t: funcname = "B_Func"#t; AddSystemCode(code); goto DoParams;
    switch(token)
    {
        case SIN: funcname = "std::sin"; goto DoParams;
        case COS: funcname = "std::cos"; goto DoParams;
        case TAN: funcname = "std::tan"; goto DoParams;
        case qbcASIN: funcname = "std::asin"; goto DoParams;
        case qbcACOS: funcname = "std::acos"; goto DoParams;
        case ATN: funcname = "std::atan"; goto DoParams;
        case qbcSINH: funcname = "std::sinh"; goto DoParams;
        case qbcCOSH: funcname = "std::cosh"; goto DoParams;
        case qbcTANH: funcname = "std::tanh"; goto DoParams;
        case qbcASINH: funcname = "std::asinh"; goto DoParams;
        case qbcACOSH: funcname = "std::acosh"; goto DoParams;
        case qbcATANH: funcname = "std::atanh"; goto DoParams;
        case qbcATAN2: funcname = "std::atan2"; goto DoParams;
        case ABS:
            if(params[0]->GetType()->is_integer())
                return GetTypeBaseNames(GetType()->Type,"").first + "(std::labs(" + CompileExpr(&*params[0]) + "))";
                funcname = "std::fabs";
            goto DoParams;
        case LOG: funcname = "std::log"; goto DoParams;
        case EXP: funcname = "std::exp"; goto DoParams;
        case SQR: funcname = "std::sqrt"; goto DoParams;
        case INT: funcname = "std::floor"; goto DoParams;
        case FIX: funcname = "std::trunc"; goto DoParams;
        case CINT: funcname = "(BAS_INT)std::round"; goto DoParams;
        case CLNG: funcname = "(BAS_LNG)std::round"; goto DoParams;
        case CSNG: funcname = "BAS_SNG"; goto DoParams;
        case CDBL: funcname = "BAS_DBL"; goto DoParams;

#include "libdata-expr_numeric.inc"
#include "libdata-expr_string.inc"

        default:
            dprintf(dp_errors, "Internal Error: Unknown expr token: %s\n", GetTokenName(token).c_str());
            assert(!"unknown expr token");
    }
    return result;
DoParams:

    // Check if we can do special optimizations
    const expr_number* p0n = params.size() >= 1 ? dynamic_cast<const expr_number*>(params[0].get()) : nullptr;
    const expr_number* p1n = params.size() >= 2 ? dynamic_cast<const expr_number*>(params[1].get()) : nullptr;
    const expr_string* p1s = params.size() >= 2 ? dynamic_cast<const expr_string*>(params[1].get()) : nullptr;

    switch(token)
    {
        case LEN:
            if(params[0]->GetType()->is_string())
                return "BAS_ULONG((" + CompileExpr(&*params[0]) + ").size())";
            return "BAS_ULONG(sizeof(" + CompileExpr(&*params[0]) + "))";

        case SPACE_S:
            if(p0n && p0n->get_i() >= 1)
                { return "std::string(" + std::to_string(p0n->get_i()) + ", ' ')"; }
            break;
        case STRING_S:
            if(p0n && p0n->get_i() >= 1
            && p1n && p1n->get_i() >= 0 && p1n->get_i() <= 255)
                { return "std::string(" + std::to_string(p0n->get_i()) + ", char(" + std::to_string(p1n->get_i()) + "))"; }
            if(p0n && p0n->get_i() >= 1
            && p1s && p1s->get().size() >= 1)
                { return "std::string(" + std::to_string(p0n->get_i()) + ", char(" + std::to_string((int)p1s->get()[0]) + "))"; }
            break;
        case ASC:
            if(GetToken() == MID_S)
            {
                // Check if the third parameter of MID_S is a constant 1
                const expr_number* p3n = dynamic_cast<const expr_number*>(GetParam(2).get());
                if(p3n && p3n->get_i() == 1)
                {
                    AddSystemCode(R"(
BAS_INT B_ASC_MID(const std::string& s, int offset)
{
    std::size_t p = offset-1; // produces unsigned
    if(unlikely(p >= s.size())) B_ERROR(5); // illegal function call
    return (unsigned char) s [ p ];
}
)");
                    // Change this function into an ASCMID.
                    std::vector<exprp> plist;
                    // temporary sharing of unique_ptr
                    plist.emplace_back(GetParam(0).get());
                    plist.emplace_back(GetParam(1).get());
                    std::vector<SubRoutine::ParamSpec> paramdefs = CreateParamDefsFromPattern("SN", plist);
                    std::string result = CompileFunctionCall(
                        lexloc,
                        "B_ASC_MID",
                        plist,
                        paramdefs,
                        false
                    );
                    plist[0].release();
                    plist[1].release();
                    return result;
                }
            }
        default: ;
    }

    // Temporary sharing of unique_ptr
    std::vector<exprp> plist;
    for(std::size_t a=0; a<params.size(); ++a) plist.emplace_back( params[a].get() );

    std::vector<SubRoutine::ParamSpec> paramdefs = CreateParamDefsFromPattern(pattern, plist);

    result += CompileFunctionCall(
        lexloc,
        funcname,
        plist,
        paramdefs,
        false
    );
    #undef tokcase

    for(std::size_t a=0; a<params.size(); ++a) if(params[a]) plist[a].release();

    return result;
}

std::string expr_lubound::Compile(ExprCompileDefs defs) const
{
    bool kind_u = token==UBOUND;

    const char* fname = kind_u?"UBOUND":"LBOUND"; // Token name used in error messages
    const char* bname = kind_u?"Ubound":"Lbound"; // Method name used in C++ code

    expr_var* v = dynamic_cast<expr_var*>(&*var);
    if(!v) { ae:dprintf(dp_errors, "%s: Error: %s must only be used on arrays\n", lexloc.stringify().c_str(),fname); return ""; }
    if(!v->isarray()) goto ae;

    const std::string& name = v->getname();
    VarContext ctx = FindVariableFromScope(name, true);
    if(!ctx) dprintf(dp_errors, "%s: Error: Cannot find definition of '%s' in expr_lubound\n", lexloc.stringify().c_str(), name.c_str());
    TypeP type = ctx.info->type;

    std::ostringstream out;

    switch(type->arraytype)
    {
        case TypeInfo::NotArray: goto ae;
        case TypeInfo::StaticArray:
        {
            if(expr_number* e = dynamic_cast<expr_number*>(&*num_index))
            {
                std::size_t index = e->get_i() - 1;

                if(index >= type->dimensions.size())
                    dprintf(dp_errors, "%s: Error: Array subscript out of range\n", lexloc.stringify().c_str());
                else
                {
                    expr_number e(
                        type->dimensions[index].first
                      + (kind_u ? type->dimensions[index].second-1 : 0) );
                    out << CompileExpr(&e);
                }
            }
            else
            {
                int min=0, max=0;
                for(std::size_t a=0; a<type->dimensions.size(); ++a)
                {
                    int v = (type->dimensions[a].first + (kind_u ? type->dimensions[a].second-1 : 0));
                    if(a==0 || v<min) min=v;
                    if(a==0 || v>max) max=v;
                }
                std::string typestr;
                if(min>=-128 && max<=127) typestr="BAS_BYTE";
                else if(min>=0 && max<=255) typestr="BAS_UBYTE";
                else if(min>=-32768 && max<=32767) typestr="BAS_INT";
                else if(min>=0 && max<=65535) typestr="BAS_UINT";
                else typestr = (min<0) ? "BAS_LNG" : "BAS_ULONG";

                AddSystemCode(R"(
    template<typename IntType>
    static inline IntType B_StaticBoundFunction(int which_index, const std::initializer_list<IntType>& bounds)
    {
        std::size_t i = which_index-1; // unsigned result
        if(unlikely(i >= bounds.size())) B_ERROR(9); // subscript out of range
        return bounds.begin()[i];
    }
)");
                out << "B_StaticBoundFunction<" << typestr
                    << ">(" << CompileExpr(&*num_index, exprdef_cast_to_int)
                    << ", {";
                for(std::size_t a=0; a<type->dimensions.size(); ++a)
                {
                    if(a) out << ",";
                    out << (type->dimensions[a].first + (kind_u ? type->dimensions[a].second-1 : 0));
                }
                out << "} )";
            }
            break;
        }
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
        {
            std::pair<std::string, std::string>
                base = GetTypeBaseNames(*type, UCase(StripTypeSuffix(name)));

            out << CPP_ARRAY_NAME(base.second)
                << '.' << bname
                << '(';
            out << CompileExpr(&*num_index, exprdef_cast_to_int);
            out << ')';
            break;
        }
    }
    return out.str();
}

// SPECIALS
std::string expr_label::Compile(ExprCompileDefs)    const { assert(!"not reached"); }
std::string expr_caseelse::Compile(ExprCompileDefs) const { assert(!"not reached"); }
std::string expr_caseis::Compile(ExprCompileDefs)   const { assert(!"not reached"); }
std::string expr_caserange::Compile(ExprCompileDefs)const { assert(!"not reached"); }
std::string expr_caseor::Compile(ExprCompileDefs)   const { assert(!"not reached"); }

// LVALUES
std::string expr_var::Compile(ExprCompileDefs defs) const
{
    TypeP basetype = type;
    if(defs != exprdef_membername)
    {
        VarContext ctx = FindVariableFromScope(name, isarray());
        if(!ctx)
        {
            dprintf(dp_errors, "%s: Error: Cannot find definition of '%s' in expr_var\n", lexloc.stringify().c_str(), name.c_str());
        }
        assert(ctx);
        assert(ctx.info);
        basetype = ctx.info->type;
    }
    assert(basetype);

    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*basetype, UCase(StripTypeSuffix(name)));

    if(defs==exprdef_membername)
        base.second = CPP_MEMBER_NAME(base.second);

    std::ostringstream out;

    switch(basetype->arraytype)
    {
        case TypeInfo::NotArray:
            out << base.second;
            break;

        case TypeInfo::StaticArray:
            out << CPP_ARRAY_NAME(base.second);
            for(std::size_t a=0; a<indexes.size(); ++a)
            {
                out << '[';
                if(basetype->dimensions[a].first != 0) out << '(';
                if(CheckStaticSubscripts)
                {
                    AddSystemCode(R"(
    template<long begin, std::size_t size, typename T>
    static inline T B_StaticBoundCheck(T value)
    {
        if(unlikely(std::size_t(value-begin) >= size)) B_ERROR(9); // subscript out of range
        return value;
    }
)");
                    out << "B_StaticBoundCheck<" << basetype->dimensions[a].first
                        << ", " << basetype->dimensions[a].second
                        << ">(";
                }
                out << CompileExpr(indexes[a].get(), exprdef_cast_to_int);
                if(basetype->dimensions[a].first != 0)
                {
                    out << ")-(" << basetype->dimensions[a].first << ')';
                }
                if(CheckStaticSubscripts)
                    out << ')';
                out << ']';
            }
            break;

        case TypeInfo::DynamicArray:
            out << CPP_ARRAY_NAME(base.second);
            if(!array_as_whole)
            {
                out << ".Index(0,0";
                for(std::size_t a=0; a<indexes.size(); ++a)
                    out << ',' << CompileExpr(&*indexes[a], exprdef_cast_to_int);
                out << ')';
            }
            break;

        case TypeInfo::ParamArray:
            out << CPP_ARRAY_NAME(base.second);
            if(!array_as_whole)
            {
                out << ".Index( {";
                for(std::size_t a=0; a<indexes.size(); ++a)
                {
                    if(a) out << ", ";
                    out << CompileExpr(&*indexes[a], exprdef_cast_to_int);
                }
                out << "} )";
            }
            break;
    }
    if(hasmember())
    {
        out << '.';
        out << member->Compile(exprdef_membername); // not CompileExpr here
    }
    return out.str();
}

std::string expr_returnvalueholder::Compile(ExprCompileDefs defs) const
{
    return FunctionReturnVarName;
}

static void VerifyLabels(const Scope& scope)
{
    for(const auto& l: scope.labels)
    {
        if(!l.second.defined)
            dprintf(dp_errors, "Error: Label %s not defined\n", l.first.c_str());
    }
}

static void VerifyProgramOptions(const stmt* main_codep)
{
    std::function<void(const class stmt*)> AnalyzeCode;
    std::function<void(const class expr*,tristate,const std::string&,unsigned)> AnalyzeExpr;

    /* If we encounter ERL, set DoERLhandling = true;
     * If we encounter RESUME or ON ERROR GOTO <not nowhere>, set DoResumeHandling = true;
     * If we encounter ON_PLAY, ON_ERROR, ON_KEY, or ON_TIMER, and the label is set, DoEventChecks = true;
     */
    bool found_resume   = false;
    bool found_on_event = false;
    bool found_on_error = false;
    bool found_erl      = false;

    AnalyzeExpr = [&](const class expr* expr, tristate, const std::string&, unsigned)
    {
        if(!expr) return;
        if(auto e = dynamic_cast<const expr_builtin_numeric*>(expr))
            if(e->GetToken() == ERL)
                found_erl = true;
        expr->ForEachSubExpr(AnalyzeExpr);
    };

    AnalyzeCode = [&](const class stmt* stmt)
    {
        if(!stmt) return;

        if(auto s = dynamic_cast<const stmt_labelref*>(stmt))
            switch(s->token)
            {
                case ON_PLAY:
                case ON_KEY:
                case ON_TIMER:
                    if(!s->label) break;
                    found_on_event = true;
                    break;

                case ON_ERROR:
                    if(!s->label) break;
                    found_on_error = true;
                    break;

                case RESUME:
                    found_resume = true;
                    break;

                default:;
            }

        stmt->ForEachSubStmt(AnalyzeCode);
        stmt->ForEachSubExpr(AnalyzeExpr);
    };

    SetCurrentScope( GetRootScope() );
    AnalyzeCode(main_codep);

    for(const auto& s: GetRootScope()->subroutines)
    {
        SetCurrentScope( s.second.sub_scope );
        AnalyzeCode(s.second.code.get());
    }
    SetCurrentScope( GetRootScope() );

    if(found_on_event && !DoEventChecks.is_true())
    {
        dprintf(dp_warnings, "ON PLAY/ON KEY/ON TIMER found, %s\n",
            DoEventChecks.is_false() ? "but will not work because manually disabled"
                                     : "event checking automatically enabled");
        if(!DoEventChecks.is_false())
            DoEventChecks = true;
    }
    if((found_resume || found_on_error) && !DoResumeHandling.is_true())
    {
        dprintf(dp_warnings, "RESUME/ON ERROR found, %s\n",
            DoResumeHandling.is_false() ? "but will not work because manually disabled"
                                        : "resume support automatically enabled");
        if(!DoResumeHandling.is_false())
            DoResumeHandling = true;
    }
    if(found_erl && !DoERLhandling.is_true() && !DoResumeHandling.is_false())
    {
        dprintf(dp_warnings, "ERL found, %s\n",
            DoResumeHandling.is_false() ? "but will not work because manually disabled"
                                        : "line number information automatically enabled");
        if(!DoERLhandling.is_false())
            DoERLhandling = true;
    }
}

void Compile(
    std::ostream& out,
    const stmt* main_codep,
    const std::vector<std::string>& global_data)
{
    Scope& scope = *GetRootScope();
    SetCurrentScope(&scope);

    VerifyLabels(scope);
    AnalyzeSubParamUsage();

    VerifyProgramOptions(main_codep);

    std::ostringstream mout;

    CompileTypeDefs(mout, scope);

    mout << "/* These are global */\n";
    CompileVars(mout, scope, true); // shared only

    mout << "/* These are local to this module */\n";
    mout << "template<bool insub>\n";
    mout << "static void B_StmtCLEAR(int=-1, int=-1, int=-1);\n";
    CompileSubDeclarations(mout, scope);

  {
    std::ostringstream main_code;
    main_code << "int main(int argc, char**argv)\n";
    main_code << "{\n";
    main_code << "for(int a=1; a<argc; ++a) { if(a>1) B_CmdLine += ' '; B_CmdLine += argv[a]; }\n";
    main_code << "B_LibInit();\n";
    main_code << "B_ERR_FRAME_MAIN_BEGIN();\n";
    main_code << CompileCode(main_codep);

    main_code << "return 0;\n";

    main_code << "B_ERR_FRAME_MAIN_END();\n";
    main_code << "}\n";

    mout << "namespace {\n";
    mout << "/* These are local to main() */\n";
    CompileVars(mout, scope, false); // local only
    mout << "}\n";

    mout << main_code.str();
  }

    /* Compile subs */
    for(const auto& s: scope.subroutines)
    {
        mout << "\n\n";

        mout << "/* BODY OF SUB " << s.first << " */\n" << std::flush;

        const SubRoutine& sub = s.second;

        if(!sub.sub_scope)
        {
            dprintf(dp_warnings, "Warning: SUB %s was declared, but never defined\n", s.first.c_str());
            continue;
        }

        Scope& subscope = *sub.sub_scope;

        SetCurrentScope(&subscope);

        VerifyLabels(subscope);

        CompileSubFunc(mout, s.first, sub, false);
        mout << "{\n";

        bool needs_sub_exit_label = false;
        std::function<void(const stmt*)> scan_stmt = [&](const stmt* s) {
            if(!s) return;
            if(auto l = dynamic_cast<const stmt_labelref*>(s))
                if(l->token == EXITSUB
                || l->token == EXITFUNCTION
                || l->token == EXITDEF)
                    { needs_sub_exit_label = true; }
            s->ForEachSubStmt(scan_stmt);
        };
        scan_stmt(&*sub.code);

        mout << "B_ERR_FRAME_SUB_BEGIN();\n";

        if(sub.returntype)
        {
            FunctionReturnVarName = CompileVarDeclaration(mout, "FunctionReturn", sub.returntype);
            mout << ";\n";
        }

        std::string sub_code = CompileCode(sub.code.get());

        CompileVars(mout, subscope, false); // local only

        mout << sub_code;

        if(needs_sub_exit_label)
            mout << ' ' << "subfunc_exit:; \n";

        if(sub.returntype)
            mout << "    return " << FunctionReturnVarName << ";\n";
        else
            mout << "    return;\n";

        mout << "B_ERR_FRAME_SUB_END();\n";

        mout << "}\n";
    }

    mout << "\n\n";


    /* FILE HEADER */
    if(!DoResumeHandling.is_true())
        out << "#define B_NO_ERROR_HANDLING 1\n";

    out << R"(
#include <qblib.hh>
#include <unordered_map>
#include <ctime>
#include <cmath>

namespace
{
    std::string B_CmdLine;
    std::size_t   B_DataPtr = 0;
)";

    for(const auto& s: SystemCode)
        if(s.find("B_RandomSeed") != s.npos)
        {
            out << "    unsigned long B_RandomSeed = 0x00050000;\n";
            break;
        }

    if(!global_data.empty())
    {
        out << "static constexpr std::size_t DATA_COUNT = " << global_data.size() << ";\n";
        out << "const char* const DATA_TABLE[DATA_COUNT] = {\n";
        unsigned line_length = 0;
        for(std::size_t a=0; a<global_data.size(); ++a)
        {
            //out << "// " << a << "\n";
            std::string s = CompileStringConst(global_data[a], StrCompileConstOnly);
            if(line_length > 0 && line_length + s.size() >= 79)
                { out << "\n"; line_length = 0; }
            else if(line_length)
                { out << ' '; ++line_length; }
            out << s;
            if(a+1 < global_data.size()) out << ',';
            line_length += s.size() + 1;
        }
        if(line_length) out << "\n";
        out << "};\n";
    }

    for(const auto& s: SystemCode) out << s << std::endl;
    out << "}\n";

    out << mout.str();

    /* Create a B_StmtCLEAR function, if needed */
    for(const auto& s: SystemCode)
        if(s.find("void B_CLEAR") != s.npos)
        {
            SetCurrentScope(GetRootScope());
            out << "template<bool insub>\n";
            out << "static void B_StmtCLEAR(int ds, int x, int ss)\n{\n";
            EmitClearVars(out, scope);
            out << "B_CLEAR(ds,x,ss, insub);\n";
            out << "}\n";
            break;
        }
}
