#define PARSER_CLIKE_MAX_CASES 4

#include "blocksetting"

#include "expr.hh"
#include "binding.hh"
#include "stmt.hh"

#include "lexer.hh"
#include "tokens.hh"
#include "compile.hh"

#include "dprintf.hh"

//#define PARSER_DEBUG

#include "parser.hh"
#include "parser_framework.hh"
#include "parser_lexiterator.hh"

#include "util.hh"

typedef Lex<QBtoken>::iterator iterator_t;
typedef scanner<iterator_t> scanner_t;

class QB_parser;

unsigned option_base = 0;

namespace
{
    const SubRoutine* in_sub = 0;
    bool in_block = false;

    struct CoordType
    {
        bool step;
        exprp x,y;
    };
    CoordType coord_res;

    struct Context
    {
        exprp exp_result;
        stmtp stm_result;
        TypeP typ_result;
    };
    Context ctx;
    std::list<Context> saved_contexts;

    void SetExpr(exprp&& new_res)
    {
        ctx.exp_result = std::move(new_res);
        OptimizeExprp(ctx.exp_result);
    }
    void SetStmt(stmtp&& new_stm)
    {
        if(new_stm && ctx.stm_result)
        {
            // Find the first statement in the chain of new_stm,
            // and set ctx.stm_result as the "prev" of that first statement.
            stmt* first = new_stm.get();
            for(;;)
            {
                stmt* prev_of_that = first->get_prev().get();
                if(!prev_of_that) break;
                first = prev_of_that;
            }
            first->set_prev( std::move(ctx.stm_result) );
        }
        ctx.stm_result = std::move(new_stm);
    }
    void SetStmt(stmtp&& new_stm, stmtp&& prev)
    {
        ctx.stm_result = std::move(prev);
        SetStmt( std::move(new_stm) );
    }
    void SetType(TypeP&& new_type)
    {
        ctx.typ_result = std::move(new_type);
    }

    /////////////////////////////////

    void DoNumber(scanner_t& scan)
    {
        TypeInfo t = TypeSuffixIntoType(scan.GetInfo().numtype);
        if(t.Type == TypeType::UserType) t.Type = TypeType::Double;
        SetExpr( exprp(new expr_number(scan.GetInfo().number, std::make_shared<TypeInfo>(t))) );
    }
    void DoString(scanner_t& scan)
    {
        SetExpr( exprp(new expr_string(scan.GetInfo().str)) );
    }

    void DoNumericLabelRef(scanner_t& scan)
    {
        long number = (scan.GetInfo().number);
        if(!number) { SetExpr(exprp(nullptr)); return; }
        SetExpr( exprp(new expr_label( std::to_string( number ) )) );
    }
    void DoIdLabelRef(scanner_t& scan)
    {
        std::string id = scan.GetInfo().str;
        if(HasTypeSuffix(id)) throw InvalidateAction("Not a label");
        SetExpr( exprp(new expr_label( DeclareAsLabelInCurrentScope(id, false) )) );
    }
    void DoNumericLabelDef(scanner_t& scan)
    {
        std::string name = std::to_string( (long) (scan.GetInfo().number) );
        if(!AnchorLabelHere(name)) throw InvalidateAction("Duplicate label");
        SetStmt( stmtp(new stmt_label(scan.GetLoc(), name)) );
    }
    void DoIdLabelDef(scanner_t& scan)
    {
        std::string id = scan.GetInfo().str;
        if(HasTypeSuffix(id)) throw InvalidateAction("Not a label");
        std::string name = DeclareAsLabelInCurrentScope(id, true);
        if(name.empty()) throw InvalidateAction("Duplicate label");
        if(!AnchorLabelHere(name)) throw InvalidateAction("Duplicate label");
        SetStmt( stmtp(new stmt_label(scan.GetLoc(), name)) );
    }

    bool Do_Dim(const LexLocation& lexloc,
                const std::string& varname,
                std::vector<std::pair<exprp, exprp>>&& dim_exprs,
                TypeP type,
                bool is_redim,
                bool is_shared,
                bool is_static)
    {
        if(type && HasTypeSuffix(varname))
        {
            throw InvalidateAction("Type suffix not allowed in DIM..AS");
        }

#ifdef BINDING_DEBUG
        fprintf(stderr, "1: %s %s(%u)\n",
            is_redim ? "redim" : "dim",
            varname.c_str(),
            (unsigned)dim_exprs.size());
#endif
        // Scalar:
        if(dim_exprs.empty())
        {
            if(is_redim)
            {
                // REDIM is invalid with scalars
                throw InvalidateAction("Invalid function call");
            }
            VarContext previous_def = FindVariableFromScope(varname, false);
            if(previous_def)
            {
                throw InvalidateAction("Duplicate definition of " + varname);
            }
            if(!type) type = CreateTypeTemplate(varname);
            return ExplicitlyDefine(varname, type, is_static, is_shared, false, lexloc).get();
        }

        // Array:
        VarContext previous_def = FindVariableFromScope(varname, true);
#ifdef BINDING_DEBUG
        if(previous_def)
            fprintf(stderr, "1b: Array type is %s\n",
                previous_def.info->type->arraytype == TypeInfo::DynamicArray
                    ? "dyn" : "other than dyn");
#endif
        bool dims_are_const = true;
        for(const auto& p: dim_exprs)
            if(!p.first->is_const() || !p.second->is_const())
            {
                dims_are_const = false;
                break;
            }

        bool create_dynamic = IsMetaDynamic() || !dims_are_const;
        if(is_redim && !previous_def)
        {
            create_dynamic = true;
        }
        if(is_redim && previous_def && previous_def.info->type->arraytype == TypeInfo::DynamicArray)
        {
            create_dynamic = true;
        }

        if(previous_def
        && (!is_redim
         || (previous_def.info->type->arraytype != TypeInfo::DynamicArray
          && previous_def.info->type->arraytype != TypeInfo::ParamArray)
           ))
        {
            throw InvalidateAction("Array already dimensioned");
        }

        is_redim = previous_def;

        /*
           At this point, possibilities are:

                previous_def  is_redim  create_dynamic

                        NONE        NO              NO
                        NONE        NO             YES
                     DYNAMIC       YES             ---

         */

#ifdef BINDING_DEBUG
        fprintf(stderr, "2: %s %s array %s (%s)\n",
            is_redim ? "redim" : "dim",
            create_dynamic ? "dyn":"stc",
            varname.c_str(),
            previous_def ? "previously known" : "previously unknown");
#endif
        bool nonconstant_initial_dimensions = false;

        if(is_redim)
        {
            /* Check the number of dimensions, and if it matches,
             * generate a REDIM runtime call. */
            if(type)
            {
                if(*type != *previous_def.info->type)
                    throw InvalidateAction("Type mismatch");
            }
            type = previous_def.info->type;

            if(type->arraytype == TypeInfo::ParamArray)
            {
                // always run REDIM, viability is checked runtime
                create_dynamic = true;
            }
            else
            {
                if(type->dimensions.size() != dim_exprs.size())
                {
                    throw InvalidateAction("Array dimensions are different");
                }
            }
        }
        else
        {
            /* Create the variable first. */

            // Default array dimensions
            if(!type) type = CreateTypeTemplate(varname);
            type->dimensions.resize(dim_exprs.size(),
                                    std::make_pair((long)option_base,
                                                   (unsigned)(10+(1-option_base))));

            /* Create the specific dimensions. */
            for(std::size_t a=0; a<dim_exprs.size(); ++a)
            {
                OptimizeExprp(dim_exprs[a].first);
                OptimizeExprp(dim_exprs[a].second);
                expr_number* begin = dynamic_cast<expr_number*> (&*dim_exprs[a].first);
                expr_number* end   = dynamic_cast<expr_number*> (&*dim_exprs[a].second);

                if(!begin || !end)
                {
                    nonconstant_initial_dimensions = true;
                    if(!create_dynamic)
                    {
                        dprintf(dp_warnings, "Warning: $STATIC array %s turned into a $DYNAMIC one because dimensions are non-constant\n",
                            varname.c_str());
                        create_dynamic = true;
                    }
                    continue;
                }

                long long int min = begin->get_i(), count = end->get_i() - min + 1;
                if(count < 0) throw InvalidateAction("Invalid function call");

                type->dimensions[a].first  = min;
                type->dimensions[a].second = count;
            }

            type->arraytype = create_dynamic ? TypeInfo::DynamicArray : TypeInfo::StaticArray;

            if(!ExplicitlyDefine(varname, type, is_static, is_shared, false, lexloc))
            {
                throw InvalidateAction("Duplicate definition of "+varname);
            }
            /* Then, if dynamic, generate a REDIM function call. */
        }

        if(create_dynamic && (nonconstant_initial_dimensions || is_redim))
        {
            expr_var* e = new expr_var(lexloc, varname);
            e->set_type(type);
            e->set_basetype(type);
            e->setwholearray();
            SetStmt( stmtp(new stmt_redim(lexloc, exprp(e), std::move(dim_exprs))) );
            //fprintf(stderr, "redim stmt created\n");
        }
        return true;
    }
};

struct Ctx1 /* enter context */
{
    void operator() () const
    {
        saved_contexts.emplace_back( std::move(ctx) );
        ctx = Context();
    }
};
struct Ctx0 /* exit context */
{
    void operator() () const
    {
        exprp e = std::move(ctx.exp_result);
        stmtp s = std::move(ctx.stm_result);
        TypeP t = std::move(ctx.typ_result);

        ctx = std::move(saved_contexts.back());
        saved_contexts.pop_back();

        /* If the subcontext produced an expression or a statement,
         * import them to the calling context
         */
        if(e)SetExpr( std::move(e) );
        if(s)SetStmt( std::move(s) );
        if(t)SetType( std::move(t) );

#ifdef PARSER_DEBUG
        dprintf(dp_rulecomego, "> pop ctx, e=%s stmt=%s[prev=%s] type=%s\n",
            ctx.exp_result ? typenameof(*ctx.exp_result).c_str() : "NULL",
            ctx.stm_result ? typenameof(*ctx.stm_result).c_str() : "NULL",
            ctx.stm_result && ctx.stm_result->get_prev() ?
                    typenameof(*ctx.stm_result->get_prev()).c_str() : "NULL",
            ctx.typ_result ? typenameof(*ctx.typ_result).c_str() : "NULL");
#endif
    }

};

template<typename RetType>
static bool ParseLibItem(QB_parser& parser, scanner_t& scan, QBtoken tok,
    const char* defs, std::unique_ptr<RetType>& result,
    std::string& matched_pattern);

template<typename StmtType>
static bool ParseLibCommand(QB_parser& parser, scanner_t& scan, QBtoken tok,
                            const char* defs)
{
    std::unique_ptr<StmtType> result;
    std::string matched_pattern;
    if(!ParseLibItem(parser, scan, tok, defs, result, matched_pattern)) return false;
    result->set_pattern(matched_pattern);
    SetStmt( std::move(result) );
    return true;
}

template<typename ExprType>
static bool ParseLibFunction(QB_parser& parser, scanner_t& scan, QBtoken tok,
                             const char* defs)
{
    std::unique_ptr<ExprType> result;
    std::string matched_pattern;
    if(!ParseLibItem(parser, scan, tok, defs, result, matched_pattern)) return false;
    result->set_pattern(matched_pattern);
    SetExpr( std::move(result) );
    return true;
}


typedef rule<scanner_t, Ctx1, Ctx0> rule_t;

class QB_parser: public parser<QB_parser>
{
    #define tk(x) ch_p(QBtoken(x))
    #define tq(x) ch_p(QBtoken(x)).expected()

    #define num_expr           expr
    #define string_expr        expr
    #define string_or_num_expr expr

    #define LambdaParseBody(x)  ((x).parse(scan))

    #define TypeCheckAny(expr) {}

    #define TypeCheckNum(expr) \
        if(expr->is_num().is_false()) \
           throw InvalidateAction("Type mismatch");

    #define TypeCheckString(expr) \
        if(expr->is_string().is_false()) \
           throw InvalidateAction("Type mismatch");

    #define LambdaExpectToken(tkn, ifnot) \
        do { if(!tq(tkn).parse(scan)) { ifnot; }; } while(0)

    #define LambdaExpectSubRule(rule, ifnot) \
        do { if(!(rule).expected().parse(scan)) { ifnot; }; } while(0)

    #define LambdaExpectAnyExpr(expr, ifnot) \
     do { if(!expr.expected().parse(scan)) { ifnot; }; \
        } while(0)

    #define LambdaExpectNumExpr(expr, ifnot) \
     do { LambdaExpectAnyExpr(expr, ifnot); \
          TypeCheckNum(ctx.exp_result); \
        } while(0)

    #define LambdaExpectStringExpr(expr, ifnot) \
     do { LambdaExpectAnyExpr(expr, ifnot); \
          TypeCheckString(ctx.exp_result); \
        } while(0)

public:
    rule_t program = LambdaParser([this] (scanner_t& scan) -> bool
    {
        /* THE MAIN RULE */

        /* Parse long statements */
        while(!scan.eof() && longstatement.parse(scan)) {}

        /* Check what was left over. */
        while(!scan.eof())
        {
            QBtoken tok = scan.get();
            switch(tok)
            {
                case COLONCHAR:
                case CRLF: break;

                case ENDIF:        throw InvalidateAction("END IF without IF");
                case ENDDEF:       throw InvalidateAction("END DEF without DEF");
                case ENDSUB:       throw InvalidateAction("END SUB without SUB");
                case ENDFUNCTION:  throw InvalidateAction("END FUNCTION without FUNCTION");
                case ENDSELECT:    throw InvalidateAction("END SELECT without SELECT");
                case LOOP:         throw InvalidateAction("LOOP without DO");
                case WEND:         throw InvalidateAction("WEND without WHILE");
                case ENDTYPE:      throw InvalidateAction("END TYPE without TYPE");
                case NEXT:         throw InvalidateAction("NEXT without FOR");
                default: throw InvalidateAction("Unexpected " + GetTokenName(tok));
            }
        }
        return true;
    });

    rule_t type = LambdaTokParser([] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                case INTEGER: GotBit16: SetType(std::make_shared<TypeInfo>(*TypeInt)); return true;
                case SINGLE:  SetType(std::make_shared<TypeInfo>(*TypeSng)); return true;
                case DOUBLE:  SetType(std::make_shared<TypeInfo>(*TypeDbl)); return true;
                case LONG:    SetType(std::make_shared<TypeInfo>(*TypeLng)); return true;
                case qb64BYTE:            GotBit8: SetType(std::make_shared<TypeInfo>(*TypeByte)); return true;
                case qb64UNSIGNEDBYTE:    GotUbit8: SetType(std::make_shared<TypeInfo>(*TypeUbyte)); return true;
                case qb64UNSIGNEDINTEGER: GotUbit32: SetType(std::make_shared<TypeInfo>(*TypeUint));  return true;
                case qb64UNSIGNEDLONG:    SetType(std::make_shared<TypeInfo>(*TypeUlong)); return true;
                case STRING:
                case qb64BIT:
                case qb64UNSIGNEDBIT:
                {
                    TypeP t;
                    if(tok == STRING)          t = std::make_shared<TypeInfo>(TypeType::String);
                    if(tok == qb64BIT)         t = std::make_shared<TypeInfo>(TypeType::Bit);
                    if(tok == qb64UNSIGNEDBIT) t = std::make_shared<TypeInfo>(TypeType::UnsignedBit);
                    if((tk(ASTERISK).parse(scan)))
                    {
                        // TODO: Accept constant-expression here
                        LambdaExpectToken(NUMBER, goto asterisk_expr_missing);
                      {
                        long len = (scan.GetInfo().number);
                        if(len < 1 || len > 65535)
                            yyerror("STRING/_BIT length must be 1..65535\n", scan.GetLoc());
                        t->fixlength = len;

                        if(tok == qb64BIT && len == 8) goto GotBit8;
                        if(tok == qb64BIT && len ==32) goto GotBit8;
                        if(tok == qb64UNSIGNEDBIT && len == 8) goto GotUbit8;
                        if(tok == qb64UNSIGNEDBIT && len ==32) goto GotUbit32;
                      }
                    asterisk_expr_missing:;
                    }
                    else if(tok != STRING)
                        t->fixlength = 1; // Default _BIT*1

                    SetType( std::move(t) );
                    return true;
                }
                case IDENTIFIER:
                {
                    TypeP t = std::make_shared<TypeInfo>(TypeType::UserType);
                    const std::string utypename = scan.GetInfo().str;
                    if(HasTypeSuffix(utypename))
                        throw InvalidateAction("A type name cannot have a type suffix");

                    t->members   = FindUserType(utypename);

                    if(!t->members)
                        throw InvalidateAction("Undefined type");

                    t->utypename = UCase(StripTypeSuffix(utypename));

                    SetType( std::move(t) );
                    return true;
                }
                default: break;
            }
            return false;
        }, "type").expected();

    rule_t anylabeltarget = LambdaTokParser([this] (QBtoken tok, scanner_t& scan) -> bool
        {
            if(tok == NUMBER)     { DoNumericLabelRef(scan); return true; }
            if(tok == IDENTIFIER) { DoIdLabelRef(scan); return true; }
            return false;
        }, "anylabeltarget");

    rule_t plain_coordinate = LambdaParser([this] (scanner_t& scan) -> bool
        {
            if(!tk(OPENPARENS).parse(scan)) return false;
            if(!num_expr.parse(scan)) return false;
            coord_res.x = std::move(ctx.exp_result);
            if(!tk(COMMA).parse(scan)) return false;
            if(!num_expr.parse(scan)) return false;
            coord_res.y = std::move(ctx.exp_result);
            if(!tk(CLOSEPARENS).parse(scan)) return false;
            return true;
        }, "coordinate");

    rule_t step_coordinate = LambdaParser([this] (scanner_t& scan) -> bool
        {
            coord_res.step = tk(STEP).parse(scan);
            return plain_coordinate.parse(scan);
        }, "step coordinate");

    // Parse a sequence that begins with a separator and ends with a separator,
    // but has zero or more statements in the middle.
    // If the separator encountered is a colon, only short-statements are accepted
    // until the next CRLF.
    rule_t separator_statements_and_separator = LambdaParser([this] (scanner_t& scan) -> bool
        {
            SetStmt(nullptr); // begin a new chain of statements
            enum { nothing,colon,crlf } got=nothing;

            if(tk(COLONCHAR).parse(scan)) got = colon;
            else if(tk(CRLF).parse(scan)) got = crlf;
            else { (tk(COLONCHAR)|tk(CRLF)).expected().parse(scan); return false; }

            bool first = true;

            iterator_t it = scan.pos();
            while(!scan.eof())
            {
                if(tk(COLONCHAR).parse(scan))
                {
                    it = scan.pos();
                    got = colon;
                    continue;
                }
                if(tk(CRLF).parse(scan))
                {
                    it = scan.pos();
                    got = crlf;
                    continue;
                }
                auto loc = scan.GetLoc();

                if(got == nothing) break;
                if(got == colon && shortstatement.parse(scan)) got = nothing;
                else if(got == crlf && longstatement.parse(scan)) got = crlf; // longstatement always ends in a crlf
                else if(got == colon && first && longblockstatement.parse(scan))
                {
                    // as mercy, we allow long IF in these contexts
                    dprintf(dp_warnings, "%s: Warning: QuickBASIC wouldn't allow long block statements after a ':' character.\n",
                        loc.stringify().c_str());
                    // not terminated
                    got = nothing;
                }
                else if(shortif.parse(scan)) got = nothing;
                else break;

                it = scan.pos();
                first = false;
            }
            if(got != nothing) return true;
            scan.revert(it);
            return false;
        }, "statements");

    // Parses a sequence that mandatorily begins with a colonchar and also mandatorily ends with a colonchar,
    // but has zero or more short-statements in the middle.
    rule_t colonchars_shortstatements_and_colonchar = LambdaParser([this] (scanner_t& scan) -> bool
        {
            SetStmt(nullptr); // begin a new chain of statements
            LambdaExpectToken(COLONCHAR, return false);
            bool got_colonchar = true;
            iterator_t it = scan.pos();
            while(!scan.eof())
            {
                if(tk(COLONCHAR).parse(scan))
                {
                    it = scan.pos();
                    got_colonchar = true;
                    continue;
                }
                if(!got_colonchar) break;
                if(!shortstatement.parse(scan)) break;
                got_colonchar=false;
            }
            if(got_colonchar) return true;
            scan.revert(it);
            return false;
        }, "statements");

    rule_t shortif = LambdaTokParser([this] (QBtoken tok, scanner_t& scan) -> bool
        {
            if(tok != IF) return false;

            const block_setting<bool> block_marker(in_block, true);

            LambdaExpectNumExpr(expr, return false);

            std::unique_ptr<stmt_if> if_stmt(new stmt_if(scan.GetLoc(), std::move(ctx.exp_result)));

            // FIXME: IF x=5 GOTO label is a valid shortif!

            LambdaExpectToken(THEN, return true);

            stmtp prev = std::move(ctx.stm_result); // Save the end of statements chain
            bool got_else = false;
            bool got_stmt = false;

            std::list<std::pair<exprp, stmtp>> elseifp;
            stmtp thenp, elsep;

            for(;;)
            {
                // Previous token is a "THEN" or a "ELSE".
                if(tk(NUMBER).parse(scan))
                {
                    /* Translate into GOTO */
                    DoNumericLabelRef(scan);

                    SetStmt(nullptr); // begin a new chain of statements
                    SetStmt( stmtp(new stmt_labelref(scan.GetLoc(), GOTO, std::move(ctx.exp_result))) );
                    got_stmt = true;

                    while(tk(COLONCHAR).parse(scan));
                }
                else
                {
                    bool stmt_ok = true;
                    while(!scan.eof())
                    {
                        if(tk(COLONCHAR).parse(scan))  { got_stmt = true; stmt_ok = true; continue; }
                        if(!stmt_ok) break;
                        if(shortstatement.parse(scan)) { got_stmt = true; stmt_ok = false; continue; }
                        if(shortif.parse(scan))        { got_stmt = true; break; }
                        break;
                    }
                }

                if(got_else)
                    elsep = std::move(ctx.stm_result);
                else
                    thenp = std::move(ctx.stm_result);

                if(got_else) break;
                if(!tk(ELSE).parse(scan)) break;
                got_else = true;
            }
            // There must be something after THEN and before CRLF, for it to be a valid shortif.
            if(!got_stmt && !got_else) return false;

            if_stmt->AddBody( std::move(thenp), std::move(elseifp), std::move(elsep) );

            SetStmt( std::move(if_stmt), std::move(prev) );
            return true;
        }, "IF clause");

    rule_t casematch = LambdaParser([this] (scanner_t& scan) -> bool
        {
            exprp prev;
            for(;;)
            {
                if(tk(ELSE).parse(scan))
                {
                    SetExpr( exprp(new expr_caseelse) );
                }
                else if(tk(IS).parse(scan))
                {
                    bool invert = false;
                    QBtoken comptok = QBtoken();
                    for(iterator_t it = scan.pos(); !scan.eof(); it = scan.pos())
                    {
                        QBtoken tok = scan.get();
                        switch(tok)
                        {
                            case NOT: invert=!invert; break;
                            case LESSTHAN:
                            case EQUALS:
                            case GREATERTHAN:
                            case NOTEQUAL:
                            case LESSOREQUAL:
                            case GREATEROREQUAL:
                                if(comptok) { tok=QBtoken(); break; } // Allow only one
                                comptok = tok;
                                break;
                            default:
                                tok=QBtoken();
                        }
                        if(tok==QBtoken()) { scan.revert(it); break; }
                    }
                    if(comptok==QBtoken()) comptok = EQUALS; // '=' is optional
                    if(invert)
                        switch(comptok)
                        {
                            case LESSTHAN:     comptok = GREATEROREQUAL; break;
                            default:
                            case EQUALS:       comptok = NOTEQUAL; break;
                            case GREATERTHAN:  comptok = LESSOREQUAL; break;
                            case NOTEQUAL:      comptok = EQUALS; break;
                            case LESSOREQUAL:   comptok = GREATERTHAN; break;
                            case GREATEROREQUAL:comptok = LESSTHAN; break;
                        }
                    LambdaExpectAnyExpr((expr), goto expr_missing);
                    SetExpr( exprp(new expr_caseis(comptok, std::move(ctx.exp_result))) );
                }
                else
                {
                    LambdaExpectAnyExpr((expr), goto expr_missing);
                    exprp e = std::move(ctx.exp_result);
                    if(tk(TO).parse(scan))
                    {
                        LambdaExpectAnyExpr((expr), goto expr_missing);
                        SetExpr( exprp(new expr_caserange( std::move(e), std::move(ctx.exp_result))) );
                    }
                    else
                        SetExpr( exprp(new expr_caseis(EQUALS, std::move(e))) );
                }
                if(prev)
                    SetExpr( exprp( new expr_caseor( std::move(prev), std::move(ctx.exp_result) ) ) );

            expr_missing:;
                if(!tk(COMMA).parse(scan)) break;

                prev = std::move(ctx.exp_result);
            }
            return true;
        }, "case option list");

    rule_t simplestatement =
        LambdaTokParser
        (
            [this] (QBtoken tok, scanner_t& scan) -> bool
            {
                const char* pattern = nullptr;

                LexLocation lexloc = scan.GetLoc();

                //fprintf(stderr, "Tok=%s\n", GetTokenName(tok).c_str());

                /* LIBSTATEMENTS */
                switch(tok)
                {
                    default:break;
                    got_stmt_libstmt:
                        return (ParseLibCommand<stmt_lib_stmt>)(*this, scan, tok, pattern);
                    got_stmt_labelref:
                        return (ParseLibCommand<stmt_labelref>)(*this, scan, tok, pattern);

                    #define tokcase(s,p,code) case s: { pattern=p; goto got_stmt_libstmt; }
#include "libdata-stmt-stmt.inc"
                    #undef tokcase

                    #define tokcase(s,p,code) case s: { pattern=p; goto got_stmt_labelref; }
#include "libdata-stmt-label.inc"
                    #undef tokcase

                    /* More complex statements */

                    case LET:
                    {
                        LambdaExpectSubRule((var_target), return true);
                        exprp target = std::move(ctx.exp_result);
                        LambdaExpectToken(EQUALS,   return true);
                        LambdaExpectSubRule((expr), return true);
                        SetStmt( stmtp(new stmt_let(lexloc, std::move(target), std::move(ctx.exp_result) )) );
                        return true;
                    }

                    /* explicit sub call */
                    case CALL:
                    {
                        LambdaExpectToken(IDENTIFIER, return true);
                        // TODO: check namespace
                        const std::string subname = scan.GetInfo().str;
                        std::vector<exprp> params;
                        if(tk(OPENPARENS).parse(scan))
                        {
                            do {
                                LambdaExpectSubRule((func_param), goto call_expr_missing);
                                params.emplace_back( std::move(ctx.exp_result) );
                            call_expr_missing:;
                            } while(tk(COMMA).parse(scan));
                            LambdaExpectToken(CLOSEPARENS, );
                        }
                        SetStmt( stmtp(new stmt_subcall(lexloc, subname, std::move(params) )) );
                        return true;
                    }

                    /* single-line DEF. multiline DEF is already caught in longstatement. */
                    case DEF:
                    {
                        if(in_sub || in_block)
                            throw InvalidateAction("SUB/FUNCTION/DEF not allowed in control statements");


                        LambdaExpectToken(FNIDENTIFIER, return true);
                        std::string subname = scan.GetInfo().str;
                        std::vector<SubRoutine::ParamSpec> params;

                        TypeP returntype = CreateTypeTemplate(subname);

                        // This parameter parsing is mostly duplicate from the multi-line
                        // SUB/FUNCTION/DEF parameter parsing, except here we won't handle
                        // some special cases, like array parameters.

                        if(tk(OPENPARENS).parse(scan) && !tk(CLOSEPARENS).parse(scan)) /* optional */
                        {
                            do {
                                SubRoutine::ParamSpec param;

                                if(tk(BYVAL).parse(scan)) { param.SetByval(); }

                                LambdaExpectToken(IDENTIFIER, goto def_param_incomplete);
                                param.name     = scan.GetInfo().str;
                                param.is_array = false;
                                param.was_modified = true; // assume worst

                                param.type = CreateTypeTemplate(param.name);
                                if(tk(AS).parse(scan))
                                {
                                    if(HasTypeSuffix(param.name))
                                    {
                                        throw InvalidateAction("Invalid AS together with type suffix");
                                    }
                                    LambdaExpectSubRule((type), return true);
                                    param.type = ctx.typ_result;
                                }
                                params.push_back(param);
                            def_param_incomplete:;
                            } while(tk(COMMA).parse(scan));
                            LambdaExpectToken(CLOSEPARENS, );
                        }
                        LambdaExpectToken(EQUALS, );

                        const SubRoutine* sub = DefineSubFunc(false, subname, returntype, params);
                        if(!sub) throw InvalidateAction("Duplicate definition of "+subname);

                        /* For DEFs, ALL variables are SHARED even if not declared SHARED */
                        EnterSubFuncScope(subname, true);

                        const block_setting<const SubRoutine*> block_marker(in_sub, sub);

                        LambdaExpectSubRule((expr), return true);

                        // The sub contains just one statement.
                        ExitSubFuncScope(subname,
                                         stmtp(new stmt_let(lexloc, exprp(new expr_returnvalueholder(lexloc, subname)),
                                                           std::move(ctx.exp_result)) ) );

                        // We did not change ctx.stm_result.
                        return true;
                    }

                    /* Note: QuickBasic allows CONSTs in control structures.
                     * But they are unconditional.
                     */
                    case CONST:
                    {
                        do {
                            LambdaExpectToken(IDENTIFIER, );
                            std::string id = scan.GetInfo().str;
                            LambdaExpectToken(EQUALS, );
                            LambdaExpectSubRule((expr), );

                            if(!id.empty() && ctx.exp_result)
                            {
                                OptimizeExprp(ctx.exp_result);
                                if(!ctx.exp_result->is_const())
                                {
                                    throw InvalidateAction("Expression is not constant");
                                }
                                int r = DefineConstant(id, std::move(ctx.exp_result));
                                if(r == -1) throw InvalidateAction("Type mismatch");
                                if(r == 0) throw InvalidateAction("Duplicate definition of "+id);
                            }
                        } while( tk(COMMA).parse(scan) );
                        return true;
                    }

                    /* Note: QuickBasic allows OPTION BASE in control structures.
                     * But they are unconditional. Also, only 1 is allowed.
                     * Attempt to use more will produce compile-time error
                     * "Array already dimensioned".
                     */
                    case OPTION:
                    {
                        LambdaExpectToken(BASE, return true);

                        if(in_sub)
                            throw InvalidateAction("OPTION BASE not allowed in SUB/FUNCTION/DEF");

                        LambdaExpectToken(NUMBER, return true);
                        if(scan.GetInfo().number != 0.0
                        && scan.GetInfo().number != 1.0)
                        {
                            throw InvalidateAction("Expected: 0 or 1\n");
                        }
                        option_base = (unsigned)scan.GetInfo().number;
                        return true;
                    }

                    /* Note: DATA cannot exist in SUBs.
                     * However, Quickbasic allows them inside control structures.
                     * But they are unconditional.
                     */
                    case DATA:
                        if(in_sub)
                            throw InvalidateAction("DATA not allowed in SUB/FUNCTION/DEF");

                        do {
                            if(tk(STRINGCONST).parse(scan))
                            {
                                //fprintf(stderr, "Data(%s)\n", scan.GetInfo().str.c_str());
                                global_data.push_back(scan.GetInfo().str);
                            }
                            else
                            {
                                //fprintf(stderr, "Data(%s)\n", "");
                                global_data.push_back(std::string());
                            }
                        } while( tk(COMMA).parse(scan) );
                        return true;

                    case SHARED:
                        if(!in_sub)
                            throw InvalidateAction("SHARED invalid outside SUB/FUNCTION");

                        do {
                            LambdaExpectToken(IDENTIFIER, );
                            std::string gotvar   = scan.GetInfo().str;
                            bool        is_array = false;
                            TypeP       type;
                            if(tk(OPENPARENS).parse(scan))
                            {
                                LambdaExpectToken(CLOSEPARENS, );
                                is_array = true;
                            }
                            if(tk(AS).parse(scan))
                            {
                                LambdaExpectSubRule((this->type), );
                                type = ctx.typ_result;
                            }
                            if(!gotvar.empty())
                                SubScopeImportSharedVariable(UCase(gotvar), type, is_array);
                        } while( tk(COMMA).parse(scan) );
                        return true;

                    case COMMON:
                        if(in_sub || in_block)
                            throw InvalidateAction("COMMON not allowed in control statements");

                        fprintf(stderr,
                            "Warning: COMMON not supported yet\n");

                        /* TODO: do */
                        return LambdaParseBody(
                               !tk(SHARED)
                             >> if_p(tk(SLASH)) [ tk(IDENTIFIER) >> SLASH ]
                             >> (tq(IDENTIFIER) % COMMA)
                             >> tq(CRLF)
                        );

                    /* data commands */
                    case DIM:
                    case REDIM:
                    case STATIC:
                    {
                        // Note: If the DIM parameters are not constants, we
                        //       should probably make the array automatically $DYNAMIC.

                        bool is_shared = false;
                        bool is_static = false;
                        if(tok != STATIC && tk(SHARED).parse(scan))
                        {
                            if(in_sub)
                                throw InvalidateAction("DIM SHARED invalid in SUB/FUNCTION");
                            // all vars in this definition will be shared
                            is_shared = true;
                        }
                        if(tok == STATIC)
                        {
                            if(!in_sub)
                                throw InvalidateAction("STATIC invalid outside SUB/FUNCTION");

                            is_static = true;
                        }
                        if(!is_shared)
                        {
                            /* If the SUB is declared STATIC, set is_static */
                            if(in_sub && in_sub->is_static)
                                is_static = true;
                        }
#ifdef BINDING_DEBUG
                        fprintf(stderr, "DIM: %s, %s, %s\n",
                            is_shared ? "Shared" : "Not shared",
                            is_static ? "Static" : "Not static",
                            in_sub ? (in_sub->is_static ? "Static sub" : "Non-static sub")
                                   : "Not sub");
#endif
                        /* DIM and REDIM.
                         * Note: REDIM can only be applied to $DYNAMIC arrays,
                         *             not scalars, utypes or $STATIC arrays.
                         *
                         * Note: If the array has not previously been declared,
                         *       declare it as $DYNAMIC if REDIM was used even if $STATIC was active.
                         */

                        do {
                            std::string varname;
                            LambdaExpectToken(IDENTIFIER, goto dim_fail_ident);

                            varname = scan.GetInfo().str;
                        dim_fail_ident:;
                            std::vector<std::pair<exprp,exprp> > dim_exprs;

                            if(tk(OPENPARENS).parse(scan))
                            {
                                /* FIXME/TODO: Should also support mere () for STATIC */
                                do {
                                    exprp first, second;

                                    LambdaExpectNumExpr((expr), goto dim_fail_second);

                                    second = std::move(ctx.exp_result);
                                dim_fail_second:;
                                    if(tk(TO).parse(scan))
                                    {
                                        first = std::move(second);
                                        LambdaExpectNumExpr((expr), goto dim_fail_second_again);
                                        second = std::move(ctx.exp_result);
                                dim_fail_second_again:;
                                    }
                                    if(!first)
                                        first.reset( new expr_number((long)option_base, TypeInt) );

                                    dim_exprs.emplace_back(std::move(first), std::move(second));
                                } while(tk(COMMA).parse(scan));
                                LambdaExpectToken(CLOSEPARENS, );
                            }

                            TypeP type;

                            if(tk(AS).parse(scan))
                            {
                                LambdaExpectSubRule((this->type), goto dim_fail_expr);
                                type = ctx.typ_result;
#ifdef BINDING_DEBUG
                                dprintf(dp_declare, "DIM AS type: %s\n",
                                    type ? type->Dump().c_str() : "NULL");
#endif
                            }

                            if(varname.empty()) goto dim_fail_expr;

                            if(!Do_Dim(lexloc, varname,
                                       std::move(dim_exprs),
                                       std::move(type),
                                       tok == REDIM,
                                       is_shared,
                                       is_static))
                            {
                                throw InvalidateAction("Invalid function call");
                            }
                        dim_fail_expr:;
                        } while(tk(COMMA).parse(scan));
                        return true;
                    }

                    case ERASE:
                    {
                        /* For static arrays, ERASE clears their contents to zero.
                         * For dynamic arrays, ERASE shrinks them to zero.
                         * For magic arrays, it delegates the work.
                         */
                        do {
                            LambdaExpectToken(IDENTIFIER, goto erase_ident_missing);
                          {
                            const std::string arrayname = scan.GetInfo().str;
                            VarContext c = FindVariableFromScope(arrayname, true);
                            if(!c.info)
                                throw InvalidateAction("Array not defined");

                            expr_var* v = new expr_var(lexloc, arrayname);
                            v->set_type(c.info->type);
                            v->set_basetype(c.info->type);
                            v->setwholearray();
                            SetStmt( stmtp(new stmt_erase(lexloc, exprp(v))) );
                          }
                        erase_ident_missing:;
                        } while(tk(COMMA).parse(scan));
                        return true;
                    }

                    case OPEN:
                    {
                        LambdaExpectStringExpr((expr),  return true );
                        exprp param1 = std::move(ctx.exp_result); /* 1:filename, 2:mode */

                        QBtoken mode = RANDOM;
                        int access=0;  // 1=read, 2=write
                        int locking=0; // 0=shared, 1=read, 2=write
                        bool access_defined=false;
                        bool locking_defined=false;
                    OpenLoop:
                        switch(scan.get())
                        {
                            case FOR:
                                switch(mode=scan.get())
                                {
                                    case RANDOM: access=3; goto OpenLoop;
                                    case OUTPUT: access=2; goto OpenLoop;
                                    case INPUT:  access=1; goto OpenLoop;
                                    case APPEND: access=2; goto OpenLoop;
                                    case BINARY: access=3; goto OpenLoop;
                                    default:
                                        throw InvalidateAction("Expected RANDOM, OUTPUT, INPUT, APPEND or BINARY");
                                }
                            case ACCESS:
                            {
                                if(access_defined) throw InvalidateAction("Duplicate ACCESS clause");
                                int allow_access=0;
                                iterator_t
                                it = scan.pos(); if(scan.get() == READ)  allow_access|=1; else scan.revert(it);
                                it = scan.pos(); if(scan.get() == WRITE) allow_access|=2; else scan.revert(it);
                                if(!allow_access) throw InvalidateAction("Expected READ or WRITE");
                                access &= allow_access;
                                if(!access) throw InvalidateAction("File must be OPENed for a purpose");
                                access_defined = true;
                                goto OpenLoop;
                            }
                            case SHARED:
                                if(locking_defined) throw InvalidateAction("Duplicate LOCK or SHARED clause");
                                locking_defined=true;
                                locking=0; goto OpenLoop;
                            case LOCK:
                            {
                                iterator_t it = scan.pos();
                                if(locking_defined) throw InvalidateAction("Duplicate LOCK or SHARED clause");
                                if(scan.get() == READ) locking|=1; else scan.revert(it);
                                it = scan.pos();
                                if(scan.get() == WRITE) locking|=2; else scan.revert(it);
                                if(!locking) throw InvalidateAction("Expected READ or WRITE");
                                locking_defined=true;
                                goto OpenLoop;
                            }
                            case AS:  break;  // it's syntax 1
                            case COMMA: break; // it's syntax 2
                            default:
                                throw InvalidateAction("Expected FOR, ACCESS, SHARED, LOCK, AS or ','");
                        }
                        bool oldsyntax = access == 0;

                        LambdaExpectSubRule((autofilenum), return true);
                        exprp filenum = std::move(ctx.exp_result);

                        exprp reclen;
                        exprp filename;
                        if(!oldsyntax) filename = std::move(param1);
                        if(oldsyntax)
                        {
                            if(param1->is_const())
                            {
                                /* Parse param1, get access mode (O/I/R/B/A) */
                                std::string m = (dynamic_cast<expr_string*>(&*param1))->get();
                                /**/ if(m=="O") { mode=OUTPUT; access=2; oldsyntax=false; }
                                else if(m=="I") { mode=INPUT;  access=1; oldsyntax=false; }
                                else if(m=="R") { mode=RANDOM; access=3; oldsyntax=false; }
                                else if(m=="B") { mode=BINARY; access=3; oldsyntax=false; }
                                else if(m=="A") { mode=APPEND; access=2; oldsyntax=false; }
                                else throw InvalidateAction("Expected string expression, or \"O\", \"I\", \"R\", \"B\" or \"A\"");
                            }

                            LambdaExpectToken(COMMA, return true);

                            LambdaExpectStringExpr((expr),  goto open_filename_expr_missing2);
                            filename = std::move(ctx.exp_result);
                        open_filename_expr_missing2:;

                            if(tk(COMMA).parse(scan))
                            {
                                LambdaExpectNumExpr((expr),  goto open_reclen_expr_missing);
                                reclen = std::move(ctx.exp_result);
                            open_reclen_expr_missing:;
                            }
                        }
                        else if(tk(LEN).parse(scan))
                        {
                            LambdaExpectToken(EQUALS, );
                            LambdaExpectNumExpr((expr),  goto open_len_expr_missing);
                            reclen = std::move(ctx.exp_result);
                        open_len_expr_missing:;
                        }

                        // We now have:
                        //  mode (token)
                        //    or param1 (string expr), which should be parsed
                        //  access (bitmask)
                        //  locking (bitmask)
                        //  filename (string expr)
                        //  reclen (null, or int expr)

                        if(oldsyntax)
                            SetStmt(stmtp(new stmt_fileopen(lexloc,
                                std::move(filename),std::move(filenum), std::move(param1), access, locking, std::move(reclen))));
                        else
                            SetStmt(stmtp(new stmt_fileopen(lexloc,
                                std::move(filename),std::move(filenum), mode,  access, locking, std::move(reclen))));
                        return true;
                    }

                    case LPRINT:
                    case PRINT:
                    case WRITE:
                    {
                        std::unique_ptr<stmt_print> print_stmt( new stmt_print(lexloc) );
                        if(tok == LPRINT) print_stmt->SetLPRINT();

                        if(tok != LPRINT && tk(HASHSYM).parse(scan))
                        {
                            // A file number is allowed with PRINT and WRITE
                            ctx.exp_result.reset( nullptr );
                            LambdaExpectSubRule((autofilenum), );
                            LambdaExpectToken(COMMA, );

                            print_stmt->SetFile( std::move(ctx.exp_result) );
                        }

                        bool was_using = false;
                        if(tok != WRITE && tk(USING).parse(scan))
                        {
                            ctx.exp_result.reset( nullptr );
                            LambdaExpectStringExpr((expr), break);
                            (tk(SEMICOLON) | tk(COMMA)).expected().parse(scan);
                            was_using = true;

                            print_stmt->SetFormat( std::move(ctx.exp_result) );
                        }

                        bool need_delim = false;
                        bool ends_bare  = true;
                        for(bool first=true; ; first=false)
                        {
                            if(tk(COMMA).parse(scan))
                            {
                                if(!was_using && tok != WRITE) // not in USING/WRITE
                                {
                                    print_stmt->AppendTab();
                                }
                                need_delim = false;
                                ends_bare  = false;
                                continue;
                            }
                            if(tk(SEMICOLON).parse(scan))
                            {
                                need_delim = false;
                                ends_bare  = false;
                                continue;
                            }
                            /* Note: need_delim not actually used here.
                               PRINT "a"b$"c" is allowed by GW-BASIC, and
                                this backward compatibility doesn't hurt.. much.
                             */
                            if(tok != WRITE)
                            {
                                if(tk(SPC).parse(scan))
                                {
                                    ctx.exp_result.reset( nullptr );
                                    LambdaExpectToken(OPENPARENS, );
                                    LambdaExpectNumExpr((expr), );
                                    LambdaExpectToken(CLOSEPARENS, );

                                    if(ctx.exp_result)
                                        print_stmt->AppendSPCfunc( std::move(ctx.exp_result) );

                                    need_delim = true;
                                    ends_bare  = true;
                                    continue;
                                }
                                if(tk(TAB).parse(scan))
                                {
                                    ctx.exp_result.reset( nullptr );
                                    LambdaExpectToken(OPENPARENS, );
                                    LambdaExpectNumExpr((expr),   );
                                    LambdaExpectToken(CLOSEPARENS, );

                                    if(ctx.exp_result)
                                        print_stmt->AppendTABfunc( std::move(ctx.exp_result) );

                                    need_delim = true;
                                    ends_bare  = true;
                                    continue;
                                }
                            }
                            if(tok == WRITE && !first)
                            {
                                print_stmt->AppendCommaCh();
                            }

                            // USING requires at least one arg.
                            if(! ( (was_using && first) ? expr.expected().parse(scan)
                                                        : expr.parse(scan) ) )
                            {
                                break;
                            }

                            exprp param = std::move(ctx.exp_result);

                            if(tok == WRITE && param->is_string())
                            {
                                print_stmt->AppendQuoteCh();
                                print_stmt->AppendExpr( std::move(param) );
                                print_stmt->AppendQuoteCh();
                            }
                            else
                            {
                                if(tok != WRITE && !was_using && param->is_num())
                                {
                                    print_stmt->AppendSpace();
                                    print_stmt->AppendExpr( std::move(param) );
                                    print_stmt->AppendSpace();
                                }
                                else
                                {
                                    /* Smart translations for several nn$ builtins */
                                    if(auto s = dynamic_cast<expr_builtin_string*> (param.get()))
                                    {
                                        if(s->GetToken() == STR_S && !was_using)
                                        {
                                            /* STR$(num) prints a space and num */
                                            print_stmt->AppendSpace();
                                            param = std::move(s->GetParam(0));
                                        }
                                        else if(s->GetToken() == LTRIM_S && !was_using)
                                        {
                                            auto s2 = dynamic_cast<expr_builtin_string*> (s->GetParam(0).get());
                                            if(s2->GetToken() == STR_S)
                                            {
                                                /* LTRIM$(STR$(num)) prints just the num */
                                                param = std::move(s2->GetParam(0));
                                            }
                                        }
                                        else if(s->GetToken() == CHR_S)
                                        {
                                            /* CHR$(numconst) can be translated to a string. */
                                            if(auto n = dynamic_cast<expr_number*> (s->GetParam(0).get()))
                                            {
                                                long long int i = n->get_i();
                                                if(i < 0 || i > 255)
                                                    yyerror("Error: Invalid parameter for CHR$", scan.GetLoc());
                                                param.reset( new expr_string( std::string( std::size_t(1), char(i) ) ) );
                                            }
                                        }
                                    }
                                    print_stmt->AppendExpr( std::move(param) );
                                }
                            }

                            need_delim = true;
                            ends_bare  = true;
                        }
                        if(ends_bare || tok==WRITE)
                            print_stmt->AppendNewline();

                        SetStmt( std::move(print_stmt) );
                        return true;
                    }

                    case INPUT:
                    case LINE_INPUT:
                    {
                        std::unique_ptr<stmt_input> input_stmt(new stmt_input(lexloc));

                        bool is_file = tk(HASHSYM).parse(scan);
                        bool is_line = tok == LINE_INPUT;

                        if(is_line) input_stmt->SetLINEINPUT();
                        if(is_file)
                        {
                            ctx.exp_result.reset( nullptr );
                            LambdaExpectSubRule((autofilenum), );
                            LambdaExpectToken(COMMA, );
                            input_stmt->SetFile( std::move(ctx.exp_result) );
                        }
                        else
                        {
                            if(tk(SEMICOLON).parse(scan))
                            {
                                input_stmt->SetKeepCursor();
                            }
                            if(tk(STRINGCONST).parse(scan))
                            {
                                DoString(scan);
                                input_stmt->SetPrompt( std::move(ctx.exp_result) );
                                (tk(SEMICOLON) | tk(COMMA)).expected().parse(scan);
                            }
                        }
                        if(is_line)
                        {
                            ctx.exp_result.reset( nullptr );
                            LambdaExpectStringExpr((var_target), );
                            if(ctx.exp_result)
                                input_stmt->AppendTarget( std::move(ctx.exp_result) );
                        }
                        else
                        {
                            do {
                                ctx.exp_result.reset( nullptr );
                                LambdaExpectSubRule((var_target), );
                                if(ctx.exp_result)
                                    input_stmt->AppendTarget( std::move(ctx.exp_result) );
                            } while(tk(COMMA).parse(scan));
                        }
                        SetStmt( std::move(input_stmt) );
                        return true;
                    }

                    case ON:
                    {
                        LambdaExpectNumExpr((expr), return true);

                        std::unique_ptr<stmt_select> select_stmt(new stmt_select(lexloc, std::move(ctx.exp_result)));

                        bool is_goto = false;
                        if(tk(GOSUB).parse(scan)) is_goto=false;
                        else if(tk(GOTO).parse(scan)) is_goto=true;
                        else throw InvalidateAction("Expected: GOTO or GOSUB");

                        for(int value=1; ; ++value)
                        {
                            lexloc = scan.GetLoc();

                            ctx.exp_result.reset( nullptr );
                            LambdaExpectSubRule((anylabeltarget), );
                            if(ctx.exp_result)
                            {
                                exprp tg = std::move(ctx.exp_result);
                                exprp ex ( new expr_caseis( EQUALS, exprp(new expr_number(value))) );

                                select_stmt->cases.emplace_back(std::move(ex), stmtp(new stmt_labelref
                                    (lexloc, is_goto?GOTO:GOSUB, std::move(tg))));
                            }
                            if(!tk(COMMA).parse(scan)) break;
                        }

                        /* TODO: according to QuickBasic, if expression is <0 or >255,
                         * illegal function call should be produced. */
                        SetStmt( std::move(select_stmt) );
                        return true;
                    }

                    case REM:
                    {
                        // REM is allows as a simple statement.
                        // This makes "IF 0 THEN REM xx" work.
                        return true;
                    }

                    case qb64DEFINE:
                    {
                        std::string as_def   = scan.GetInfo().numtype;
                        std::string charlist = scan.GetInfo().str;

                        if(as_def.empty())
                        {
                            ctx.typ_result = ( nullptr );
                            LambdaExpectToken(AS,       );
                            LambdaExpectSubRule((type), );
                            if(ctx.typ_result)
                            {
                                as_def = TypeIntoTypeSuffix(*ctx.typ_result);
                                if(as_def.empty())
                                    throw InvalidateAction("Type " + ctx.typ_result->Dump() + " cannot be used in _DEFINE");
                            }
                        }
                        SetDefVarType(charlist, as_def);
                        return true;
                    }
                }
                return false;
            }
        )
    |   /* Assign statements without "LET" */
        LambdaParser([this] (scanner_t& scan) -> bool
        {
            if(!(var_target).parse(scan)) return false;
            LexLocation lexloc = scan.GetLoc();
            exprp target = std::move(ctx.exp_result);
            if(!tk(EQUALS).parse(scan)) return false;
            LambdaExpectSubRule((expr), return true);
            exprp source = std::move(ctx.exp_result);

            /* TODO: Check type more carefully */
            TypeP tt = target->GetType(), st = source->GetType();
            if(tt && st
            && ((tt->is_array()
              != st->is_array())
             || (tt->is_string()
              && st->is_num())
             || (tt->is_num()
              && st->is_string())))
            {
                throw InvalidateAction("Type mismatch");
            }
            SetStmt(stmtp(new stmt_let(lexloc, std::move(target), std::move(source))));
            return true;
        })
    |   /* SUB call without "CALL" */
        LambdaParser([this] (scanner_t& scan) -> bool
        {
            LexLocation lexloc = scan.GetLoc();
            if(!tk(IDENTIFIER).parse(scan)) return false;

            const std::string subname = scan.GetInfo().str;
            // TODO: check namespace

            std::vector<exprp> params;
            bool expected = false;
            do {
                ctx.exp_result.reset( nullptr );

                if(expected)
                    func_param.expected().parse(scan);
                else
                    if(!func_param.parse(scan)) break;

                if(ctx.exp_result)
                    params.emplace_back( std::move(ctx.exp_result) );

                expected = true;
            } while(tk(COMMA).parse(scan));
            SetStmt(stmtp(new stmt_subcall(lexloc, subname, std::move(params))));
            return true;
        });

    /**** LVAL */
    rule_t var_target = LambdaParser([this] (scanner_t& scan) -> bool
        {
            /* It cannot be a CONST (parse error)
             * It cannot be a FUNCTIONNAME (fail rule without error)
             *   However, it can be the name of current function,
             *   or the name of current DEF.
             */
            LexLocation lexloc = scan.GetLoc();
            bool res = this->ParseVarDef(scan,
                // handle_const
                [&] (expr_var&&, exprp&& ) -> bool
                {
                    throw InvalidateAction("Invalid use of a CONST");
                },
                // handle_sub
                [&] (expr_var&& var, const SubRoutine& s, bool params) -> bool
                {
                    if(!s.returntype)
                    {
#ifdef BINDING_DEBUG
                        //dprintf(dp_declare, "Sub call?");
#endif
                        return false; // It may be a sub call instead!
                    }

                    const std::string name = var.getname();
                    const std::string basename = UCase(StripTypeSuffix(name));

                    if(params || basename != GetCurrentSubBaseName())
                    {
                        throw InvalidateAction("Invalid function call syntax");
                    }

                    SetExpr( exprp(new expr_returnvalueholder(lexloc, name)) );
#ifdef BINDING_DEBUG
                    //fprintf(stderr, "retval \n");
#endif
                    return true;
                }
            );
            if(!res) return false;


            return true;
        }, "lvalue");

    rule_t autofilenum = LambdaParser([this] (scanner_t& scan) -> bool
        {
            if(tk(HASHSYM).parse(scan)) { /* good, though omission is not an error */ }
            LambdaExpectNumExpr(expr, return false);
            return true;
        }, "file number");

    rule_t expr = LambdaParser([this] (scanner_t& scan) -> bool
        {
            struct ExprPrioItem
            {
                exprp    exp;
                QBtoken  tok;
                unsigned prio;
            };

            iterator_t it = scan.pos();

            /* PRIORITY LEVELS:
             *
             *  1 IMP
             *  2 EQV
             *  3 XOR
             *  4 OR
             *  5 AND
             *  6 unary NOT
             *  7 comparisons
             *  8 + -
             *  9 MOD  _SHL _SHR
             * 10 \
             * 11 * /
             * 12 unary -  (using token "IF" here)
             * 13 ^
             */
            unsigned prio = 0;
            std::vector<ExprPrioItem> priostack;

            #define BINARY_OP(level, op) \
                while(prio >= level) PRIO_RETURN(); \
                PRIO_ENTER_OP(level, op)

            auto PRIO_ENTER_OP = [&prio,&priostack] (int level, QBtoken op)
            {
                /*
                fprintf(stderr, "prio=%u, enter %u(%s)\n",
                    (prio), level,
                    GetTokenName(op).c_str());
                */
                ExprPrioItem tmp;
                tmp.prio = (prio);
                tmp.tok = op;
                tmp.exp = std::move(ctx.exp_result);
                prio = level;
                priostack.emplace_back( std::move(tmp) );
            };

            auto PRIO_RETURN = [&prio,&priostack] ()
            {
                if(priostack.empty())
                {
                    fprintf(stderr, "error: PRIO_RETURN called when empty (prio=%u)\n",
                        (prio));
                }
                ExprPrioItem& top = priostack.back();
                switch(top.tok)
                {
                    case NOT:
                        if(!ctx.exp_result->is_num()) goto TypeMismatch;
                        SetExpr(exprp(new expr_unary(NOT, std::move(ctx.exp_result))));
                        break;
                    case IF: // unary minus
                        if(!ctx.exp_result->is_num()) goto TypeMismatch;
                        SetExpr(exprp(new expr_unary(MINUS, std::move(ctx.exp_result))));
                        break;
                    default:
#ifdef PARSER_DEBUG
                        fprintf(stderr, "Creating %s <%s>, <%s>\n",
                            GetTokenName(top.tok).c_str(),
                            typenameof(*top.exp).c_str(),
                            typenameof(*ctx.exp_result).c_str()
                        );
#endif
                        bool string_ok = false;
                        switch(top.tok)
                        {
                            case PLUS: case LESSTHAN: case GREATERTHAN: case EQUALS:
                            case NOTEQUAL: case LESSOREQUAL: case GREATEROREQUAL:
                                string_ok = true;
                            default: break;
                        }
                        if(string_ok)
                        {
                            if(ctx.exp_result->is_num()
                            != top.exp->is_num()) goto TypeMismatch;
                        }
                        else
                        {
                            if(top.exp->is_num().is_false()) goto TypeMismatch;
                            if(ctx.exp_result->is_num().is_false()) goto TypeMismatch;
                        }
                        switch(top.tok)
                        {
                            case IMP: // a IMP b = b OR NOT a
                              { exprp p = std::move(ctx.exp_result);
                                SetExpr(exprp(new expr_unary(NOT, std::move(top.exp))));
                                SetExpr(exprp(new expr_binary(OR, std::move(ctx.exp_result), std::move(p))));
                                break; }
                            case EQV: // a EQV b = NOT(a XOR b)
                              { SetExpr(exprp(new expr_binary(XOR, std::move(top.exp), std::move(ctx.exp_result))));
                                SetExpr(exprp(new expr_unary(NOT, std::move(ctx.exp_result))));
                                break; }
                            default:
                                SetExpr(exprp(new expr_binary(top.tok, std::move(top.exp), std::move(ctx.exp_result))));
                        }
                        break;
                }
                prio = top.prio;
                priostack.pop_back();
                return;
            TypeMismatch:
                throw InvalidateAction("Type mismatch\n");
            };

            bool first = true;
        Continues:
            if(!scan.eof()) /* check unary */
            {
                iterator_t unary_it = scan.pos();
                switch(scan.get())
                {
                    case NOT:   PRIO_ENTER_OP(6, NOT); goto Continues;
                    case MINUS: PRIO_ENTER_OP(6, IF); goto Continues;
                    default: scan.revert(unary_it);
                }
            }
            if(first)
                { first=false; if(!(expr_meat).parse(scan)) goto Fail; }
            else
                { LambdaExpectSubRule((expr_meat), goto Fail); }
        Continued:
            if(!scan.eof()) /* check binary */
            {
                iterator_t opr_it = scan.pos();
                auto opr = scan.get();
                switch(opr)
                {
                    case IMP: BINARY_OP(1, opr); goto Continues;
                    case EQV: BINARY_OP(2, opr); goto Continues;
                    case XOR: BINARY_OP(3, opr); goto Continues;
                    case OR:  BINARY_OP(4, opr); goto Continues;
                    case AND: BINARY_OP(5, opr); goto Continues;
                    // unary NOT
                    case NOTEQUAL: case LESSOREQUAL: case GREATEROREQUAL:
                    case GREATERTHAN: case LESSTHAN: case EQUALS:
                              BINARY_OP(7, opr); goto Continues;
                    case PLUS: case MINUS:
                              BINARY_OP(8, opr); goto Continues;
                    case MOD: case qbcSHL: case qbcSHR:
                              BINARY_OP(9, opr); goto Continues;
                    case BKSLASH: BINARY_OP(10, opr); goto Continues;
                    case ASTERISK: case SLASH:
                              BINARY_OP(11, opr); goto Continues;
                    // unary -
                    case CIRCUMFLEX: BINARY_OP(13, opr); goto Continues;

                    case NUMBER:
                        // Handle x = x-3 where -3 is a NUMBER token
                        if(scan.GetInfo().had_sign)
                        {
                            exprp tmp = std::move(ctx.exp_result);
                            DoNumber(scan);
                            exprp e = std::move(ctx.exp_result);
                            SetExpr( std::move(tmp) );
                            BINARY_OP(8, PLUS);
                            SetExpr( std::move(e) );
                            goto Continued;
                        }
                        break;

                    default: break;
                }
                scan.revert(opr_it);
            }
            while(!priostack.empty()) { PRIO_RETURN(); }
            return true;
        Fail:
            scan.revert(it);
            return false;
        }, "expression");

    rule_t func_param = LambdaParser([this] (scanner_t& scan) -> bool
        {
            iterator_t it = scan.pos();
            LexLocation lexloc = scan.GetLoc();
            if(tk(IDENTIFIER).parse(scan))
            {
                const std::string arrayname = scan.GetInfo().str;
                if(tk(OPENPARENS).parse(scan)
                && tk(CLOSEPARENS).parse(scan))
                {
                    // Magic array!
                    VarContext c = FindVariableFromScope(arrayname, true);
                    if(!c.info)
                        throw InvalidateAction("Array not defined");
                    expr_var* v = new expr_var(lexloc, arrayname);
                    v->set_type(c.info->type);
                    v->set_basetype(c.info->type);
                    v->setwholearray();
                    SetExpr( exprp(v) );
                    return true;
                }
            }
            scan.revert(it);
            return (expr).parse(scan);
        }, "SUB/FUNCTION parameter");

    rule_t longblockstatement = LambdaTokParser([this] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                case IF:
                {
                    const block_setting<bool> block_marker(in_block, true);

                    LambdaExpectNumExpr((expr), break);

                    std::unique_ptr<stmt_if> if_stmt(new stmt_if(scan.GetLoc(), std::move(ctx.exp_result)));

                    /* Use tk instead of tq, to enable shortif without problems */
                    if(!tk(THEN).parse(scan)) break;
                    if(!tk(CRLF).parse(scan)) break;

                    // Now that we are definitely in longstatement context (we saw a CRLF),
                    // we can begin using expected().

                    stmtp prev = std::move(ctx.stm_result); // Save the end of statements chain

                    std::list<std::pair<exprp, stmtp>> elseifp;
                    stmtp thenp, elsep;

                    SetStmt(nullptr); // begin a new chain of statements
                    while(!scan.eof() && longstatement.parse(scan)) {}
                    thenp = std::move(ctx.stm_result);

                    while(tk(ELSEIF).parse(scan))
                    {
                        LambdaExpectNumExpr((expr), );
                        exprp ex = std::move(ctx.exp_result);

                        LambdaExpectToken(THEN, );

                        SetStmt(nullptr); // begin a new chain of statements
                        while(!scan.eof() && longstatement.parse(scan)) {}

                        if(ex)
                            elseifp.emplace_back( std::move(ex), std::move(ctx.stm_result) );
                    }
                    if(tk(ELSE).parse(scan))
                    {
                        SetStmt(nullptr); // begin a new chain of statements
                        while(!scan.eof() && longstatement.parse(scan)) {}
                        elsep = std::move(ctx.stm_result);
                    }

                    if_stmt->AddBody( std::move(thenp), std::move(elseifp), std::move(elsep) );

                    SetStmt( std::move(if_stmt), std::move(prev) );
                    return tq(ENDIF).parse(scan);
                }

                case SELECT:
                    return parse_select(scan, false);

                case FOR:
                    return parse_for(scan, false);

                 /* do while ... loop
                    do until ... loop
                    do ... loop
                    do ... loop while
                    do ... loop until */
                case DO:
                case WHILE:
                    return parse_doloop(scan, false, tok);

                /* Note: SUBs cannot exist in SUBs. */
                case SUB:
                case FUNCTION:
                case DEF:
                case DECLARE:
                {
                    bool is_declare = false;
                    if(tok == DECLARE)
                    {
                        is_declare = true;
                        iterator_t it = scan.pos();
                        tok = tk(SUB).parse(scan) ? SUB : FUNCTION;
                        scan.revert(it);
                        if(!(tk(SUB) | tk(FUNCTION)).expected().parse(scan)) return true;
                    }

#ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Declare sub/func/def...\n");
#endif
                    if(in_sub || in_block)
                        throw InvalidateAction("SUB/FUNCTION/DEF/DECLARE not allowed in control statements");

                    QBtoken endtoken = tok==SUB ? ENDSUB : ENDFUNCTION;

                    LexLocation identifier_lexlocation = scan.GetLoc();

                    if(tok == DEF)
                    {
                        LambdaExpectToken(FNIDENTIFIER, return true);
                        endtoken = ENDDEF;
                    }
                    else
                    {
                        LambdaExpectToken(IDENTIFIER, return true);
                    }
                    std::string subname = scan.GetInfo().str;
                    std::vector<SubRoutine::ParamSpec> params;
                    bool is_cdecl = false;
                    bool is_static = false;
                    std::string alias;

                    if((tok == SUB) && HasTypeSuffix(subname))
                        throw InvalidateAction("Invalid SUB name");

                    if(  (is_declare ? tq(OPENPARENS).parse(scan)
                                      : tk(OPENPARENS).parse(scan) )
                    && !tk(CLOSEPARENS).parse(scan)) /* optional, but mandatory for DECLAREs */
                    {
                        do {
                            SubRoutine::ParamSpec param;

                            if(tk(BYVAL).parse(scan)) { param.SetByval(); }
                            else if(tk(SEG).parse(scan)) { param.SetSeg(); }

                            LambdaExpectToken(IDENTIFIER, goto sub_param_identifier_missing);

                            param.name     = scan.GetInfo().str;
                        sub_param_identifier_missing:;
                            param.is_array = false;
                            param.was_modified = true; // assume worst
                            if(tk(OPENPARENS).parse(scan))
                            {
                                LambdaExpectToken(CLOSEPARENS, );
                                param.is_array = true;
                            }

                            LexLocation lexloc = scan.GetLoc();
                            param.type = CreateTypeTemplate(param.name);
                            if(tk(AS).parse(scan))
                            {
                                if(HasTypeSuffix(param.name))
                                    yyerror("Invalid AS together with type suffix");
                                if(tk(ANY).parse(scan))
                                {
                                    if(!is_declare || !param.is_array)
                                        dprintf(dp_warnings, "%s: Warning: AS ANY is only valid with arrays in DECLARE\n",
                                                lexloc.stringify().c_str());
                                    param.type = nullptr;
                                }
                                else
                                {
                                    ctx.typ_result = ( nullptr );
                                    LambdaExpectSubRule((type), );
                                    param.type = ctx.typ_result;
                                }
                            }
                            if(param.type && param.is_array)
                            {
                                param.type->arraytype = TypeInfo::ParamArray;
                            }
                            if(!param.name.empty())
                                params.emplace_back( std::move(param) );
                        } while(tk(COMMA).parse(scan));
                        LambdaExpectToken(CLOSEPARENS, );
                    }

                    if(!is_declare && tk(STATIC).parse(scan)) // optional
                    {
                        if(tok == DEF)
                            yyerror("DEFs cannot be 'STATIC'");
                        else
                            is_static = true;
                    }

                    if(is_declare)
                    {
                        if(tk(CDECL).parse(scan)) { is_cdecl = true;  }
                        if(tk(ALIAS).parse(scan))
                        {
                            LambdaExpectToken(STRINGCONST, goto sub_alias_missing);
                            alias = scan.GetInfo().str;
                        sub_alias_missing:;
                        }
                    }

                    iterator_t it = scan.pos();

                    if(tok == DEF)
                    {
                        if(!tk(CRLF).parse(scan)) return false; // no harm, try an oneliner format later.
                    }
                    else
                    {
                        if(tq(CRLF).parse(scan)) // Expect CRLF here.
                        {
                            if(is_declare)
                            {
                                // Revert the CRLF so that subrule "separator"
                                // will match properly.
                                scan.revert(it);
                            }
                        }
                    }

                    TypeP returntype;
                    if(tok == FUNCTION || tok == DEF)
                        returntype = CreateTypeTemplate(subname);

                    auto sub = DefineSubFunc(is_declare, subname, returntype, params, is_static, is_cdecl, alias);
                    if(!sub)
                        throw InvalidateAction("Duplicate definition of "+subname);

                    if(is_declare)
                        (const_cast<SubRoutine*>(sub))->declared_at = identifier_lexlocation;
                    else
                        (const_cast<SubRoutine*>(sub))->defined_at = identifier_lexlocation;

                    if(!is_declare)
                    {
                        /* For DEFs, ALL variables are SHARED even if not declared SHARED */
                        EnterSubFuncScope(subname, tok == DEF);

                        const block_setting<const SubRoutine*> block_marker(in_sub, sub);

                        stmtp prev = std::move(ctx.stm_result); // Save the end of statements chain

                        SetStmt(nullptr); // begin a new chain of statements
                        while(!scan.eof() && longstatement.parse(scan)) {}

                        ExitSubFuncScope(subname, std::move(ctx.stm_result));

                        if(!tk(endtoken).parse(scan))
                        {
                            // Check if there is another type of end-token instead.
                            if((tk(ENDSUB) | tk(ENDFUNCTION) | tk(ENDDEF)).expected().parse(scan))
                            {
                                dprintf(dp_warnings, "Warning: Expected: %s\n", GetTokenName(endtoken).c_str());
                            }
                        }

                        // Reset the statement context from earlier.
                        // Note that we did not add any new statements to the chain.
                        SetStmt( std::move(prev), nullptr );
                    }
                    return true;
                }

                /* Note: TYPE cannot exist in SUBs.
                 * However, Quickbasic allows them inside control structures.
                 * But they are unconditional.
                 */
                case TYPE:
                {
                    if(in_sub)
                        throw InvalidateAction("TYPE not allowed in SUB/FUNCTION/DEF");

                    LambdaExpectToken(IDENTIFIER, return true);
                    const std::string utypename = scan.GetInfo().str;

                    if(HasTypeSuffix(utypename))
                        throw InvalidateAction("TYPE name cannot have a type suffix");

                    UtypeP utype = std::make_shared<UserType>();

                    LambdaExpectToken(CRLF, break);
                    while(!scan.eof())
                    {
                        // empty line is ok in typedef
                        if(tk(CRLF).parse(scan)) continue;

                        if(!tk(IDENTIFIER).parse(scan)) break;

                        while(!scan.eof())
                        {
                            const std::string membername = scan.GetInfo().str;
                            if(HasTypeSuffix(membername))
                                throw InvalidateAction("TYPE member cannot have a type suffix");

                            LambdaExpectToken(AS, break);
                            LambdaExpectSubRule((type), break);

                            if(!utype->DeclareMember(membername, ctx.typ_result))
                                throw InvalidateAction("Duplicate definition of "+membername);

                            if(!tk(COMMA).parse(scan)) break;

                            LambdaExpectToken(IDENTIFIER, break);
                        }
                        LambdaExpectToken(CRLF, break);
                    }
                    LambdaExpectToken(ENDTYPE, return true);

                    if(!DefineUserType(utypename, utype))
                    {
                        throw InvalidateAction("Duplicate definition of "+utypename);
                    }

                    return true;
                }

                default: break;
            }
            return false;
        });

    rule_t longstatement =
        /* A label (that does not end in a newline/terminator) */
        LambdaTokParser(
        [this] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                /* a label statement */
                /* must be checked before simplestatement */
                case NUMBER:
                    DoNumericLabelDef(scan); return true;
                case IDENTIFIER:
                    // Labels are only be recognized in the beginning of line (long statements).
                    // This still conflicts with parameterless SUB calls though.
                    if(!tk(COLONCHAR).parse(scan)) break;
                    DoIdLabelDef(scan); return true;

                // filler
                case COLONCHAR: return true;
                case CRLF: return true;

                case metaDYNAMIC: SetMetaDynamic(); return true;
                case metaSTATIC:  SetMetaStatic(); return true;
                default:break;
            }
            return false;
        })
        /* Or, a statement (that must end in a newline/terminator) */
    | ( longblockstatement
      | simplestatement
      | shortif
     ) >> LambdaParser([this] (scanner_t& scan) -> bool
          {
              if(scan.eof())
              {
                  yyerror("File must end in a newline");
                  return true;
              }
              if(tk(COLONCHAR).parse(scan))
              {
                  // If the statement ends in a ':', only short-statements are accepted until a linefeed.
                  bool stmt_ok = true;
                  while(!scan.eof())
                  {
                      if(tk(COLONCHAR).parse(scan))  { stmt_ok = true; continue; }
                      if(!stmt_ok) break;
                      if(shortstatement.parse(scan)) { stmt_ok = false; continue; }
                      if(shortif.parse(scan))        { break; }
                      break;
                  }
                  // To make short-form NEXTs work, we allow the omission of CRLF if a NEXT is coming instead.
                  // Actually, this goes also for LOOP and WEND. None of these need to be in the beginning
                  // of the line even if the block inside was a set of longstatements.
                  auto p = scan.pos();
                  if((tk(NEXT) | tk(LOOP) | tk(WEND)).parse(scan)) { scan.revert(p); return true; }
              }
              // longstatements must always be terminated with a CRLF.
              LambdaExpectToken(CRLF, return true);
              return true;
          });

    rule_t shortstatement =
        /* statements that can be included between online THEN and ELSE */
        /* Does not include shortif. Does not include trailing COLONCHARs. */
        simplestatement
    |   LambdaTokParser(
        [this] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                case FOR:
                    return parse_for(scan, true);

                case SELECT:
                    return parse_select(scan, true);

                 /* do while ... loop
                    do until ... loop
                    do ... loop
                    do ... loop while
                    do ... loop until */
                case DO:
                case WHILE:
                    return parse_doloop(scan, true, tok);

                default: break;
            }
            return false;
        }, "statement");


    rule_t expr_meat =
        OPENPARENS >> expr >> CLOSEPARENS

    |   LambdaTokParser(
        [this] (QBtoken tok, scanner_t& scan) -> bool
        {
            const char* pattern = 0;
            /* LIBFUNCTIONS */
            switch(tok)
            {
                default:break;
                got_expr_builtin_numeric:
                    return (ParseLibFunction<expr_builtin_numeric>)((*this), scan, tok, pattern);
                got_expr_builtin_string:
                    return (ParseLibFunction<expr_builtin_string>)((*this), scan, tok, pattern);

                #define tokcase(s,p,code) case s: Do(p)
                #define Do(p) { pattern=p; goto got_expr_builtin_numeric; }

                     // numeric -> numeric functions
#include "libdata-expr_numeric.inc"

                // These 7 functions are handled separately in compile.cc,
                // and for convenience they are also removed from libdata-expr_numeric.inc .
                case COS: case SIN: case ATN: case TAN:
                case ABS: case LOG: case EXP: case SQR:
                case qbcASIN: case qbcACOS:
                case qbcASINH: case qbcACOSH: case qbcATANH:
                case qbcSINH: case qbcCOSH: case qbcTANH:
                case INT: case FIX:
                case CINT: case CLNG: case CSNG: case CDBL:
                    Do("(N)"); goto got_expr_builtin_numeric;
                case qbcATAN2:
                    Do("(N,N)"); goto got_expr_builtin_numeric;

                #undef Do
                #define Do(p) { pattern=p; goto got_expr_builtin_string; }

#include "libdata-expr_string.inc"

                #undef Do
                #undef tokcase

                case LBOUND:
                case UBOUND:
                {
                    LexLocation lexloc = scan.GetLoc();
                    LambdaExpectToken('(',        return true);
                    LambdaExpectToken(IDENTIFIER, return true);
                    const std::string arrayname = scan.GetInfo().str;
                    VarContext c = FindVariableFromScope(arrayname, true);
                    if(!c.info)
                        throw InvalidateAction("Array not defined");

                    expr_var* v = new expr_var(lexloc, arrayname);
                    v->set_type(c.info->type);
                    v->setwholearray();
                    exprp p1 ( v );

                    ctx.exp_result.reset( nullptr );
                    if(tk(COMMA).parse(scan))
                    {
                        LambdaExpectNumExpr((expr), return true);
                    }
                    LambdaExpectToken(CLOSEPARENS, );
                    SetExpr(exprp(new expr_lubound(lexloc, tok, std::move(p1), std::move(ctx.exp_result))));
                    return true;
                }

                case NUMBER:      { DoNumber(scan); return true; }
                case STRINGCONST: { DoString(scan); return true; }
            }
            return false;
        }, "a constant or a built-in function")

    |   /**** RVAL */
        /* Var or const or function call.
         * Collect it into an expr_var first. Then decide if to convert
         * it to expr_userfun.
         *
         * Function calls look just like array variables.
         * Only context will tell whether it should be converted into a FUNCTION.
         */
        LambdaParser([this] (scanner_t& scan) -> bool
        {
            return (this->ParseVarDef)(scan,
                // handle_const
                [&] (expr_var&&, exprp&& c) -> bool
                {
                    SetExpr( std::move(c) );
                    return true;
                },
                // handle_sub
                [&] (expr_var&& var, const SubRoutine& s, bool) -> bool
                {
                    if(!s.returntype)
                    {
                        throw InvalidateAction("Invalid use of a SUB");
                    }
                    SetExpr( exprp( std::move(var).convert_to_functioncall(s.returntype) ) );
                    return true;
                }
            );
        }
        , "rvalue");

    std::function<bool(scanner_t&, bool)> parse_for = [this](scanner_t& scan, bool short_statement) -> bool
    {
        const block_setting<bool> block_marker(in_block, true);

        if(!(this->ParseVarDef)(scan,
            // handle_const
            [&] (expr_var&&, exprp&& ) -> bool
            {
                throw InvalidateAction("Invalid use of a CONST");
            },
            // handle_sub
            [&] (expr_var&&, const SubRoutine&, bool) -> bool
            {
                throw InvalidateAction("Invalid use of a SUB name");
            }
        )) return false;

        LambdaExpectToken(EQUALS, return true);

        exprp var  = std::move(ctx.exp_result);

        const expr_var& vardef = *dynamic_cast<const expr_var*>(var.get());

        const std::string& loopvar = vardef.getname();
        TypeP vartype = vardef.GetType();
        if(!vartype) throw InvalidateAction("Duplicate definition");
        if(!vartype->is_num()) throw InvalidateAction("Type mismatch");
        if(vardef.isarray()) throw InvalidateAction("FOR loop variable cannot be an array member");

        LambdaExpectNumExpr((expr),  return true );
        exprp begin = std::move(ctx.exp_result);

        LambdaExpectToken(TO, return true);
        LambdaExpectNumExpr((expr),  return true );
        std::unique_ptr<stmt_for> for_stmt(new stmt_for(scan.GetLoc(), std::move(var),
                                                        std::move(begin), std::move(ctx.exp_result)) );

        if(tk(STEP).parse(scan))
        {
            LambdaExpectNumExpr((expr),  return true );
            for_stmt->step = std::move(ctx.exp_result);
        }
        else
            for_stmt->step.reset( new expr_number(1, TypeByte) );

        stmtp prev = std::move(ctx.stm_result); // Save the end of statements chain

        if(short_statement)
        {
            LambdaExpectSubRule((colonchars_shortstatements_and_colonchar), return false);
        }
        else
        {
            LambdaExpectSubRule((separator_statements_and_separator), return false);
        }

        for_stmt->bodyp = std::move(ctx.stm_result);

        LambdaExpectToken(NEXT, return true);

        if(tk(IDENTIFIER).parse(scan))
        {
            std::string gotvar = scan.GetInfo().str;
            if(gotvar != loopvar)
                throw InvalidateAction
                    ("NEXT without FOR (expected "+loopvar+", got "+gotvar+")");
        }

        for_stmt->Optimize();
        SetStmt( std::move(for_stmt), std::move(prev) );

        return true;
    };

    std::function<bool(scanner_t&, bool)> parse_select = [this](scanner_t& scan, bool short_statement) -> bool
    {
        const block_setting<bool> block_marker(in_block, true);

        LambdaExpectToken(CASE, return false);

        LambdaExpectAnyExpr((expr), return true);

        std::unique_ptr<stmt_select> select_stmt(new stmt_select(scan.GetLoc(), std::move(ctx.exp_result)));

        stmtp prev = std::move(ctx.stm_result); // Save the end of statements chain

        if(!(tk(CRLF) | tk(COLONCHAR)).expected().parse(scan)) return true;

        for(;;)
        {
            while(tk(CRLF).parse(scan)) {}
            if(!tk(CASE).parse(scan)) break;

            LambdaExpectSubRule((casematch), break);
            exprp cond = std::move(ctx.exp_result);

            if(short_statement)
            {
                LambdaExpectSubRule((colonchars_shortstatements_and_colonchar), break);
            }
            else
            {
                LambdaExpectSubRule((separator_statements_and_separator), return false);
            }

            select_stmt->cases.emplace_back( std::move(cond), std::move(ctx.stm_result) );
        }
        SetStmt( std::move(select_stmt), std::move(prev) );
        return tq(ENDSELECT).parse(scan);
    };

    std::function<bool(scanner_t&, bool, QBtoken)> parse_doloop = [this](scanner_t& scan, bool short_statement, QBtoken tok) -> bool
    {
        const block_setting<bool> block_marker(in_block, true);

        bool has_starting_condition = false;
        bool negated_condition      = false;

        std::unique_ptr<stmt_doloop> doloop_stmt(new stmt_doloop(scan.GetLoc()));
        if(tok == WHILE)
        {
            has_starting_condition = true;
            doloop_stmt->DisableEXITDO();
        }
        else
        {
            if(tk(WHILE).parse(scan))
                { has_starting_condition = true; }
            else if(tk(UNTIL).parse(scan))
                { has_starting_condition = negated_condition = true; }
        }

        if(has_starting_condition)
        {
            LambdaExpectNumExpr((expr), goto condition_missing1);
            if(negated_condition)
                ctx.exp_result.reset( new expr_unary(NOT, std::move(ctx.exp_result)) );
            doloop_stmt->precondition = std::move(ctx.exp_result);
          condition_missing1:;
        }

        stmtp prev = std::move(ctx.stm_result); // Save the end of statements chain

        if(short_statement)
        {
            LambdaExpectSubRule((colonchars_shortstatements_and_colonchar), return true);
        }
        else
        {
            LambdaExpectSubRule((separator_statements_and_separator), return false);
        }

        doloop_stmt->bodyp = std::move(ctx.stm_result);

        LambdaExpectToken( tok==WHILE ? WEND : LOOP, return true );

        if(tok != WHILE && !has_starting_condition)
        {
            bool has_ending_condition = false;
            if(tk(WHILE).parse(scan))
                { has_ending_condition = true; }
            else if(tk(UNTIL).parse(scan))
                { has_ending_condition = negated_condition = true; }

            if(has_ending_condition)
            {
                LambdaExpectNumExpr((expr), break);
                if(negated_condition)
                    ctx.exp_result.reset( new expr_unary(NOT, std::move(ctx.exp_result)) );
                doloop_stmt->postcondition = std::move(ctx.exp_result);
            }
        }

        SetStmt( std::move(doloop_stmt) , std::move(prev) );
        return true;
    };

    /* ParseVarDef:
     *   bool handle_const(expr_var&& var, exprp&& c)
     *      The returning will be delegated to handle_const,
     *      if the expression happens to be a compile-time constant.
     *      c is a pointer to expression containing the constant value.
     *      The pointer can be moved, changed, etc.
     *
     *   bool handle_sub(expr_var&& var, const SubRoutine& s, bool params)
     *      The returning will be delegated to handle_sub,
     *      if it is determined that the expression is a sub-function call.
     *      var = the expression that is actually a function call.
     *      s   = the subfunction that we think var is actually invoking.
     *      params = true if var includes "()".
     *
     *   During both of these calls, var may be stripped of its possessions,
     *   but it will not be deallocated, and a pointer to it will not be stored anywhere.
     */
    bool ParseVarDef(scanner_t& scan,
                     const std::function<bool(expr_var&&, exprp&&)>& handle_const,
                     const std::function<bool(expr_var&&, const SubRoutine& s, bool params)>& handle_sub)
    {
        LexLocation context = scan.GetLoc();

        if(true) // scope for ambiguous id parsing
        {
            /* Parse a potential variable reference */
            iterator_t it = scan.pos();

            if(false)
            {
            Fail:
                // Fail quietly, because this rule is an alternative
                scan.revert(it);
                return false;
            }

            /* This merely parses stuff, without context or type info */
            std::unique_ptr<expr_var> root(nullptr);
            expr_var* pivot = nullptr;

            bool is_fn = false;

            do {
                LexLocation lexloc = scan.GetLoc();

                if(tk(FNIDENTIFIER).parse(scan))
                {
                    if(root)
                        throw InvalidateAction("TYPEs cannot contain FN members");
                    is_fn = true;
                }
                else
                {
                    if(!tk(IDENTIFIER).parse(scan)) goto Fail;
                    if(is_fn)
                        throw InvalidateAction("FNs cannot contain TYPE elements");
                }

                std::string identifier = scan.GetInfo().str;

                if(root)
                {
                    if(HasTypeSuffix(identifier))
                        throw InvalidateAction("TYPE member cannot have a type suffix");
                }

                std::unique_ptr<expr_var> v(new expr_var(lexloc, identifier));

                if(tk(OPENPARENS).parse(scan))
                {
                    /* indexes */
                    do {
                        LambdaExpectAnyExpr((func_param), goto Fail);
                        v->addindex( std::move(ctx.exp_result) );
                    } while(tk(COMMA).parse(scan));

                    LambdaExpectToken(CLOSEPARENS, goto Fail);
                }

                if(!pivot)
                {
                    root.reset(pivot = v.release());
                }
                else
                {
                    /* It's a member. Therefore this whole construct
                     * is a user type. User types cannot contain type suffixes.
                     */
                    auto p = v.release();
                    pivot->addmember(p);
                    pivot = p;
                }
            } while(tk(PERIOD).parse(scan));

            SetExpr( std::move(root) );
        }
        // ctx.exp_result now contains the expr_var.

        /* If it has indexes, it can be ARRAY/FUNCTION/DEF name.
         * If it has no indexes, it can be CONST/SCALAR/FUNCTIONNAME.
         * If it has members, it was a user type.
         */
        expr_var* var = dynamic_cast<expr_var*> (&*ctx.exp_result);
        if(!var) throw InvalidateAction("Invalid result from ambiguous_id");

        const std::string name = var->getname();

        bool is_static = in_sub && in_sub->is_static;

        bool can_be_scalar = true;
        bool can_be_array  = true;
        bool can_be_funct  = true;
        bool can_be_const  = true;

        if(var->hasmember())
        {
            // If it has '.', it cannot be a CONST, a FUNCTION or a DEF.
            // It can still be an array or a scalar.
            can_be_const  = false;
            can_be_funct  = false;
        }

        if(var->isarray())
        {
            // If it has '(', it cannot be a CONST or a scalar.
            // It can still be an array, a FUNCTION call, or a DEF call.
            can_be_const  = false;
            can_be_scalar = false;
            for(std::size_t b = var->numindexes(), a=0; a<b; ++a)
                if(auto v = dynamic_cast<expr_var*> (&* var->getindex(a)))
                    if(v->iswholearray())
                    {
                        // "a()" is only valid in a function call.
                        // It cannot be an array index.
                        can_be_array = false;
                    }
        }
        else
        {
            //can_be_array = false; -- it can still be an array!
            // TODO: Explain how. Maybe the GET statement?
        }

        /* Create the relevant type of expression */

#ifdef BINDING_DEBUG
        dprintf(dp_declare, "%s may be: %s %s %s %s\n",
            name.c_str(),
            can_be_scalar ? "scalar" : "",
            can_be_array  ? "array" : "",
            can_be_funct  ? "funct" : "",
            can_be_const  ? "const" : "");
#endif
        if(can_be_const)
        {
            // true = allocates a new expr.
            class expr* c = FindConstant(name, true);
            if(c) return handle_const( std::move(*var), exprp(c) );
        }
        if(can_be_funct)
        {
            const SubRoutine* s = FindSubFunc(name);
            if(s) return handle_sub( std::move(*var), *s, var->isarray() );
        }

        TypeP got_var;
        if(can_be_scalar)
        {
            VarContext c = FindVariableFromScope(name, false);
            if(c.info && c.info->type)
            {
                got_var = c.info->type;
            }
            else
            {
                // Implicitly define. Can however only implicitly define standard types.
                if(var->hasmember())
                    throw InvalidateAction("Not defined as a user type (missing DIM?): " + name);
                if(!var->isarray())
                {
                    got_var = ImplicitlyDefineScalar(name, is_static, false, nullptr, context);
#ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Tried to implicitly define scalar %s - %s\n",
                        name.c_str(), got_var?"success":"fail");
#endif
                }
            }
        }
        if(can_be_array && !got_var)
        {
            VarContext c = FindVariableFromScope(name, true);
            if(c.info && c.info->type)
            {
                got_var = c.info->type;
            }
            else
            {
                // Implicitly define. Can however only implicitly define standard types.
                if(var->hasmember())
                    throw InvalidateAction("Not defined as a user type (missing DIM?): " + name);
                if(var->isarray())
                {
                    got_var = ImplicitlyDefineArray(name,
                                                    var->numindexes(),
                                                    option_base, 10,
                                                    is_static,false,nullptr,context);
#ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Tried to implicitly define array %s - %s\n",
                        name.c_str(), got_var?"success":"fail");
#endif
                }
            }
        }
        if(!got_var)
        {
            throw InvalidateAction("Duplicate definition?");
        }

        var->set_basetype(got_var);

        TypeP pivot_found   = std::make_shared<TypeInfo>(*got_var);
        expr_var* pivot_var = var;

        for(;;)
        {
    #ifdef BINDING_DEBUG
            dprintf(dp_declare, "DeeperCheck: pivot(%s)=%s\n",
                name.c_str(), pivot_found->Dump().c_str());
    #endif
            /* Check the members and number of array dimensions for match */
            if(pivot_found->arraytype != TypeInfo::ParamArray)
            {
                if(pivot_var->numindexes() != pivot_found->dimensions.size())
                    throw InvalidateAction("Array dimensions do not match");

                // addressing an element, not the array
                pivot_found->dimensions.clear();
                pivot_found->arraytype = TypeInfo::NotArray;
            }
            else
            {
                // Parameter arrays are more complicated, because their
                // dimensions are unknown
                // Assume the dimensions were matching. Check runtime!
                pivot_found->dimensions.clear();
                pivot_found->arraytype = TypeInfo::NotArray;
            }

            if(!pivot_var->hasmember())
                break;

            if(!pivot_found->members)
                throw InvalidateAction("Cannot pass a struct as rvalue");

            expr_var& mem = pivot_var->getmember();
            const std::string mem_name = mem.getname();

            TypeP memtype = pivot_found->members->GetMember(mem_name);
            if(!memtype) throw InvalidateAction("No such member variable");
            mem.set_type(memtype);
            mem.set_basetype(memtype);

            pivot_found = memtype;
            pivot_var   = &mem;
        }

        var->set_type(pivot_found);
        // var is typecast pointer from ctx.exp_result.
        // The actual pointer in ctx.exp_result is not modified.

        return true;
    }

public:
    QB_parser()
    {
    #ifdef PARSER_DEBUG
        program.setname("program");
        type.setname("type");
        expr.setname("expr");
        anylabeltarget.setname("anylabeltarget");
        step_coordinate.setname("step_coordinate");
        plain_coordinate.setname("plain_coordinate");
        longstatement.setname("longstatement");
        longblockstatement.setname("longblockstatement");
        shortstatement.setname("shortstatement");
        shortif.setname("shortif");
        casematch.setname("casematch");
        simplestatement.setname("simplestatement");
        expr_meat.setname("expr_meat");
        var_target.setname("var_target");
        autofilenum.setname("autofilenum");
        func_param.setname("func_param");
        colonchars_shortstatements_and_colonchar.setname("colonchars_shortstatements_and_colonchar");
        separator_statements_and_separator.setname("separator_statements_and_separator");
    #endif
    }

    bool parse(scanner_t& scanner) const
    {
        return program.parse(scanner);
    }

    #undef expr
};

// lowercase means optional
// S       string
// N       numeric
// X       string or numeric
// L       anylabeltarget
// R       anylabeltarget where zero constitutes as no parameter
// P       step coordinate                                              param: flag,x,y
// C       coordinate (no step)                                         param: x,y
// V       var_target
// VS      var_target that must be a string
// VN      var_target that must be numeric
// A       array with optional indexes
// F       file number
// ZG      GOSUB                                     (ON PLAY/KEY/TIMER)
// ZH      LIST/OFF/ON                               (KEY)              param: 0/-1/1
// ZK      STOP/OFF/ON                               (PLAY,TIMER)       param: 0/-1/1
// ZL      B/BF                                      (LINE)             param: 1/2 [0]
// ZO      GOTO                                      (ON ERROR)
// ZA      AS                                        (FIELD)
// ZP      XOR/PSET/PRESET/AND/OR, XOR being default (PSET/PRESET)      param: 0/1/2/3/4 [0]
// ZS      SCREEN                                    (VIEW)
// ZT      TO                                        (VIEW PRINT)

// ,-=()   literals
// %       any >=1 number of following separated by comma
// ?%      any >=0 number of following separated by comma
//
// |       another alternative for entire syntax
// ?       if next token is not matched, parsing skips successfully
//         to the next closing parenthesis or end of pattern
//
//
// spaces are ignored
template<typename RetType>
static bool ParseLibItem(QB_parser& parser, scanner_t& scan, QBtoken tok,
                         const char* defs, std::unique_ptr<RetType>& result,
                         std::string& matched_pattern)
{
    std::vector<exprp> params;
    std::vector<bool> lvaluemap;
    iterator_t it = scan.pos();

    LexLocation lexloc = scan.GetLoc();

    #define MATCH()   matched_pattern += std::toupper(MatchType)
    #define NOMATCH() matched_pattern += std::tolower(MatchType)
    /*
      Matched pattern keywords:
        S = simple (string)
        N = simple (num)
        X = S or N
        L = label
        P = stepcoordinate (three nums)
        C = coordinate (two nums)
        V = lvalue
    */

    const char* loopbegin = 0;
    bool        loop_opt = false;
    matched_pattern.clear();
  reloop:
    while(*defs)
    {
        bool optional = loop_opt;
        char MatchType = 0;
    reswitch:

        // FIXME: In case of optional parameters (such as with CLOSE),
        // figure out a way to suppress "Expected:" errors

        iterator_t curit = scan.pos();
        if(*defs == '|') break;
        //fprintf(stderr, "'%c' (%s) (%s)? - optionality=%d\n", *defs, defs, GetTokenName(tok).c_str(), optional);
        switch(*defs)
        {
            case ' ': ++defs; goto reswitch;
            case '?':
                optional = true; ++defs; goto reswitch;
            case 'S': case 's':
                MatchType='S';
                if(!parser.expr.parse(scan)
                || ctx.exp_result->is_string().is_false())
                { if(*defs=='S')goto Failed;
                IgnoreFail: scan.revert(curit); NOMATCH();
                  params.push_back(nullptr); break; }
                MATCH(); params.emplace_back(std::move(ctx.exp_result)); break;
            case 'N': case 'n':
                MatchType='N';
                if(!parser.expr.parse(scan)
                || ctx.exp_result->is_num().is_false())
                { if(*defs=='N')goto Failed; goto IgnoreFail; }
                MATCH(); params.emplace_back(std::move(ctx.exp_result)); break;
            case 'X': case 'x':
                MatchType='X';
                if(!parser.expr.parse(scan))
                { if(*defs=='X')goto Failed; goto IgnoreFail; }
                MATCH(); params.emplace_back(std::move(ctx.exp_result)); break;
            case 'L': case 'l':
                MatchType='L';
                if(!parser.anylabeltarget.parse(scan))
                { if(*defs=='L')goto Failed; goto IgnoreFail; }
                MATCH(); params.emplace_back(std::move(ctx.exp_result)); break;
            case 'R': case 'r':
                MatchType='L';
              { if(!parser.anylabeltarget.parse(scan))
                { if(*defs=='R')goto Failed; goto IgnoreFail; }
                expr_label* labelp = dynamic_cast<expr_label*>(&*ctx.exp_result);
                if(labelp && labelp->get_label().empty()) ctx.exp_result = nullptr;
                if(ctx.exp_result) { MATCH(); } else { NOMATCH(); }
                params.emplace_back(std::move(ctx.exp_result)); break; }
            case 'P': case 'p':
                MatchType='P';
              { if(!parser.step_coordinate.parse(scan))
                { if(*defs=='P')goto Failed; goto IgnoreFail; }
                MATCH();
                params.emplace_back(new expr_number(coord_res.step));
                params.emplace_back(std::move(coord_res.x));
                params.emplace_back(std::move(coord_res.y));
                break; }
            case 'C': case 'c':
                MatchType='C';
              { if(!parser.plain_coordinate.parse(scan))
                { if(*defs=='C')goto Failed; goto IgnoreFail; }
                MATCH();
                params.emplace_back(std::move(coord_res.x));
                params.emplace_back(std::move(coord_res.y));
                break; }
            case 'V': case 'v':
                MatchType='V';
                if(!parser.var_target.parse(scan))
                { if(*defs=='V')goto Failed; goto IgnoreFail; }
                MATCH();
                params.emplace_back(std::move(ctx.exp_result));
                lvaluemap.resize(params.size());
                lvaluemap[params.size()-1] = true;
                if(defs[1]=='S') { ++defs; TypeCheckString(ctx.exp_result); }
                if(defs[1]=='N') { ++defs; TypeCheckNum(ctx.exp_result);    }
                break;
            case 'F': case 'f':
                MatchType='N';
                if(!parser.autofilenum.parse(scan))
                { if(*defs=='F')goto Failed; goto IgnoreFail; }
                MATCH();
                params.emplace_back(std::move(ctx.exp_result)); break;
            case 'A': case 'a':
            {
                /* TODO */
                /* array name, possibly with starting offset */
                MatchType='A';
                if(!parser.expr.parse(scan))
                { if(*defs=='A')goto Failed; goto IgnoreFail; }
                MATCH(); params.emplace_back(std::move(ctx.exp_result)); break;
                break;
            }
            case ',': case '-':
            case '=': case '(':
            case ')':
                if(!tk(*defs).parse(scan)) goto Failed;
                break;
            case '%':
                loopbegin = ++defs;
                loop_opt = optional;
                goto reswitch;
            case 'Z': case 'z':
            {
                switch(defs[1])
                {
                    case 'P': case 'p': // XOR/PSET/PRESET/AND/OR
                        MatchType='N';
                        MATCH();
                        if(tk(XOR).parse(scan)) params.emplace_back(new expr_number(0));
                        else if(tk(PSET).parse(scan)) params.emplace_back(new expr_number(1));
                        else if(tk(PRESET).parse(scan)) params.emplace_back(new expr_number(2));
                        else if(tk(AND).parse(scan)) params.emplace_back(new expr_number(3));
                        else if(tk(OR).parse(scan)) params.emplace_back(new expr_number(4));
                        else params.emplace_back(new expr_number(0));
                        ++defs; break;
                    case 'T': case 't': // TO
                        if(!tk(TO).parse(scan))
                            { Zfail: if(*defs=='Z') goto Failed; ++defs; goto IgnoreFail; }
                        ++defs; break;
                    case 'L': case 'l': // B/BF
                        MatchType='N';
                      { if(!tk(IDENTIFIER).parse(scan))
                            goto Zfail;
                        const std::string id = UCase(scan.GetInfo().str);
                        if(id == "B") { MATCH(); params.emplace_back(new expr_number(1)); }
                        else if(id == "BF") { MATCH(); params.emplace_back(new expr_number(2)); }
                        else goto Zfail;
                        ++defs; break; }
                    case 'S': case 's': // SCREEN
                        MatchType='N';
                        if(tk(SCREEN).parse(scan)) params.emplace_back(new expr_number(1));
                        else params.emplace_back(new expr_number(0));
                        MATCH();
                        ++defs; break;
                    case 'K': case 'k': // STOP/OFF/ON
                        MatchType='N';
                        if(tk(STOP).parse(scan)) params.emplace_back(new expr_number(0));
                        else if(tk(OFF).parse(scan)) params.emplace_back(new expr_number(-1));
                        else if(tk(ON).parse(scan)) params.emplace_back(new expr_number(1));
                        else goto Zfail;
                        MATCH();
                        ++defs; break;
                    case 'H': case 'h': // LIST/OFF/ON
                        MatchType='N';
                        if(tk(LIST).parse(scan)) params.emplace_back(new expr_number(0));
                        else if(tk(OFF).parse(scan)) params.emplace_back(new expr_number(-1));
                        else if(tk(ON).parse(scan)) params.emplace_back(new expr_number(1));
                        else goto Zfail;
                        MATCH();
                        ++defs; break;
                    case 'G': case 'g': // GOSUB
                        if(!tk(GOSUB).parse(scan))
                            goto Zfail;
                        ++defs; break;
                    case 'O': case 'o': // GOTO
                        if(!tk(GOTO).parse(scan))
                            goto Zfail;
                        ++defs; break;
                    case 'A': case 'a': // AS
                        if(!tk(AS).parse(scan))
                            goto Zfail;
                        ++defs; break;
                }
            }
        }
        //fprintf(stderr, "- ok\n");
        if(false)
        {
        Failed:
            //fprintf(stderr, "- fail\n");
            if(optional)
            {
                while(*defs && *defs != '|' && *defs != ')') ++defs;
                if(*defs == ')') { continue; }
                if(!*defs) break;
            }
            while(*defs && *defs!='|') ++defs;
            if(*defs=='|')
            {
                ++defs;
                scan.revert(it);
                params.clear();
                lvaluemap.clear();
                matched_pattern.clear();
                continue;
            }
            return false;
        }
        ++defs;
    }
    if(loopbegin && tk(COMMA).parse(scan))
    {
        defs = loopbegin;
        goto reloop;
    }

    result.reset( new RetType(lexloc, tok) );
    result->set_params( std::move(params) );
    result->set_lvalue_map(lvaluemap);

    return true;
}

/************************************/

#include <iostream> // For std::cout
#include <fstream>

bool ParseAndCompile(const std::string& filename, const std::string& outfilename)
{
    /*setbuf(stdin,NULL);*/
    FILE* fp = stdin;
    bool want_stdin = filename.empty() || filename == "-";

    if(!want_stdin)
    {
        fp = std::fopen(filename.c_str(), "rt");
        if(!fp)
        {
            std::perror(filename.c_str());
            return false;
        }
    }

    yyStartFile(fp, want_stdin ? "<stdin>" : filename.c_str());

    Lex<QBtoken> lexer;
    scanner_t scan(lexer.begin(), lexer.end());
    QB_parser pa;

    pa.parse(scan);

    if(!scan.eof())
        yyerror("Incomplete parse\n", scan.pos().GetLoc());

    if(outfilename.empty() || outfilename == "-")
        Compile(std::cout, &*ctx.stm_result, global_data);
    else
    {
        std::ofstream out;
        out.open(outfilename, std::ios_base::out | std::ios_base::trunc);
        Compile(out, &*ctx.stm_result, global_data);
    }
    return true;
}
