#ifndef bqtQBlibBittypesHH
#define bqtQBlibBittypesHH

#include <stdint.h>

template<unsigned nbits,bool is_signed,
         bool fitsin8  = (nbits<=8),
         bool fitsin16 = (nbits<=16),
         bool fitsin32 = (nbits<=32) >
struct B_OptimalType
{
    typedef uint_least64_t result;
};

template<unsigned nbits>
struct B_OptimalType<nbits,false,true,true,true>
{
    typedef uint_least8_t result;
};

template<unsigned nbits>
struct B_OptimalType<nbits,false,false,true,true>
{
    typedef uint_least16_t result;
};

template<unsigned nbits>
struct B_OptimalType<nbits,false,false,false,true>
{
    typedef uint_least32_t result;
};

template<unsigned nbits>
struct B_OptimalType<nbits,true, true,true,true>
{
    typedef int_least8_t result;
};

template<unsigned nbits>
struct B_OptimalType<nbits,true, false,true,true>
{
    typedef int_least16_t result;
};

template<unsigned nbits>
struct B_OptimalType<nbits,true, false,false,true>
{
    typedef int_least32_t result;
};

template<unsigned nbits>
struct B_OptimalType<nbits,true, false,false,false>
{
    typedef int_least64_t result;
};


/* This implements a bitfield that can be put into an union.
 *
 *   bitno     = starting bit number
 *   nbits     = number of bits
 *   is_signed = whether this datatype acts as "signed" or "unsigned" type
 *   totalbits = number of bits total in the regbitset.
 *
 *   
 
 */
template<unsigned bitno, unsigned nbits, bool is_signed, unsigned totalbits>
struct RegBitSet
{
    /* R: The POD-datatype that represents the value of this bit-variable. */
    using R = typename B_OptimalType<nbits,is_signed>::result;

    /* Bitmask: A binary number that has nbits 1-bits. pow(2, nbits) - 1 */
    static constexpr unsigned long long mask = ((1ull << (nbits/2)) << (nbits-(nbits/2))) - 1ull;
    static constexpr unsigned long long signbit = (1ull << (nbits-1));
private:
    R data;

public:
    operator R() const
    {
        R data = 0; /* FIXME BROKEN UNIMPLEMENTED */

        if(is_signed && (data & signbit)) return -(mask & -data);
        return data & mask;
    }

    template<typename U>
    RegBitSet& operator=(U val)
    {
        /* Poke the given value into the data.
         * If nbits=1, cast val as bool before doing the work.
         */
        if(nbits == 1) val = (bool)val; else val &= mask;

        /* FIXME BROKEN UNIMPLEMENTED */

        return *this;
    }

    /* FIXME: Optimized solutions are available for +, -, &, | and ^, not involving >> operations */

    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator+= (U b) { return *this = operator R() + b; }
    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator-= (U b) { return *this = operator R() - b; }
    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator*= (U b) { return *this = operator R() * b; }
    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator/= (U b) { return *this = operator R() / b; }
    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator%= (U b) { return *this = operator R() % b; }
    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator&= (U b) { return *this = operator R() & b; }
    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator|= (U b) { return *this = operator R() | b; }
    template<typename U>
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator^= (U b) { return *this = operator R() ^ b; }

    template<typename U>
    auto operator+ (U b) const -> decltype(R() + b)  { return operator R() + b; }
    template<typename U>
    auto operator- (U b) const -> decltype(R() - b)  { return operator R() - b; }
    template<typename U>
    auto operator* (U b) const -> decltype(R() * b)  { return operator R() * b; }
    template<typename U>
    auto operator/ (U b) const -> decltype(R() / b)  { return operator R() / b; }
    template<typename U>
    auto operator% (U b) const -> decltype(R() % b)  { return operator R() % b; }
    template<typename U>
    auto operator& (U b) const -> decltype(R() & b)  { return operator R() & b; }
    template<typename U>
    auto operator| (U b) const -> decltype(R() | b)  { return operator R() | b; }
    template<typename U>
    auto operator^ (U b) const -> decltype(R() ^ b)  { return operator R() ^ b; }
    R operator~() const { return ~operator R(); }
    bool operator!() const { return !operator R(); }

    R operator++ (int) { R result = operator R(); operator += (1); return result; }
    R operator-- (int) { R result = operator R(); operator -= (1); return result; }
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator++ () { return operator+=(1); }
    RegBitSet<bitno,nbits,is_signed,totalbits>& operator-- () { return operator-=(1); }
};

/* This implements standalone _BIT types, with wrap-around semantics */
template<typename T, unsigned nbits>
class B_BIT
{
    static constexpr unsigned long long mask = ((1ull << (nbits/2)) << (nbits-(nbits/2))) - 1ull;
    static constexpr unsigned long long signbit = (1ull << (nbits-1));
    static constexpr bool is_signed = T(-1) < 0;

private:
    T data = T();

public:
    B_BIT() : data() {}
    B_BIT(T value) : data(value) {}

    B_BIT<T,nbits>& operator+= (const B_BIT<T,nbits>& b) { data += b.data; return *this; }
    B_BIT<T,nbits>& operator-= (const B_BIT<T,nbits>& b) { data -= b.data; return *this; }
    B_BIT<T,nbits>& operator*= (const B_BIT<T,nbits>& b) { data *= b.data; return *this; }
    B_BIT<T,nbits>& operator/= (const B_BIT<T,nbits>& b) { data /= b.data; return *this; }
    B_BIT<T,nbits>& operator%= (const B_BIT<T,nbits>& b) { data %= b.data; return *this; }
    B_BIT<T,nbits>& operator&= (const B_BIT<T,nbits>& b) { data &= b.data; return *this; }
    B_BIT<T,nbits>& operator|= (const B_BIT<T,nbits>& b) { data |= b.data; return *this; }
    B_BIT<T,nbits>& operator^= (const B_BIT<T,nbits>& b) { data ^= b.data; return *this; }

    template<typename U>
    auto operator+ (U b) const -> decltype(T() + b)  { return operator T() + b; }
    template<typename U>
    auto operator- (U b) const -> decltype(T() - b)  { return operator T() - b; }
    template<typename U>
    auto operator* (U b) const -> decltype(T() * b)  { return operator T() * b; }
    template<typename U>
    auto operator/ (U b) const -> decltype(T() / b)  { return operator T() / b; }
    template<typename U>
    auto operator% (U b) const -> decltype(T() % b)  { return operator T() % b; }
    template<typename U>
    auto operator& (U b) const -> decltype(T() & b)  { return operator T() & b; }
    template<typename U>
    auto operator| (U b) const -> decltype(T() | b)  { return operator T() | b; }
    template<typename U>
    auto operator^ (U b) const -> decltype(T() ^ b)  { return operator T() ^ b; }

    B_BIT<T,nbits> operator~() const { return ~data; } // FIXME: Should we return T instead?
    bool operator!() const { return !operator T(); }

    T operator++ (int) { T result = operator T(); operator += (1); return result; }
    T operator-- (int) { T result = operator T(); operator -= (1); return result; }
    B_BIT<T,nbits>& operator++ () { return operator+=(1); }
    B_BIT<T,nbits>& operator-- () { return operator-=(1); }

    operator T() const
    {
        if(is_signed && (data & signbit)) return -(mask & -data);
        return data & mask;
    }
    operator bool() const { return operator T(); }
};

#endif
