#include "error.hh"
#include "exec.hh"

#include <cstring>
#include <errno.h>

/* This table of error messages comes almost directly
 * from QuickBASIC. Most of the messages are obsolete,
 * but they're maintained for compatibility with older
 * BASIC versions such as GW-BASIC.
 * (!) indicates that the message does not exist in QB.
 */
static const char* const ErrorMessages[] =
{
    "NEXT without FOR", //1,
    "Syntax error", //2,
    "RETURN without GOSUB", //3,
    "Out of DATA", //4,
    "Illegal function call", //5,
    "Overflow", //6,
    "Out of memory", //7,
    "Label not defined", //8,
    "Subscript out of range", //9,
    "Duplicate definition", //10,
    "Division by zero", //11,
    "Illegal in direct mode", //12,
    "Type mismatch", //13,
    "Out of string space", //14,
    "String too long", // 15 (!)
    "String formula too complex", //16,
    "Cannot continue", //17,
    "Function not defined", //18,
    "No RESUME", //19,
    "RESUME without error", //20,
    "Unprintable error", // 21
    "Missing operand", // 22 (!)
    "Line buffer overflow", // 23 (!)
    "Device timeout", //24,
    "Device fault", //25,
    "FOR without NEXT", //26,
    "Out of paper", //27,
    0,
    "WHILE without WEND", //29,
    "WEND without WHILE", //30,
    0,
    0,
    "Duplicate label", //33,
    0,
    "Subprogram not defined", //35,
    "Subprogram already in use", //36 (!),
    "Argument-count mismatch", //37,
    "Array not defined", //38,
    "CASE ELSE expected", //39,
    "Variable required", //40,
    0,0,0,0,
    0,0,0,0,
    0,
    "FIELD overflow", //50,
    "Internal error", //51,
    "Bad file name or number", //52,
    "File not found", //53,
    "Bad file mode", //54,
    "File already open", //55,
    "FIELD statement active", //56,
    "Device I/O error", //57,
    "File already exists", //58,
    "Bad record length", //59,
    0,
    "Disk full", //61,
    "Input past end of file", //62,
    "Bad record number", //63,
    "Bad file name", //64,
    0,
    "Direct statement in file", //66,
    "Too many files", //67,
    "Device unavailable", //68,
    "Communication-buffer overflow", //69,
    "Permission denied", //70,
    "Disk not ready", //71,
    "Disk-media error", //72,
    "Advanced feature unavailable", //73,
    "Rename across disks", //74,
    "Path/File access error", //75,
    "Path not found", //76,
    "Deadlock", //77,

    "Too many open files", // 78 (QBC)
    "Stack overflow", // 79 (QBC)
    "STOP" // 80 (QBC)
};

void* B_ErrorHandler = nullptr;
bool B_ErrorHandling_Enabled = true;
int B_ERR_VALUE;
int B_ERL_VALUE;

static void* B_ResumeRepeat = nullptr;
static void* B_ResumeIgnore = nullptr;

void B_ERROR_Init2(const B_ErrorPacket_Full& err)
{
    B_ERL_VALUE = err.line;
    if(!B_ErrorHandler)
    {
        constexpr BAS_INT NumMessages = sizeof(ErrorMessages) / sizeof(*ErrorMessages);
        int code = B_ERR_VALUE;
        if(code < 1
        || code > NumMessages
        || !ErrorMessages[code-1]) code = 21; // Unprintable error

        fprintf(stderr, "%s (error %d) at %s:%d:%d\n",
            ErrorMessages[code-1], (int)B_ERR_VALUE,
            err.filename ? err.filename : "(unknown)", (int)B_ERL_VALUE, err.column);

        //B_StmtSYSTEM(code);
        B_LibDone();
        std::abort();
    }
    B_ResumeRepeat = err.repeat;
    B_ResumeIgnore = err.ignore;
}
void B_ERROR_Init(const B_ErrorPacket_Brief& err)
{
    B_ERL_VALUE = -1;
    if(!B_ErrorHandler)
    {
        constexpr BAS_INT NumMessages = sizeof(ErrorMessages) / sizeof(*ErrorMessages);
        int code = B_ERR_VALUE;
        if(code < 1
        || code > NumMessages
        || !ErrorMessages[code-1]) code = 21; // Unprintable error

        fprintf(stderr, "%s (error %d)\n", ErrorMessages[code-1], (int)B_ERR_VALUE);

        //B_StmtSYSTEM(code);
        B_LibDone();
        std::abort();
    }
    B_ResumeRepeat = err.repeat;
    B_ResumeIgnore = err.ignore;
}


void* B_RESUME_Execute(int mode)
{
    if(!B_ResumeIgnore) B_ERROR(20);/* RESUME WITHOUT ERROR*/
    void* result = nullptr;;
    switch(mode)
    {
        case 0: result = B_ResumeIgnore; break;
        case 1: result = B_ResumeRepeat; break;
    }
    B_ResumeIgnore = nullptr;
    B_ResumeRepeat = nullptr;
    B_ERR_VALUE = 0;
    B_ERL_VALUE = 0;
    return result;
}

int B_ERROR_TranslateErrno()
{
    switch(errno)
    {
        case EPERM:  return(70);  // permission denied
        case EACCES: return(75);  // path/file access error
        case EEXIST: return(58);  // file already exists
        case EFAULT: return(57);  // seg fault? FIXME ("Device I/O error")
        case ENAMETOOLONG:
        case EFBIG:  return(15);  // string too long
        case EISDIR: return(75);  // path/file access error
        case ELOOP:  return(16);  // too many loops (string formula too complex)
        case ENFILE:
        case EMFILE: return(78);  // too many open files
        case ENODEV: return(71);  // disk not ready? no device
        case ENXIO:  return(71);  // disk not ready? no device
        case ENOENT: return(53);  // file not found
        case ENOSPC: return(61);  // disk full
        case ENOTDIR: return(75);  // path/file access error
        case EROFS:  return(75);  // -"- (read only filesystem)
        case ETXTBSY: return(55);  // file already open
        case EIO: return(57);  // device I/O error
        case ENOMEM: return(7); // Out of memory
        default:  break;
    }
    return(21); // unprintable
}

#undef B_StmtERROR

void B_StmtERROR(int code)
{
    B_ERR_VALUE = code;
    B_ErrorPacket_Brief err = {nullptr,nullptr};
    B_ERROR_Init(err);
}

void B_ERROR(int code)
{
    B_ERR_VALUE = code;
    if(B_ErrorHandling_Enabled) throw B_ErrorEv{};

    B_ErrorPacket_Brief err = {nullptr,nullptr};
    B_ERROR_Init(err);
    // should not reach here!
    std::abort();
}
