#include <map>
#include <memory>

#include "gfx.hh"
#include "error.hh"

#include "gfx_api.hh"
#include "gfx_term.hh"
#include "gfx_gfx.hh"

static std::shared_ptr<Screen> CurScreen = nullptr;
static std::map<int, std::shared_ptr<Screen>> OtherScreens;

void GFX_Init()
{
    Term_Init();
    OtherScreens[0] = std::make_shared<TerminalScreen>();
    CurScreen = OtherScreens[0];
}
void GFX_Done()
{
    OtherScreens.clear();
    Term_Done();
}

int B_FuncCSRLIN()
{
    std::pair<int,int> pos = CurScreen->GetCursorPos();
    return pos.second;
}
int B_FuncPOS(int)
{
    std::pair<int,int> pos = CurScreen->GetCursorPos();
    return pos.first;
}

int B_FuncWIDTH()
{
    std::pair<int,int> pos = CurScreen->GetWinSize();
    return pos.first;
}

void GFX_OutS(const char* buffer, size_t nbytes)
{
    CurScreen->OutS(buffer, nbytes);
}

void B_StmtVIEWPRINT(int from, int to)
{
    CurScreen->SetViewPrint(from, to);
}

int B_FuncSCREEN(int y,int x, int colorflag)
{
    std::pair<int,int> chr = CurScreen->ReadScreen(x,y);
    if(colorflag)
        return chr.second;
    return chr.first;
}

void B_StmtCLS(int type)
{
    CurScreen->CLS(type);
}

void B_StmtWIDTH(int w,int h)
{
    CurScreen->SetWinSize(std::pair<int,int>(w,h), false);
}

void B_StmtSCREEN(int mode,int color,int apage,int vpage)
{
    color=color; // We ignore the color parameter.

    if(mode < 0)
    {
        auto i = OtherScreens.find(mode);
        if(i == OtherScreens.end()) B_ERROR(5); // Illegal function call
        CurScreen = i->second;
        return;
    }

    /*
     SCREEN GWIDTH TWIDTH   ATTRIBUTES  PALETTE   PAGES
     0              40x25   16,16,16    64        8
     0              40x43   16,16,16    64        8
     0              40x50   16,16,16    64        8
     0              80x25   16,16,16    64        8
     0              80x43   16,16,16    64        8
     0              80x50   16,16,16    64        8

     1      320x200 40x25   4,4         16        1
     2      640x200 80x25   2,2         16        1
     3      720x348 80x25   1,1         -         2
     4      640x400 80x25   16,1        -         1
     7      320x200 40x25   16,16       16        8
     8      640x200 80x25   16,16       16        4

     9      640x350 80x25   16,16       64        2
     9      640x350 80x43   16,16       64        2

     10     640x350 80x25   4,4         9         2
     10     640x350 80x43   4,4         9         2

     11     640x480 80x30   2,2         262144    1
     11     640x480 80x60   2,2         262144    1

     12     640x480 80x30   16,16       262144    1
     12     640x480 80x60   16,16       262144    1

     13     320x200 40x25   256,256     262144    1
    */

    std::pair<int,int> oldwidth = CurScreen->GetWinSize();

    if(mode == 0)
    {
        if(!dynamic_cast<TerminalScreen*>(CurScreen.get()))
        {
            std::shared_ptr<TerminalScreen> p = std::make_shared<TerminalScreen> ();
            p->SetPages(8);
            OtherScreens[0] = CurScreen = p;
        }
    }
    else
    {
        int w=0, h=0; // text dimensions
        int xs=0, ys=0;   // gfx dimensions
        int fc=0, bc=0;   // default colors
        int cm=0, pm=0;   // color mode, palette mode
        int p=0; // pages

        unsigned cpmode = 0;

        switch(mode)
        {
            case 1:
                // pal: 0x00,0x3B,0x3D,0x3F
                w=40; h=25; xs=320; ys=200; fc=3; bc=0; p=1; cpmode = CM_CGA + PM_EGA; break;
            case 2:
                // pal: 0x00,0x3F
                w=80; h=25; xs=640; ys=200; fc=1; bc=0; p=1; cpmode = CM_Mono + PM_EGA; break;
            case 3:
                w=80; h=25; xs=720; ys=348; fc=1; bc=0; p=2; cpmode = CM_Mono + PM_Mono; break;
            case 4:
                w=80; h=25; xs=640; ys=400; fc=1; bc=0; p=1; cpmode = CM_Mono + PM_EGA; break;
            case 7:
                w=40; h=25; xs=320; ys=200; fc=15;bc=0; p=8; cpmode = CM_EGA + PM_EGA; break;
            case 8:
                w=80; h=25; xs=640; ys=200; fc=15;bc=0; p=4; cpmode = CM_EGA + PM_EGA; break;
            case 9:
                // color mapping: 0123 -> 0145
                w=80; h=25; xs=640; ys=350; fc=15;bc=0; p=2; cpmode = CM_EGA + PM_EGA64; break;
            case 10:
                // pal: 0,8,0,0,24,24,0,0,0,8,0,0,0,24,0,0
                // color mapping: 0123 -> 0145
                w=80; h=25; xs=640; ys=350; fc=3; bc=0; p=2; cpmode = CM_S10 + PM_S10; break;
            case 11:
                // pal: 0,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,0
                w=80; h=30; xs=640; ys=480; fc=1; bc=0; p=1; cpmode = CM_Mono + PM_VGA; break;
            case 12:
                // pal: 0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16
                w=80; h=30; xs=640; ys=480; fc=15;bc=0; p=1; cpmode = CM_EGA + PM_VGA; break;
            case 13:
                // pal: 256 colors
                w=40; h=25; xs=320; ys=200; fc=15;bc=0; p=1; cpmode = CM_S13 + PM_VGA; break;
            default:
                B_ERROR(5); // Illegal function call
        }

        GraphicScreen* gfx = dynamic_cast<GraphicScreen*>(CurScreen.get());
        if(!gfx)
        {
            std::shared_ptr<GraphicScreen> p = std::make_shared<GraphicScreen> ();
            OtherScreens[0] = CurScreen = p;
            gfx = p.get();
        }
        gfx->SetModeNumber(mode);
        gfx->SetGFXresolution(xs,ys, cpmode);
        gfx->SetPages(p);
        gfx->SetWinSize(std::pair<int,int>(w,h), false);
        gfx->SetColor(fc,bc);
    }

    CurScreen->SetWinSize(oldwidth, true); // try to approximate the same size
    if(apage >= 0) CurScreen->SetAPage(apage);
    if(vpage >= 0) CurScreen->SetVPage(vpage);
}

int B_Funcqb64NEWIMAGE(int width, int height, int bpp)
{
    std::shared_ptr<GraphicScreen> p = std::make_shared<GraphicScreen> ();

    unsigned cpmode, fc=0,bc=0;

    switch(bpp)
    {
        case 1: fc=3; bc=0; cpmode = CM_CGA + PM_EGA; break;
        case 2: fc=1; bc=0; cpmode = CM_Mono + PM_EGA; break;
        case 3: fc=1; bc=0; cpmode = CM_Mono + PM_Mono; break;
        case 4: fc=1; bc=0; cpmode = CM_Mono + PM_EGA; break;
        case 7: fc=15; bc=0; cpmode = CM_EGA + PM_EGA; break;
        case 8: fc=15; bc=0; cpmode = CM_EGA + PM_EGA; break;
        case 9: fc=15; bc=0; cpmode = CM_EGA + PM_EGA64; break;
        case 10: fc=3; bc=0; cpmode = CM_S10 + PM_S10; break;
        case 11: fc=1; bc=0; cpmode = CM_Mono + PM_VGA; break;
        case 12: fc=15; bc=0; cpmode = CM_EGA + PM_VGA; break;
        case 13: fc=15; bc=0; cpmode = CM_S13 + PM_VGA; break;
        case 256: fc=15; bc=0; cpmode = CM_S13 + PM_VGA; break;
        case 32:  fc=0xFFFFFF; bc=0; cpmode = CM_RGB32 + PM_Mono; break;
        default:
            B_ERROR(5); // Illegal function call
    }

    p->SetModeNumber(-1);
    p->SetGFXresolution(width, height, cpmode);
    p->SetPages(8);
    p->SetWinSize(std::pair<int,int> (width,height), false);
    p->SetColor(fc, bc);
    int screen_number = OtherScreens.begin()->first - 1;
    OtherScreens[screen_number] = p;
    return screen_number;
}

void B_Stmtqb64DONTBLEND(int)
{
    // FIXME TODO: Unimplemented
}

void B_StmtCOLOR(int fg,int bg,int border)
{
    if(CurScreen->GetModeNumber() == 1)
    {
        // In screen 1, the following exceptions apply:
        // fg means background
        // bg means palette
        CurScreen->SetColor(0, fg, border);
        if(bg >= 0) switch(bg & 1)
        {
            case 0:
                // set the following palettes:
                //   1 = 2(green); 2 = 4(red); 3 = 6(yellow)
                //FIXME
                break;
            case 1:
                // set the following palettes:
                //   1 = 3(cyan); 2 = 5(magenta); 3=7(white)
                //FIXME
                break;
        }
        return;
    }

    CurScreen->SetColor(fg, bg, border);
}

void B_StmtLOCATE(int row,int col,int c_vis,int c_start,int c_stop)
{
    if(row != -1 || col != -1)
    {
        std::pair<int,int> pos;
        if(row == -1 || col == -1) pos = CurScreen->GetCursorPos();
        if(col != -1) pos.first = col;
        if(row != -1) pos.second = row;
        CurScreen->SetCursorPos(pos);
    }
    if(c_vis != -1)
        CurScreen->SetCursorVisibility(c_vis);
    if(c_start != -1 || c_stop != -1)
        CurScreen->SetCursorSize(c_start, c_stop);
}

void B_StmtLOCATE(int row,int col,int c_vis)
{
    B_StmtLOCATE(row,col,c_vis,-1,-1);
}

void B_StmtLOCATE(int row,int col)
{
    B_StmtLOCATE(row,col,-1);
}

void B_ScreenPOKE(int x,int y, int colorflag, int value)
{
    CurScreen->ScreenPOKE(x,y,colorflag,value);
}


void B_StmtPALETTE(int index, int value)
{
    CurScreen->SetPalette(index, value);
}

void B_StmtLINE(bool step1, double x1,double y1,
                bool step2, double x2,double y2,
                int color, int fill, int style)
{
    std::pair<int,int> coord1 = CurScreen->TranslateCoord(step1,x1,y1);
    std::pair<int,int> coord2 = CurScreen->TranslateCoord(step2,x2,y2);

    if(color==-1) color=CurScreen->GetFGC();

    if(fill==1)
    {
        CurScreen->Rectangle(coord1.first,coord1.second,
                             coord2.first,coord2.second,
                             color,style);
    }
    else if(fill==2)
    {
        CurScreen->FilledRectangle(coord1.first,coord1.second,
                                   coord2.first,coord2.second,
                                   color);
    }
    else
    {
        CurScreen->Line(coord1.first,coord1.second,
                        coord2.first,coord2.second,
                        color,style);
    }
}

void B_StmtCIRCLE(bool step, double x,double y, double radius, int color, double start, double end, double aspect)
{
    std::pair<int,int> coord = CurScreen->TranslateCoord(step,x,y);
    int scaled_radius = radius;
    Screen::CircleRenderer r(coord.first,coord.second, scaled_radius, &*CurScreen);
    if(start  != -1) r.SetBeginAngle(start);
    if(end    != -1) r.SetEndAngle(end);
    if(aspect != -1) r.SetAspect(aspect);
    if(color==-1) color=CurScreen->GetFGC();
    r.Render(color);
}

void B_StmtPSET(bool step,double x,double y, int color)
{
    if(color==-1) color=CurScreen->GetFGC();
    std::pair<int,int> coord = CurScreen->TranslateCoord(step,x,y);
    CurScreen->SetPixel(coord.first,coord.second, color);
}

void B_StmtPRESET(bool step,double x,double y, int color)
{
    if(color == -1) color = CurScreen->GetBGC();
    B_StmtPSET(step,x,y, color);
}

int B_FuncPOINT(double x,double y)
{
    std::pair<int,int> coord = CurScreen->TranslateCoord(false,x,y);
    return CurScreen->GetPixel(coord.first,coord.second);
}

const std::string B_FuncINKEY_S(bool wait)
{
    return CurScreen->Inkeys(wait);
}

int B_ScancodeInput()
{
    return CurScreen->RawInput();
}
