#ifndef bqtQBlibIOemuHH
#define bqtQBlibIOemuHH

#include <sys/io.h>
# define inportb(x) inb(x)
# define outportb(p,x) outb(x,p)
#if defined(__BORLANDC__) || defined(DJGPP)
# include <dos.h>
#endif

#include <vector>
#include <sched.h> // for sched_yield()
#include <sys/time.h>

#include "error.hh"
#include "gfx.hh"
#include "ioemu.hh"
#include "play.hh"

/************/

static bool DO_VGA_PALETTE_EMULATION = true;
static bool DO_VGA_TEXTMODE_EMULATION = true;
static bool DO_VGA_MODE13_EMULATION = true;
static bool DO_ADLIB_EMULATION = true;
static bool DO_BIOSMEM_EMULATION = true; // FIXME: unimplemented (should at least do &H46C)
static bool DO_KEYBOARD_EMULATION = true;

int B_DefSeg = 0;

class IOEmulator
{
public:
    virtual ~IOEmulator() { }
    virtual bool DoINP(int /*port*/, int& /*result*/) { return false; }
    virtual bool DoOUT(int /*port*/, int /*value*/)   { return false; }
    virtual bool DoPEEK(int /*addr*/, int& /*result*/) { return false; }
    virtual bool DoPOKE(int /*addr*/, int /*value*/)  { return false; }
};

static class VGA_Palette_Emulator: public IOEmulator
{
private:
    unsigned palette_color_index=0, subindex=0;
    unsigned palette_value=0;
public:
    virtual bool DoINP(int port, int& result)
    {
        if(!DO_VGA_PALETTE_EMULATION) return false;
        switch(port)
        {
            case 0x3C9:
                result = 0; // FIXME
                return true;
        }
        return false;
    }
    virtual bool DoOUT(int port, int value)
    {
        if(!DO_VGA_PALETTE_EMULATION) return false;
        switch(port)
        {
            case 0x3C8:
                palette_color_index=value;
                subindex=0;
                palette_value=0;
                return true;
            case 0x3C9:
                palette_value = ((palette_value << 8) | ((value&63)*255/63)) & 0xFF;
                if(++subindex >= 3)
                {
                    subindex=0;
                    B_StmtPALETTE(palette_color_index++, palette_value);
                    palette_value=0;
                }
                return true;
        }
        return false;
    }
} VGA_Palette_Emulator;

static class VGA_Textmode_Emulator: public IOEmulator
{
public:
    virtual bool DoPEEK(int addr, int& result)
    {
        if(addr >= 0xB8000 && addr <= 0xBAFFF && DO_VGA_TEXTMODE_EMULATION)
        {
            // Do transparent text mode poke emulation
            int offs = addr - 0xB8000;

            // Cannot use B_FuncWIDTH() here, because the terminal
            // width may be random but code usually assumes that
            // we have a width of 80, because that's how DOS works.

            int wid = 80;//B_FuncWIDTH();

            int y = offs / (wid*2);
            int x = (offs / 2) % wid;
            int colorflag = offs % 2;

            //fprintf(stderr, "DoPEEK(%d,%d)\n",x,y);

            if(y < 25 && x < 80)
                result = B_FuncSCREEN(y+1, x+1, colorflag);
            return true;
        }
        return false;
    }
    virtual bool DoPOKE(int addr, int value)
    {
        if(addr >= 0xB8000 && addr <= 0xBAFFF && DO_VGA_TEXTMODE_EMULATION)
        {
            // Do transparent text mode poke emulation
            int offs = addr - 0xB8000;

            int wid = 80;//B_FuncWIDTH();

            int y = offs / (wid*2);
            int x = (offs / 2) % wid;
            int colorflag = offs % 2;

            //Afprintf(stderr, "DoPEEK(%d,%d)\n",x,y);
            if(y < 25 && x < 80)
                B_ScreenPOKE(x+1,y+1,colorflag, value);
            return true;
        }
        return false;
    }
} VGA_Textmode_Emulator;

static class VGA_Mode13_Emulator: public IOEmulator
{
public:
    virtual bool DoPOKE(int addr, int value)
    {
        if(addr >= 0xA0000 && addr <= 0xAFA00 && DO_VGA_MODE13_EMULATION)
        {
            int offs = addr - 0xA0000;
            int x = offs % 320, y = offs / 320;
            B_ScreenPOKE(x, y, false, value);
            return true;
        }
        return false;
    }
    virtual bool DoPEEK(int addr, int& result)
    {
        if(addr >= 0xA0000 && addr <= 0xAFA00 && DO_VGA_MODE13_EMULATION)
        {
            int offs = addr - 0xA0000;
            int x = offs % 320, y = offs / 320;
            result = B_FuncPOINT(x,y);
            return true;
        }
        return false;
    }
} VGA_Mode13_Emulator;

static class BIOSMEM_Emulator: public IOEmulator
{
public:
    virtual bool DoPEEK(int addr, int& result)
    {
        if(addr >= 0x400 && addr <= 0x4FF) { result = 0; return true; }
        return false;
    }
    virtual bool DoPOKE(int addr, int value)
    {
        if(addr >= 0x400 && addr <= 0x4FF) return true;
        return false;
    }
} BIOSMEM_Emulator;

static class Adlib_Emulator: public IOEmulator
{
public:
    virtual bool DoINP(int port, int& result)
    {
        if(port < 0x388 || port > 0x38B || !DO_ADLIB_EMULATION) return false;
        result = B_AdlibIn(port&3);
        return true;
    }

    virtual bool DoOUT(int port, int value)
    {
        if(port < 0x388 || port > 0x38B || !DO_ADLIB_EMULATION) return false;
        B_AdlibOut(port&3, value);
        return true;
    }
} Adlib_Emulator;

static class KeyboardEmulator: public IOEmulator
{
public:
    virtual bool DoINP(int port, int& result)
    {
        if(port == 0x60 && DO_KEYBOARD_EMULATION)
        {
            result = B_ScancodeInput();
            return true;
        }
        return false;
    }
} Keyboard_Emulator;


/************/

extern "C" {

static void GetPerm(int port)
{
    static std::vector<bool> permissions(0x400, false);
    if(permissions[port])
    {
        return;
    }
    int i = ioperm(port, 1, 1);
    if(i == -1)
        B_ERROR_Errno();
    permissions[port] = true;
}

int B_FuncINP(int port)
{
    { int res;
      if(VGA_Palette_Emulator.DoINP(port,res)) return res;
      if(Adlib_Emulator.DoINP(port,res)) return res;
      if(Keyboard_Emulator.DoINP(port,res)) return res;
    }

    GetPerm(port);
    return inportb(port);
}

void B_StmtOUT(int port, int value)
{
    { if(VGA_Palette_Emulator.DoOUT(port,value)) return;
      if(Adlib_Emulator.DoOUT(port,value)) return;
    }

    GetPerm(port);
    outportb(port, value);
}

void B_StmtWAIT(int port, int andvalue, int xorvalue=0)
{
    for(;;)
    {
        int value = (B_FuncINP(port) & andvalue) ^ xorvalue;
        if(value != 0) break;
        sched_yield();
    }
}

int B_FuncPEEK(int addr)
{
    addr += B_DefSeg * 16;

    { int res=0;
      if(VGA_Textmode_Emulator.DoPEEK(addr,res)) return res;
      if(VGA_Mode13_Emulator.DoPEEK(addr,res)) return res;
      if(BIOSMEM_Emulator.DoPEEK(addr,res)) return res;
    }
    B_ERROR(70); // Permission denied
}

void B_StmtPOKE(int addr, int value)
{
    addr += B_DefSeg * 16;

    { if(VGA_Textmode_Emulator.DoPOKE(addr,value)) return;
      if(VGA_Mode13_Emulator.DoPOKE(addr,value)) return;
      if(BIOSMEM_Emulator.DoPOKE(addr,value)) return;
    }

    B_ERROR(70); // Permission denied
}

}

#endif
