#ifndef bqtStmtHH
#define bqtStmtHH

#include <list>
#include <memory>
#include <functional>

#include "tokens.hh"
#include "expr.hh"
#include "lexer.hh"

typedef std::unique_ptr<class stmt> stmtp;

class stmt
{
    stmtp prev;
    LexLocation lexloc;
public:
    stmt(const LexLocation& ll);
    virtual ~stmt();

    void set_prev(stmtp&& p);

    const stmtp& get_prev() const { return prev; }

    virtual bool MayErr() const { return true; }

    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const { func(prev.get()); }
    virtual void ForEachSubExpr(const expr_visitor& ) const {  }

    const LexLocation& GetLoc() const { return lexloc; }
};

class stmt_if: public stmt
{
public:
    exprp condition;
    stmtp thenp, elsep;
public:
    stmt_if(const LexLocation& ll, exprp&& c): stmt(ll), condition(std::move(c)) { }

    void AddBody(stmtp&& thp, std::list<std::pair<exprp, stmtp>>&& elseifp, stmtp&& elsp);

    virtual std::string Compile() const;
    virtual bool MayErr() const { return false; } // Structural statement: Errors are handled separately 

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};
class stmt_select: public stmt
{
public:
    exprp ref;
    std::list<
        std::pair<exprp, stmtp>
             > cases;
public:
    stmt_select(const LexLocation& ll, exprp&& r): stmt(ll), ref(std::move(r)) { }

    virtual std::string Compile() const;
    virtual bool MayErr() const { return false; } // Structural statement: Errors are handled separately 

    bool IsEntirelyIntegerCase() const;
    bool IsEntirelyGotoCase() const;
    bool IsEntirelyGosubCase() const;
    bool IsEntirelyStringCase() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};
class stmt_doloop: public stmt
{
public:
    exprp precondition;
    exprp postcondition;
    stmtp bodyp;
    bool is_while;
public:
    stmt_doloop(const LexLocation& ll): stmt(ll), is_while(false) { }

    virtual std::string Compile() const;
    virtual bool MayErr() const { return false; } // Structural statement: Errors are handled separately 

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
    void DisableEXITDO() { is_while = true; }
};
class stmt_for: public stmt
{
public:
    exprp var;
    exprp begin, end, step;
    stmtp bodyp;
    bool begin0step1loop;
public:
    stmt_for(const LexLocation& ll, exprp&& v,
                                    exprp&& b,
                                    exprp&& e)
        : stmt(ll), var(std::move(v)), begin(std::move(b)), end(std::move(e)),
          begin0step1loop(false)
    {
    }

    void Optimize()
    {
        // Optimize a common use-case: When the loop runs from 0 to n,
        // there is a likelihood that n is an expression of type: limit-1
        // we probably want to do 0 < limit rather than 0 <= limit-1
        // as the for-condition.
        if(auto bv = dynamic_cast<expr_number*> (&*begin))
            if(auto sv = dynamic_cast<expr_number*> (&*step))
                if(bv->get_f() == 0 && sv->get_f() == 1)
                {
                    end.reset( new expr_binary(PLUS, std::move(end), exprp(new expr_number(1, TypeUbyte))) );
                    OptimizeExprp(end);
                    begin0step1loop = true;
                }
    }

    virtual std::string Compile() const;
    virtual bool MayErr() const { return false; } // Structural statement: Errors are handled separately 

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};
class stmt_labelref: public stmt
{
public:
    QBtoken token;
    // GOTO, GOSUB or ON_ERROR, or RESTORE
    // Tokens:
    //  GOTO, GOSUB, ON_ERROR, RESTORE
    //  RETURN
    //  EXITDEF, EXITDO, EXITSUB, EXITFUNCTION, EXITFOR
    //  RESUME, RESUMENEXT
    //  STOP, END
    exprp label; // May only be expr_label
    // label only used by GOTO, GOSUB, ON_ERROR, RESTORE,
    //                    ON_TIMER, ON_KEY, ON_PLAY
    //                    RESUME and RETURN
    exprp param = nullptr;

    stmt_labelref(const LexLocation& ll, QBtoken tok, exprp&& l = nullptr) :
        stmt(ll), token(tok), label(std::move(l)) { }

    void set_label(exprp&& l) { label = std::move(l); }

    virtual bool MayErr() const
    {
        return token == STOP
            || token == END
            || token == RESUME
            || token == RETURN
            || (param && param->MayErr());
    }

    void set_params(std::vector<exprp>&& p)
    {
        //if(p.size() >= 1)
        label = p.size() >= 1 ? std::move(p[p.size()-1]) : nullptr; // last param
        param = p.size() >= 2 ? std::move(p[0])          : nullptr;
    }

    void set_lvalue_map(const std::vector<bool>& ) {}
    void set_pattern(const std::string&) { }

public:
    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};
class stmt_let: public stmt
{
    exprp target;
    exprp value;
public:
    stmt_let(const LexLocation& ll, exprp&& t, exprp&& v): stmt(ll), target(std::move(t)), value(std::move(v)) { }
public:
    virtual std::string Compile() const;
    virtual bool MayErr() const { return target->MayErr() || value->MayErr(); }

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;

    exprp& get_target() { return target; }
    exprp& get_value()  { return value; }
};

class stmt_print: public stmt // PRINT, LPRINT, PRINT #, WRITE #, PRINT USING.
{
public:
    exprp target;
    bool is_lprint; // Printer?
    exprp using_string;

    std::list<std::pair<QBtoken, exprp> > params;

    // Note: When parsing the USING format, SPC/TAB/CRLF do not consume a pattern.
public:
    stmt_print(const LexLocation& ll) : stmt(ll), is_lprint(false) { }

    void SetFile(exprp&& p)       { target = std::move(p); }
    void SetLPRINT()              { is_lprint = true; }
    void SetFormat(exprp&& p)     { using_string = std::move(p); }
    void AppendExpr(exprp&& p)    { params.emplace_back((QBtoken)0, std::move(p)); OptimizeTail(); }

    // Cursor position control:
    void AppendSPCfunc(exprp&& p) { params.emplace_back(SPC, std::move(p)); }
    void AppendTABfunc(exprp&& p) { params.emplace_back(TAB, std::move(p)); }
    void AppendTab()     { params.emplace_back(TAB,  nullptr); } // The ',' operator
    void AppendNewline() { params.emplace_back(CRLF, nullptr); }
    // For number printing:
    void AppendSpace()   { AppendExpr(std::unique_ptr<expr>(new expr_string(" "))); }
    // For WRITE:
    void AppendQuoteCh() { AppendExpr(std::unique_ptr<expr>(new expr_string("\""))); }
    void AppendCommaCh() { AppendExpr(std::unique_ptr<expr>(new expr_string(","))); }
private:
    void OptimizeTail();
public:
    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class stmt_input: public stmt // INPUT, INPUT #, LINE INPUT, LINE INPUT #
{
public:
    exprp source;
    bool is_linput; // Line input? Line input has no prompt.
    exprp prompt;   // Default prompt is "?". Not used for LINEINPUT.
    bool keep_cursor;
    std::list<exprp> params;
public:
    stmt_input(const LexLocation& ll) : stmt(ll), is_linput(false), keep_cursor(false) { }

    void SetFile(exprp&& p)      { source = std::move(p); }
    void SetLINEINPUT()          { is_linput = true; }
    void SetPrompt(exprp&& p)    { prompt = std::move(p); }
    void SetKeepCursor()         { keep_cursor=true; }
    void AppendTarget(exprp&& p) { params.emplace_back(std::move(p)); }
public:
    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class stmt_fileopen: public stmt
{
    exprp filename;
    exprp filenumber;
    QBtoken mode;   // RANDOM(access=1-3),OUTPUT(access=2),INPUT(access=1),APPEND(access=2),BINARY(access=1-3)
    exprp mode_str; // Evaluated at runtime, used in GW-BASIC syntax
    int access;     // 0=mode_str is used,    1=read, 2=write, 3=read+write
    int locking;    // 0=undefined or shared, 1=read, 2=write, 3=read+write
    exprp reclen;   // LEN=expression
public:
    stmt_fileopen(const LexLocation& ll, exprp&& fn,exprp&& num, exprp&& mstr, int ac, int lk, exprp&& rl)
        : stmt(ll), filename(std::move(fn)),filenumber(std::move(num)), mode_str(std::move(mstr)), access(ac), locking(lk), reclen(std::move(rl))
    {
    }
    stmt_fileopen(const LexLocation& ll, exprp&& fn,exprp&& num, QBtoken md, int ac, int lk, exprp&& rl)
        : stmt(ll), filename(std::move(fn)),filenumber(std::move(num)), mode(md), access(ac), locking(lk), reclen(std::move(rl))
    {
    }
public:
    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class stmt_redim: public stmt
{
    exprp target;
    std::vector<std::pair<exprp, exprp>> dimensions;
public:
    stmt_redim(const LexLocation& ll, exprp&& t, std::vector<std::pair<exprp,exprp>>&& d)
        : stmt(ll), target(std::move(t)), dimensions(std::move(d))
    {
    }
public:
    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class stmt_erase: public stmt
{
    exprp target;
public:
    stmt_erase(const LexLocation& ll, exprp&& t) : stmt(ll), target(std::move(t))
    {
    }
public:
    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class stmt_subcall: public stmt
{
    std::string name;
    std::vector<exprp> params;
public:
    stmt_subcall(const LexLocation& ll, const std::string& n) : stmt(ll), name(n) { }
    stmt_subcall(const LexLocation& ll, const std::string& n,
                 std::vector<exprp>&& p) : stmt(ll), name(n), params(std::move(p)) { }
public:
    const std::string& getname() const { return name; }

    virtual bool MayErr() const
    {
        for(size_t a=0, b=params.size(); a<b; ++a)
            if(params[a]->MayErr()) return true;
        return false;
    }

    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class stmt_label: public stmt
{
    std::string label;
public:
    stmt_label(const LexLocation& ll) : stmt(ll), label(0) { }
    stmt_label(const LexLocation& ll, const std::string& l) : stmt(ll), label(l) { }

public:
    virtual std::string Compile() const;
    virtual bool MayErr() const { return false; }

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};

class stmt_lib_stmt: public stmt
{
    QBtoken token;
    std::vector<exprp> params;
    std::vector<bool> lvaluemap;
    std::string pattern;
public:
    /* Note: Params must not be written into */

    stmt_lib_stmt(const LexLocation& ll, QBtoken t, const std::string& pat = {})
        : stmt(ll), token(t), pattern(pat) {}
    stmt_lib_stmt(const LexLocation& ll, QBtoken t,exprp&& p, const std::string& pat = {})
        : stmt(ll), token(t), pattern(pat) { params.emplace_back(std::move(p)); }
    stmt_lib_stmt(const LexLocation& ll, QBtoken t,exprp&& p1,exprp&& p2, const std::string& pat = {})
        : stmt(ll), token(t), pattern(pat) {params.emplace_back(std::move(p1));
                                            params.emplace_back(std::move(p2));}
    stmt_lib_stmt(const LexLocation& ll, QBtoken t,exprp&& p1,exprp&& p2,exprp&& p3, const std::string& pat = {})
        : stmt(ll), token(t), pattern(pat) {params.emplace_back(std::move(p1));
                                            params.emplace_back(std::move(p2));
                                            params.emplace_back(std::move(p3));}
    stmt_lib_stmt(const LexLocation& ll, QBtoken t,exprp&& p1,exprp&& p2,exprp&& p3,exprp&& p4, const std::string& pat = {})
        : stmt(ll), token(t), pattern(pat) {params.emplace_back(std::move(p1));
                                            params.emplace_back(std::move(p2));
                                            params.emplace_back(std::move(p3));
                                            params.emplace_back(std::move(p4));}

    void set_params(std::vector<exprp>&& p) { params=std::move(p); }
    void set_lvalue_map(const std::vector<bool>& l) { lvaluemap=l; }
    void set_pattern(const std::string& p) { pattern=p; }

public:
    virtual std::string Compile() const;

    virtual void ForEachSubStmt(const std::function<void(const stmt*)>& func) const;
    virtual void ForEachSubExpr(const expr_visitor& func) const;
};



#endif
