#include "types.hh"
#include "util.hh"

#include "dprintf.hh"

static std::string DefVarType[26+1]; // A-Z

void SetDefVarType(const std::string& charlist, const std::string& type)
{
    //fprintf(stderr, "DEF-%s characters %s\n", type.c_str(), charlist.c_str());

    for(char c: charlist)
    {
        if(c >= 'a' && c <= 'z') c += 'A' - 'a';
        if(c >= 'A' && c <= 'Z') DefVarType[c - 'A'] = type;
    }
}

const TypeP TypeInt = std::make_shared<TypeInfo>(TypeType::Integer);
const TypeP TypeSng = std::make_shared<TypeInfo>(TypeType::Single);
const TypeP TypeDbl = std::make_shared<TypeInfo>(TypeType::Double);
const TypeP TypeLng = std::make_shared<TypeInfo>(TypeType::Long);
const TypeP TypeStr = std::make_shared<TypeInfo>(TypeType::String);

const TypeP TypeByte  = std::make_shared<TypeInfo>(TypeType::Byte);
const TypeP TypeUbyte = std::make_shared<TypeInfo>(TypeType::UnsignedByte);
const TypeP TypeUint  = std::make_shared<TypeInfo>(TypeType::UnsignedInteger);
const TypeP TypeUlong = std::make_shared<TypeInfo>(TypeType::UnsignedLong);

bool TypeInfo::operator==(const TypeInfo& b) const
{
    return Type==b.Type
        && arraytype==b.arraytype
        && fixlength==b.fixlength
        && dimensions==b.dimensions
        && !members==!b.members
        && (!members || *members == *b.members)
      ;
}


const std::string TypeInfo::Dump() const
{
    std::string res;
    switch(Type)
    {
        case TypeType::Integer: res += "INTEGER"; break;
        case TypeType::Long:    res += "LONG"; break;
        case TypeType::Single:  res += "SINGLE"; break;
        case TypeType::Double:  res += "DOUBLE"; break;
        case TypeType::String:  res += "STRING"; break;

        case TypeType::Bit:              res += "_BIT"; break;
        case TypeType::Byte:             res += "_BYTE"; break;
        case TypeType::UnsignedBit:      res += "_UNSIGNED _BIT"; break;
        case TypeType::UnsignedByte:     res += "_UNSIGNED _BYTE"; break;
        case TypeType::UnsignedInteger:  res += "_UNSIGNED INTEGER"; break;
        case TypeType::UnsignedLong:     res += "_UNSIGNED LONG"; break;

        case TypeType::UserType:
        {
            res += "{";
            if(!members) res += "NULL";
            else for(std::size_t a=0; a<members->size(); ++a)
            {
                if(a) res += ',';
                res += (*members)[a].first;
                res += ':';
                TypeP t = (*members)[a].second;
                res += t ? t->Dump() : "NULL";
            }
            res += "}";
            break;
        }
    }
    if(fixlength) res += "*" + std::to_string(fixlength);

    switch(arraytype)
    {
        case TypeInfo::DynamicArray: res = "$DYNAMIC " + res; goto DoArray;
        case TypeInfo::StaticArray: res = "$STATIC " + res; goto DoArray;
        case TypeInfo::ParamArray: res = "$MAGICAL " + res; goto DoArray;
        DoArray:
            res += "(";
            for(unsigned a=0; a<dimensions.size(); ++a)
            {
                if(a>0) res += ",";
                res += std::to_string(dimensions[a].first);
                res += " TO ";
                res += std::to_string(dimensions[a].first + dimensions[a].second - 1);
            }
            res += ")";
            break;
       case TypeInfo::NotArray: ;
    }
    return res;
}

bool UserType::operator==(const UserType& b)const
{
    for(std::size_t a=0; a<size(); ++a)
        if( (*this)[a].first   != b[a].first
        || !(*this)[a].second  != !b[a].second
        || (b[a].second && *(*this)[a].second != *b[a].second)
          )
            return false;

     return Names == b.Names;
}

bool UserType::DeclareMember(const std::string& name, TypeP p)
{
    const std::string basename = UCase(StripTypeSuffix(name));

    if(Names.find(basename) != Names.end()) return false;
    Names.insert(std::make_pair(basename, size()));
    (*this).push_back(std::make_pair(basename, p));
    return true;
}

TypeP UserType::GetMember(const std::string& name) const
{
    const std::string basename = UCase(StripTypeSuffix(name));

    auto i = Names.find(basename);
    if(i != Names.end())
    {
        auto a = i->second;
        if(a < size())
            if( (*this)[a].first == basename) return (*this)[a].second;
    }
    return 0;
}



bool HasTypeSuffix(const std::string& varname)
{
    return !CalcTypeSuffix(varname, false).empty();
}

std::string CalcTypeSuffix(const std::string& name, bool force)
{
    std::size_t suffix_length = 0;
    bool width_ok = false;

    unsigned width_width=0;
    while(name.size() > suffix_length)
    {
        char c = name[name.size() - 1 - suffix_length];
        if(c < '0' || c > '9') break;
        ++width_width;
        ++suffix_length;
    }

    while(name.size() > suffix_length)
    {
        char c = name[name.size() - 1 - suffix_length];
        // Possible suffices:
        //   ! single
        //   $ string
        //   # double
        //   % integer
        //   %% _byte
        //   & long
        //   ` _bit
        //   ~% _unsigned integer
        //   ~%% _unsigned _byte
        //   ~& _unsigned long
        //   ~` _unsigned _bit
        if(c == '!' || c == '#') { ++suffix_length; break; }
        if(c == '$' || c == '~') { width_ok = true; ++suffix_length; break; }
        if(c == '%' || c == '&') { ++suffix_length; continue; }
        break;
    }

    if(!width_ok && width_width) suffix_length = 0;

    std::string suffix = name.substr(name.size() - suffix_length);
    if(suffix.empty() && force)
    {
        if(name[0] == '_')
        {
            dprintf(dp_warnings, "Warning: Unrecognized QB64 extension identifier '%s', treating as INTEGER variable\n", name.c_str());
            return "%";
        }
        suffix += DefVarType[ toupper(name[0]) - 'A' ];
    }
    return suffix;
}

std::string StripTypeSuffix(const std::string& varname)
{
    return varname.substr(0, varname.size() - CalcTypeSuffix(varname,false).size());
}

std::string TypeIntoTypeSuffix(const TypeInfo& type)
{
    std::string result;
    switch(type.Type)
    {
        case TypeType::Single:          result = "!"; break;
        case TypeType::Double:          result = "#"; break;
        case TypeType::String:          result = "$"; break;
        case TypeType::Integer:         result = "%"; break;
        case TypeType::Long:            result = "&"; break;
        case TypeType::UnsignedInteger: result = "~%"; break;
        case TypeType::UnsignedLong:    result = "~&"; break;
        case TypeType::Byte:            result = "%%"; break;
        case TypeType::UnsignedByte:    result = "~%%"; break;
        case TypeType::Bit:             result = "`"; break;
        case TypeType::UnsignedBit:     result = "~`"; break;
        case TypeType::UserType: break;
    }
    if(type.fixlength) result += std::to_string(type.fixlength);
    return result;
}

TypeInfo TypeSuffixIntoType(const std::string& type)
{
    #define m(s,a,b,c) (a)+(b)*0x100+(c)*0x10000+(s)*0x1000000
    TypeInfo t(TypeType::UserType);
    unsigned width_ok = 0;
    switch(m(type.size(), type.size() > 0 ? type[0] : 0,
                          type.size() > 1 ? type[1] : 0,
                          type.size() > 2 ? type[2] : 0))
    {
        case m(1,'!',0,0):   t.Type = TypeType::Single; break;
        case m(1,'#',0,0):   t.Type = TypeType::Double; break;
        case m(1,'$',0,0):   t.Type = TypeType::String; width_ok = 1; break;
        case m(1,'%',0,0):   t.Type = TypeType::Integer; break;
        case m(1,'&',0,0):   t.Type = TypeType::Long; break;
        case m(1,'`',0,0):   t.Type = TypeType::Bit; width_ok = 1; break;
        case m(2,'~','%',0): t.Type = TypeType::UnsignedInteger; break;
        case m(2,'~','&',0): t.Type = TypeType::UnsignedLong; break;
        case m(2,'~','`',0): t.Type = TypeType::UnsignedBit; width_ok = 2; break;
        case m(2,'%','%',0): t.Type = TypeType::Byte; break;
        case m(3,'~','%','%'): t.Type = TypeType::UnsignedByte; break;
    }
    #undef m
    if(width_ok)
        for(; type.size() > width_ok && type[width_ok] >= '0' && type[width_ok] <= '9'; ++width_ok)
            t.fixlength = t.fixlength * 10 + (type[width_ok] - '0');
    //fprintf(stderr, "TypeSuffixIntoType(%s)=%d\n", type.c_str(), (int)t.Type);
    return t;
}

TypeP CreateTypeTemplate(const std::string& name)
{
    std::string suffix = CalcTypeSuffix(name, true);
    TypeInfo t = TypeSuffixIntoType(suffix);
    TypeP result;
    if(t.Type != TypeType::UserType) result = std::make_shared<TypeInfo>(t);
    return result;
}
